/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "project.h"
#include "palette.h"
#include "property.h"
#include "gbwidget.h"
#include "utils.h"
#include "editor.h"

/* The pixmap to use for placeholders */
#include "graphics/placeholder.xpm"
GdkPixmap* placeholder_pixmap;

#define MIN_WIDGET_WIDTH		16
#define MIN_WIDGET_HEIGHT		16
#define MAX_INITIAL_WIDGET_WIDTH	200
#define MAX_INITIAL_WIDGET_HEIGHT	200
#define DEFAULT_WIDGET_WIDTH		50
#define DEFAULT_WIDGET_HEIGHT		50

#define PLACEHOLDER_WIDTH  60
#define PLACEHOLDER_HEIGHT 60

/* The size of the selection handles in the corners of widgets */
#define GB_CORNER_WIDTH   6
#define GB_CORNER_HEIGHT  6

/* The grid (for fixed containers) */
#define GB_GRID_DOTS	1
#define GB_GRID_LINES	2
gboolean		editor_show_grid = TRUE;
static gint		editor_grid_horz_spacing = 8;
static gint		editor_grid_vert_spacing = 8;
static gboolean		editor_grid_style = GB_GRID_DOTS;


/* Snapping to the grid */
#define GB_SNAP_TOP	1 << 1
#define GB_SNAP_BOTTOM	1 << 2
#define GB_SNAP_LEFT	1 << 3
#define GB_SNAP_RIGHT	1 << 4
/*#define GB_SNAP_CENTER	1 << 5  maybe in future*/
gboolean		editor_snap_to_grid   = TRUE;
static gint		editor_snap_to_grid_x = GB_SNAP_LEFT | GB_SNAP_RIGHT;
static gint		editor_snap_to_grid_y = GB_SNAP_TOP | GB_SNAP_BOTTOM;


/* Dragging (in a fixed container) - remembers which part of the widget is
   being dragged, the offset of the mouse (used when moving) and the initial
   widget rectangle (used when resizing) */
#define GB_DRAG_NONE    1
#define GB_TOP_LEFT     2
#define GB_TOP_RIGHT    3
#define GB_BOTTOM_LEFT  4
#define GB_BOTTOM_RIGHT 5
#define GB_MIDDLE       6
static gint		drag_action;
static GtkWidget       *dragging_widget = NULL;
static gint		drag_offset_x;
static gint		drag_offset_y;
static gint		drag_widget_x1, drag_widget_y1,
			drag_widget_x2, drag_widget_y2;

/* The list of selected widgets */
static GList	       *selected_widgets = NULL;

/* The cursors used when selecting/adding/moving/resizing widgets */
static GdkCursor       *cursor_selector;
static GdkCursor       *cursor_add_widget;
static GdkCursor       *cursor_add_to_fixed;
static GdkCursor       *cursor_move;
static GdkCursor       *cursor_top_left;
static GdkCursor       *cursor_top_right;
static GdkCursor       *cursor_bottom_left;
static GdkCursor       *cursor_bottom_right;



/* Static functions */
static gint close_window			(GtkWidget	    *widget,
						 GdkEvent	    *event,
						 gpointer	     data);

static gint editor_on_key_press_event		(GtkWidget	    *widget,
						 GdkEventKey	    *event,
						 gpointer	     data);
static gint editor_on_key_release_event		(GtkWidget	    *widget,
						 GdkEventKey	    *event,
						 gpointer	     data);

static gint editor_on_button_press		(GtkWidget	    *widget,
						 GdkEventButton	    *event,
						 gpointer	     data);
static gint editor_on_button_release		(GtkWidget	    *widget,
						 GdkEvent	    *event,
						 gpointer	     data);

static gint editor_on_motion_notify		(GtkWidget	    *widget,
						 GdkEvent	    *event,
						 gpointer	     data);

static gint placeholder_button_press		(GtkWidget	    *parent,
						 GdkEventButton	    *event,
						 gpointer	     data);
static void placeholder_replace			(GtkWidget	    *placeholder);

static gint get_notebook_page			(GtkWidget	    *notebook,
						 GtkWidget	    *widget);

static void add_widget_to_fixed			(GtkWidget	    *parent,
						 gint		     x,
						 gint		     y);
static void add_widget_to_fixed_finish		(GtkWidget	    *new_widget,
						 GbWidgetNewData    *data);

static gint expose_widget			(GtkWidget	    *widget,
						 GdkEventExpose	    *event,
						 gpointer	     data);
static gint draw_widget				(GtkWidget	    *widget,
						 GdkRectangle	    *area,
						 gpointer	     data);
static void draw_widget_focus			(GtkWidget	    *widget,
						 gpointer	     data);
static void paint_widget			(GtkWidget	    *widget);

static void delete				(GtkWidget	    *widget);
static void delete_placeholder			(GtkWidget	    *widget);
static void delete_widget			(GtkWidget	    *widget);


void
editor_init(GtkWidget *main_window)
{
  GdkBitmap *mask;

  /* Create all cursors needed */
  cursor_selector     = gdk_cursor_new(GDK_TOP_LEFT_ARROW);
  cursor_add_widget   = gdk_cursor_new(GDK_PLUS);
  cursor_add_to_fixed = gdk_cursor_new(GDK_TCROSS);
  cursor_move         = gdk_cursor_new(GDK_FLEUR);
  cursor_top_left     = gdk_cursor_new(GDK_TOP_LEFT_CORNER);
  cursor_top_right    = gdk_cursor_new(GDK_TOP_RIGHT_CORNER);
  cursor_bottom_left  = gdk_cursor_new(GDK_BOTTOM_LEFT_CORNER);
  cursor_bottom_right = gdk_cursor_new(GDK_BOTTOM_RIGHT_CORNER);

  /* Create placeholder pixmap */
  placeholder_pixmap = gdk_pixmap_create_from_xpm_d(main_window->window, &mask,
						    &main_window->style->bg[GTK_STATE_NORMAL],
						    placeholder_xpm);
  if (!placeholder_pixmap) g_warning("Couldn't create placeholder pixmap\n");
}


GtkWidget*
editor_new_window ()
{
  GbWidgetNewData *data;
  GtkWidget *window;

  data = g_new(GbWidgetNewData, 1);
  data->widget = NULL;
  data->is_a_popup_window = FALSE;
  window = gb_widget_new("GtkWindow", data);
  gb_widget_free_new_data(data);
  g_return_val_if_fail(window != NULL, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(close_window), NULL);
  return window;
}

GtkWidget*
editor_new_dialog ()
{
  GbWidgetNewData *data;
  GtkWidget *window;

  data = g_new(GbWidgetNewData, 1);
  data->widget = NULL;
  window = gb_widget_new("GtkDialog", data);
  gb_widget_free_new_data(data);
  g_return_val_if_fail(window != NULL, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(close_window), NULL);
  return window;
}

GtkWidget*
editor_new_popup ()
{
  GbWidgetNewData *data;
  GtkWidget *window;

  data = g_new(GbWidgetNewData, 1);
  data->widget = NULL;
  data->is_a_popup_window = TRUE;
  window = gb_widget_new("GtkWindow", data);
  gb_widget_free_new_data(data);
  g_return_val_if_fail(window != NULL, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(close_window), NULL);
  return window;
}

GtkWidget*
editor_new_filesel ()
{
  GbWidgetNewData *data;
  GtkWidget *window;

  data = g_new(GbWidgetNewData, 1);
  data->widget = NULL;
  window = gb_widget_new("GtkFileSelection", data);
  gb_widget_free_new_data(data);
  g_return_val_if_fail(window != NULL, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(close_window), NULL);
  gtk_widget_show(window);
  return window;
}

GtkWidget*
editor_new_colorsel ()
{
  GbWidgetNewData *data;
  GtkWidget *window;

  data = g_new(GbWidgetNewData, 1);
  data->widget = NULL;
  window = gb_widget_new("GtkColorSelectionDialog", data);
  gb_widget_free_new_data(data);
  g_return_val_if_fail(window != NULL, NULL);
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(close_window), NULL);
  gtk_widget_show(window);
  return window;
}

GtkWidget*
editor_new_popupmenu ()
{
  show_message_box("Sorry - Not implemented yet");
  return NULL;
  /*
  GtkWidget *window = gtk_window_new(GTK_WINDOW_POPUP);
  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(close_window), NULL);
  gtk_widget_show(window);
  return window;
  */
}


gint
close_window (GtkWidget  *widget,
	      GdkEvent   *event,
	      gpointer    data)
{
  gint x, y;
  /* remember position - doesn't work properly yet - moves down each time */
  gdk_window_get_position(widget->window, &x, &y);
  gtk_widget_set_uposition(widget, x, y);
  gtk_widget_hide(widget);
  return TRUE;
}


void
redraw_windows(GtkWidget* widget, gpointer data)
{
  gtk_widget_queue_draw(widget);
}


/*
 * Grid settings
 */
gboolean
editor_get_show_grid()
{
  return editor_show_grid;
}


void
editor_toggle_show_grid(GtkWidget* widget, gpointer data)
{
  editor_show_grid = GTK_CHECK_MENU_ITEM(widget)->active;
  project_foreach_component(redraw_windows, NULL);
}


void
on_grid_options_ok(GtkWidget *widget, gpointer data)
{
  GtkWidget *window, *spinbutton, *button;

  window = gtk_widget_get_toplevel(widget);
  spinbutton = gtk_object_get_data(GTK_OBJECT(window), "spinbutton1");
  editor_grid_horz_spacing = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
  spinbutton = gtk_object_get_data(GTK_OBJECT(window), "spinbutton2");
  editor_grid_vert_spacing = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));

  button = gtk_object_get_data(GTK_OBJECT(window), "button1");
  if (GTK_TOGGLE_BUTTON(button)->active) editor_grid_style = GB_GRID_DOTS;
  else editor_grid_style = GB_GRID_LINES;

  gtk_widget_destroy(window);

  /* redraw all windows */
  project_foreach_component(redraw_windows, NULL);
}

void
editor_show_grid_settings_dialog(GtkWidget* widget, gpointer data)
{
  GtkWidget *window, *vbox, *table, *label, *button, *hbbox;
  GtkWidget *separator, *spinbutton;
  GtkObject *adjustment;
  GSList *group;

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Grid Options");
  gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect_object (GTK_OBJECT (window), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (window));

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_container_border_width (GTK_CONTAINER (vbox), 10);
  gtk_widget_show(vbox);

  table = gtk_table_new (2, 3, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 5);
  gtk_widget_show(table);

  label = gtk_label_new ("Horizontal Spacing:");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE (table), label, 0, 1, 0, 1,
		   GTK_FILL, 0, 5, 5);
  gtk_widget_show(label);

  adjustment = gtk_adjustment_new(editor_grid_horz_spacing, 1, 1000, 1, 10, 10);
  spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 0, 0);
  gtk_table_attach(GTK_TABLE (table), spinbutton, 1, 3, 0, 1,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(spinbutton);
  gtk_object_set_data(GTK_OBJECT(window), "spinbutton1", spinbutton);

  label = gtk_label_new ("Vertical Spacing:");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE (table), label, 0, 1, 1, 2,
		   GTK_FILL, 0, 5, 5);
  gtk_widget_show(label);

  adjustment = gtk_adjustment_new(editor_grid_vert_spacing, 1, 1000, 1, 10, 10);
  spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 0, 0);
  gtk_table_attach(GTK_TABLE (table), spinbutton, 1, 3, 1, 2,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(spinbutton);
  gtk_object_set_data(GTK_OBJECT(window), "spinbutton2", spinbutton);

  table = gtk_table_new (1, 3, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
  gtk_widget_show(table);

  label = gtk_label_new ("Grid Style:");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE (table), label, 0, 1, 0, 1,
		   GTK_FILL, 0, 5, 5);
  gtk_widget_show(label);

  button = gtk_radio_button_new_with_label(NULL, "Dots");
  if (editor_grid_style == GB_GRID_DOTS)
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_table_attach(GTK_TABLE (table), button, 1, 2, 0, 1,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(button);
  gtk_object_set_data(GTK_OBJECT(window), "button1", button);
  group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));

  button = gtk_radio_button_new_with_label(group, "Lines");
  if (editor_grid_style == GB_GRID_LINES)
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_table_attach(GTK_TABLE (table), button, 2, 3, 0, 1,
		   GTK_EXPAND | GTK_FILL, 0, 5, 1);
  gtk_widget_show(button);
  gtk_object_set_data(GTK_OBJECT(window), "button2", button);

  separator = gtk_hseparator_new();
  gtk_box_pack_start (GTK_BOX (vbox), separator, TRUE, TRUE, 5);
  gtk_widget_show(separator);

  /* Buttons at bottom */
  hbbox = gtk_hbutton_box_new();
  gtk_box_pack_start (GTK_BOX (vbox), hbbox, TRUE, TRUE, 2);
  gtk_widget_show(hbbox);

  button = gtk_button_new_with_label("OK");
  gtk_container_add(GTK_CONTAINER(hbbox), button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_grid_options_ok),
		      NULL);
  
  button = gtk_button_new_with_label("Cancel");
  gtk_container_add(GTK_CONTAINER(hbbox), button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(window));

  gtk_widget_show(window);
}


/*
 * Snap settings
 */
gboolean
editor_get_snap_to_grid()
{
  return editor_snap_to_grid;
}


void
editor_toggle_snap_to_grid(GtkWidget* widget, gpointer data)
{
  editor_snap_to_grid = GTK_CHECK_MENU_ITEM(widget)->active;
}


void
on_snap_options_ok(GtkWidget *widget, gpointer data)
{
  GtkWidget *window, *button;

  window = gtk_widget_get_toplevel(widget);
  editor_snap_to_grid_x = 0;
  editor_snap_to_grid_y = 0;
  button = gtk_object_get_data(GTK_OBJECT(window), "button1");
  if (GTK_TOGGLE_BUTTON(button)->active)
    editor_snap_to_grid_x |= GB_SNAP_LEFT;
  button = gtk_object_get_data(GTK_OBJECT(window), "button2");
  if (GTK_TOGGLE_BUTTON(button)->active)
    editor_snap_to_grid_x |= GB_SNAP_RIGHT;
  button = gtk_object_get_data(GTK_OBJECT(window), "button3");
  if (GTK_TOGGLE_BUTTON(button)->active)
    editor_snap_to_grid_y |= GB_SNAP_TOP;
  button = gtk_object_get_data(GTK_OBJECT(window), "button4");
  if (GTK_TOGGLE_BUTTON(button)->active)
    editor_snap_to_grid_y |= GB_SNAP_BOTTOM;

  gtk_widget_destroy(window);
}

void
editor_show_snap_settings_dialog(GtkWidget* widget, gpointer data)
{
  GtkWidget *window, *vbox, *table, *label, *button, *hbbox, *separator;

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "Snap Options");
  gtk_window_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect_object (GTK_OBJECT (window), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (window));

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_container_border_width (GTK_CONTAINER (vbox), 10);
  gtk_widget_show(vbox);

  table = gtk_table_new (4, 2, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 5);
  gtk_widget_show(table);

  /* Horizontal snapping */
  label = gtk_label_new ("Horizontal Snapping:");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE (table), label, 0, 1, 0, 1,
		   GTK_FILL, 0, 5, 5);
  gtk_widget_show(label);

  button = gtk_check_button_new_with_label("Left");
  if (editor_snap_to_grid_x & GB_SNAP_LEFT)
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_table_attach(GTK_TABLE (table), button, 1, 2, 0, 1,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(button);
  gtk_object_set_data(GTK_OBJECT(window), "button1", button);

  button = gtk_check_button_new_with_label("Right");
  if (editor_snap_to_grid_x & GB_SNAP_RIGHT)
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_table_attach(GTK_TABLE (table), button, 1, 2, 1, 2,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(button);
  gtk_object_set_data(GTK_OBJECT(window), "button2", button);

  /* Vertical snapping */
  label = gtk_label_new ("Vertical Snapping:");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE (table), label, 0, 1, 2, 3,
		   GTK_FILL, 0, 5, 5);
  gtk_widget_show(label);

  button = gtk_check_button_new_with_label("Top");
  if (editor_snap_to_grid_y & GB_SNAP_TOP)
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_table_attach(GTK_TABLE (table), button, 1, 2, 2, 3,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(button);
  gtk_object_set_data(GTK_OBJECT(window), "button3", button);

  button = gtk_check_button_new_with_label("Bottom");
  if (editor_snap_to_grid_y & GB_SNAP_BOTTOM)
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(button), TRUE);
  gtk_table_attach(GTK_TABLE (table), button, 1, 2, 3, 4,
		   GTK_EXPAND | GTK_FILL, 0, 5, 5);
  gtk_widget_show(button);
  gtk_object_set_data(GTK_OBJECT(window), "button4", button);


  separator = gtk_hseparator_new();
  gtk_box_pack_start (GTK_BOX (vbox), separator, TRUE, TRUE, 5);
  gtk_widget_show(separator);

  /* Buttons at bottom */
  hbbox = gtk_hbutton_box_new();
  gtk_box_pack_start (GTK_BOX (vbox), hbbox, TRUE, TRUE, 2);
  gtk_widget_show(hbbox);

  button = gtk_button_new_with_label("OK");
  gtk_container_add(GTK_CONTAINER(hbbox), button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_snap_options_ok),
		      NULL);
  
  button = gtk_button_new_with_label("Cancel");
  gtk_container_add(GTK_CONTAINER(hbbox), button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(window));

  gtk_widget_show(window);
}

gint
snap_top_edge(gint y)
{
  if (editor_snap_to_grid && (editor_snap_to_grid_y & GB_SNAP_TOP)) {
    y += editor_grid_vert_spacing / 2;
    y -= y % editor_grid_vert_spacing;
  }
  return y;
}

gint
snap_bottom_edge(gint y)
{
  if (editor_snap_to_grid && (editor_snap_to_grid_y & GB_SNAP_BOTTOM)) {
    y += editor_grid_vert_spacing / 2;
    y -= y % editor_grid_vert_spacing;
  }
  return y;
}

gint
snap_left_edge(gint x)
{
  if (editor_snap_to_grid && (editor_snap_to_grid_x & GB_SNAP_LEFT)) {
    x += editor_grid_horz_spacing / 2;
    x -= x % editor_grid_horz_spacing;
  }
  return x;
}

gint
snap_right_edge(gint x)
{
  if (editor_snap_to_grid && (editor_snap_to_grid_x & GB_SNAP_RIGHT)) {
    x += editor_grid_horz_spacing / 2;
    x -= x % editor_grid_horz_spacing;
  }
  return x;
}


GtkWidget*
editor_new_placeholder()
{
  GtkWidget *placeholder = gtk_drawing_area_new();
  gtk_widget_set_events(placeholder,
			gtk_widget_get_events(placeholder)
			| GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK
			| GDK_BUTTON_RELEASE_MASK
			| GDK_POINTER_MOTION_MASK | GDK_BUTTON1_MOTION_MASK);
  gtk_widget_set_usize(placeholder, PLACEHOLDER_WIDTH, PLACEHOLDER_HEIGHT);
  gtk_widget_show(placeholder);
  editor_add_draw_signals(placeholder);
  gtk_signal_connect_after (GTK_OBJECT (placeholder), "button_press_event",
		      GTK_SIGNAL_FUNC(placeholder_button_press), &placeholder);
  /*gtk_signal_connect_after (GTK_OBJECT (placeholder), "button_release_event",
		      GTK_SIGNAL_FUNC(button_release), &placeholder);*/
  gtk_signal_connect_after (GTK_OBJECT (placeholder), "motion_notify_event",
		      GTK_SIGNAL_FUNC(editor_on_motion_notify), &placeholder);
  gtk_object_set_data(GTK_OBJECT(placeholder), GB_PLACEHOLDER_KEY, "True");
  return placeholder;
}


gint
placeholder_button_press(GtkWidget *placeholder,
			 GdkEventButton *event,
			 gpointer data)
{
  if (event->button == 1) {
    if (palette_is_selector_on())
      editor_select_widget(placeholder, event->state, event->x, event->y);
    else placeholder_replace(placeholder);
    return TRUE;
  } else if (event->button == 3) {
    gb_widget_show_popup_menu(placeholder, event);
    return TRUE;
  }
  return FALSE;
}


void
placeholder_finish_replace(GtkWidget *new_widget, GbWidgetNewData *data)
{
  editor_clear_selection(NULL);
  /* Replace placeholder in parent container */
  gb_widget_replace_child(data->widget->parent, data->widget, new_widget);
  gb_widget_show_properties(new_widget);
}


void
placeholder_replace(GtkWidget *placeholder)
{
  char *class_name;
  GtkWidget *new_widget;
  GbWidgetNewData *data;

  class_name = palette_get_widget_class();
  g_return_if_fail(class_name != NULL);
  palette_reset_selection();

  data = g_new(GbWidgetNewData, 1);
  data->callback = placeholder_finish_replace;
  data->widget = placeholder;
  new_widget = gb_widget_new(class_name, data);
  if (new_widget) {
    placeholder_finish_replace(new_widget, data);
    gb_widget_free_new_data(data);
  }
}


/* Struct only used for is_child_at callback */
typedef struct _GbFindChildAtData GbFindChildAtData;

struct _GbFindChildAtData
{
  gint		 x;
  gint		 y;
  GtkWidget	*found_child;
};  
   

/* Note: returns last found child if children overlap */
void
find_child_at(GtkWidget *widget, GbFindChildAtData *data)
{
  if ((widget->allocation.x <= data->x)
      && (widget->allocation.y <= data->y)
      && (widget->allocation.x + widget->allocation.width > data->x)
      && (widget->allocation.y + widget->allocation.height > data->y)) {
    /* Make sure this is a GbWidget, not a child window of a specialized
     widget or a button's label. */
    if (gtk_object_get_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY)) {
      data->found_child = widget;
    }
  }
}


gint
editor_on_button_press(GtkWidget *widget,
		       GdkEventButton *event,
		       gpointer user_data)
{
  GtkWidget *widget_to_select, *real_widget;
  GbFindChildAtData data;
  gint x, y;

  x = event->x;
  y = event->y;
  if (event->button == 1) {
    if (palette_is_selector_on()) {
      /* If widget is a container, see if a child should really be selected,
	 i.e. if the mouse is in the child's allocated area. */
      widget_to_select = real_widget = widget;

      if (GB_IS_HOLDING_WIDGET(widget)) real_widget = GTK_BIN(widget)->child;

      if (GTK_IS_CONTAINER(real_widget)) {
	data.x = x;
	data.y = y;
	data.found_child = NULL;
	gtk_container_foreach(GTK_CONTAINER(real_widget),
			      (GtkCallback) find_child_at, &data);
	if (data.found_child) {
	  widget_to_select = data.found_child;
	  /* Need to translate mouse coords. for widgets with own windows */
	  if (!GTK_WIDGET_NO_WINDOW(widget_to_select)) {
	    x -= widget_to_select->allocation.x;
	    y -= widget_to_select->allocation.y;
	  }
        }
      }
      editor_select_widget(widget_to_select, event->state, x, y);
      return TRUE;
    } else if (GTK_IS_FIXED(widget)) {
      /* but workaround? - need to translate coords if mouse button was pressed
	 in a child widget */
      GdkWindow *event_window;
      gint win_x, win_y;
      event_window = event->window;
      while (event_window && event_window != widget->window) {
	gdk_window_get_position(event_window, &win_x, &win_y);
	x += win_x;
	y += win_y;
	event_window = gdk_window_get_parent(event_window);
      }
      g_return_val_if_fail(event_window == widget->window, FALSE);

      add_widget_to_fixed(widget, x, y);
      return TRUE;
    }
  } else if (event->button == 3) {
    gb_widget_show_popup_menu(widget, event);
    return TRUE;
  }
  return FALSE;
}


gint
editor_on_button_release(GtkWidget *widget,
			 GdkEvent *event,
			 gpointer data)
{
  if (dragging_widget) {
    dragging_widget = NULL;
    drag_action = GB_DRAG_NONE;
    gtk_grab_remove(widget);
  }
  return FALSE;
}


void
add_widget_to_fixed(GtkWidget *parent, gint x, gint y)
{
  char *class_name;
  GtkWidget *new_widget;
  GbWidgetNewData *data;

  class_name = palette_get_widget_class();
  g_return_if_fail(class_name != NULL);

  palette_reset_selection();

  data = g_new(GbWidgetNewData, 1);
  data->callback = add_widget_to_fixed_finish;
  data->widget = parent;
  data->x = x;
  data->y = y;
  new_widget = gb_widget_new(class_name, data);
  if (new_widget) {
    add_widget_to_fixed_finish(new_widget, data);
    gb_widget_free_new_data(data);
  }
}


void
add_widget_to_fixed_finish(GtkWidget *new_widget, GbWidgetNewData *data)
{
  GtkRequisition requisition = { 0, 0 };
  gint x = data->x;
  gint y = data->y;
  gint w, h;

  gtk_widget_size_request(new_widget, &requisition);
  w = requisition.width;
  h = requisition.height;
  if (w == 0) w = DEFAULT_WIDGET_WIDTH;
  if (h == 0) h = DEFAULT_WIDGET_HEIGHT;
  w = MAX(w, MIN_WIDGET_WIDTH);
  w = MIN(w, MAX_INITIAL_WIDGET_WIDTH);
  h = MAX(h, MIN_WIDGET_HEIGHT);
  h = MIN(h, MAX_INITIAL_WIDGET_HEIGHT);

  /* Calculate real position & size according to grid & snapping */
  /* NOTE: only snaps top & left for now */
  if (editor_snap_to_grid) {
    if (editor_snap_to_grid_x & GB_SNAP_LEFT) {
      /*x += editor_grid_horz_spacing / 2;*/
      x -= x % editor_grid_horz_spacing;
    }
    if (editor_snap_to_grid_y & GB_SNAP_TOP) {
      /*y += editor_grid_vert_spacing / 2;*/
      y -= y % editor_grid_vert_spacing;
    }
  }

  gtk_widget_set_usize(new_widget, w, h);
  gtk_widget_set_uposition(new_widget, x, y);
  gtk_widget_show(new_widget);
  gtk_container_add(GTK_CONTAINER(data->widget), new_widget);
  gb_widget_show_properties(new_widget);
}


/*
 * Clears all currently selected widgets, except the given widget.
 * Returns TRUE if given widget is selected.
 */
gint
editor_clear_selection(GtkWidget *leave_widget)
{
  GList *child = selected_widgets, *next_child;
  GtkWidget *widget;
  gint selected = FALSE;

  while(child) {
    next_child = child->next;
    widget = child->data;
    if (widget == leave_widget) {
      selected = TRUE;
    } else {
      selected_widgets = g_list_remove(selected_widgets, widget);
      editor_refresh_widget(widget);
    }
    child = next_child;
  }
  return selected;
}

void
editor_remove_widget_from_selection(GtkWidget *widget)
{
  g_list_remove(selected_widgets, widget);
}


gint
get_position_in_widget(GtkWidget *widget, gint x, gint y)
{
  gint width = widget->allocation.width;
  gint height = widget->allocation.height;
  if (x < GB_CORNER_WIDTH && y < GB_CORNER_HEIGHT)
    return GB_TOP_LEFT;
  if (x > width - GB_CORNER_WIDTH && y < GB_CORNER_HEIGHT)
    return GB_TOP_RIGHT;
  if (x < GB_CORNER_WIDTH && y > height - GB_CORNER_HEIGHT)
    return GB_BOTTOM_LEFT; 
  if (x > width - GB_CORNER_WIDTH && y > height - GB_CORNER_HEIGHT)
    return GB_BOTTOM_RIGHT; 
  return GB_MIDDLE;
}

void
raise_fixed_child(GtkWidget *widget)
{
  /* If widget hasn't got a window, move it to the back of the parent fixed's
     children. If it has got a window, raise it. */
  if (GTK_WIDGET_NO_WINDOW(widget)) {
    GList *children, *child;
    GtkFixedChild *data;
    children = child = GTK_FIXED(widget->parent)->children;
    while (child) {
      data = child->data;
      if (data->widget == widget) {
	GTK_FIXED(widget->parent)->children = g_list_remove(children, data);
	GTK_FIXED(widget->parent)->children = g_list_append(children, data);
	break;
      }
      child = child->next;
    }
  } else {
    gdk_window_raise(widget->window);
  }
}

void
editor_select_widget(GtkWidget *widget, guint modifiers, gint x, gint y)
{
  gint already_selected, page;
  GtkWidget *select_widget, *ancestor;

  /* reset any move/resize action */
  drag_action = GB_DRAG_NONE;

  /* Shift + selection. Step through parents of this widget, until a selected
     widget is found. Then clear selection & select that widgets parent.
     If no parents were selected, or the top-level parent was selected,
     select this widget.
     NOTE: remember to skip widgets with no window, using parent event box.
     */
  if (modifiers & GDK_SHIFT_MASK) {
    gint found = FALSE;
    ancestor = widget;
    while (ancestor) {
      if (g_list_find(selected_widgets, ancestor)) {
	found = TRUE;
	break;
      }
      ancestor = ancestor->parent;
    }
    if (found && ancestor->parent != NULL) {
      select_widget = ancestor->parent;
      /* If widget is not a GbWidget, i.e. it has no GbWidgetData, skip it */
      while (select_widget) {
	if (gtk_object_get_data(GTK_OBJECT(select_widget), GB_WIDGET_DATA_KEY))
	  break;
	select_widget = select_widget->parent;
      }
    } else {
      select_widget = widget;
    }
    g_return_if_fail(select_widget != NULL);

    gb_widget_show_properties(select_widget);
    if (GTK_WIDGET_NO_WINDOW(select_widget))
      select_widget = select_widget->parent;
    g_return_if_fail(select_widget != NULL);
    already_selected = editor_clear_selection(select_widget);
    if (already_selected) return;
    selected_widgets = g_list_append(selected_widgets, select_widget);
    project_clear_component_selection();
    editor_refresh_widget(select_widget);
    return;
  }

  /* Control + selection. If widget is currently selected, deslect it, else
     add it to the selected widgets
     */
  if (modifiers & GDK_CONTROL_MASK) {
    /* If widget is currently selected, deslect it, else add it to selected */
    if (g_list_find(selected_widgets, widget)) {
      selected_widgets = g_list_remove(selected_widgets, widget);
    } else {
      gb_widget_show_properties(widget);
      selected_widgets = g_list_append(selected_widgets, widget);
      project_clear_component_selection();
    }
    editor_refresh_widget(widget);
    return;
  }

  /* Normal selection. Clear all currently selected widgets, except this one.
     Then select this one, if it wasn't already selected.
     Also remember where the button press occurred, in case widget is being
     moved or resized (in a fixed container).
     */
  gb_widget_show_properties(widget);

  already_selected = editor_clear_selection(widget);
  if (!already_selected) {
    selected_widgets = g_list_append(selected_widgets, widget);
    project_clear_component_selection();
    /* If parent is a fixed container, move widget to front */
    if (widget->parent && GTK_IS_FIXED(widget->parent))
      raise_fixed_child(widget);

    /* If the widget or an ancestor is a notebook tab, show the page */
    ancestor = widget;
    while (ancestor->parent) {
      if (GTK_IS_NOTEBOOK(ancestor->parent)) {
	page = get_notebook_page(ancestor->parent, ancestor);
	if (page != -1) gtk_notebook_set_page(GTK_NOTEBOOK(ancestor->parent),
					      page);
	break;
      }
      ancestor = ancestor->parent;
    }

    editor_refresh_widget(widget);
  }
  drag_action = get_position_in_widget(widget, x, y);
  drag_offset_x = x;
  drag_offset_y = y;

  /* NOTE: this will only work in a fixed */
  drag_widget_x1 = widget->allocation.x;
  drag_widget_y1 = widget->allocation.y;
  drag_widget_x2 = drag_widget_x1 + widget->allocation.width;
  drag_widget_y2 = drag_widget_y1 + widget->allocation.height;
}


gint get_notebook_page			(GtkWidget	    *notebook,
					 GtkWidget	    *widget)
{
  GList *children;
  GtkNotebookPage *page;
  gint page_num = 0;

  children = GTK_NOTEBOOK(notebook)->children;
  while (children) {
    page = children->data;
    if (page->tab_label == widget) {
      return page_num;
    }
    children = children->next;
    page_num++;
  }
  return -1;
}


gint
do_drag_action(GtkWidget *widget, GdkEvent *event)
{
  gint mouse_x, mouse_y, new_x = 0, new_y = 0, new_width = 0, new_height = 0;
  GtkWidget *real_widget;

  if (dragging_widget == NULL) {
    dragging_widget = widget;
    gtk_grab_add(widget);
  } else {
    if (dragging_widget != widget) return FALSE;
  }

  gtk_widget_get_pointer(widget->parent, &mouse_x, &mouse_y);
  gtk_container_block_resize(GTK_CONTAINER(widget->parent));

  switch (drag_action) {
  case GB_TOP_LEFT:
    new_x = snap_left_edge(mouse_x);
    new_y = snap_top_edge(mouse_y);
    new_width = drag_widget_x2 - new_x;
    new_height = drag_widget_y2 - new_y;
    if (new_width < MIN_WIDGET_WIDTH) {
      new_width = MIN_WIDGET_WIDTH;
      new_x = drag_widget_x2 - new_width;
    }
    if (new_height < MIN_WIDGET_HEIGHT) {
      new_height = MIN_WIDGET_HEIGHT;
      new_y = drag_widget_y2 - new_height;
    }
    break;

  case GB_TOP_RIGHT:
    new_x = drag_widget_x1;
    new_y = snap_top_edge(mouse_y);
    new_width = snap_right_edge(mouse_x) - new_x;
    new_height = drag_widget_y2 - new_y;
    if (new_width < MIN_WIDGET_WIDTH) {
      new_width = MIN_WIDGET_WIDTH;
    }
    if (new_height < MIN_WIDGET_HEIGHT) {
      new_height = MIN_WIDGET_HEIGHT;
      new_y = drag_widget_y2 - new_height;
    }
    break;

  case GB_BOTTOM_LEFT:
    new_x = snap_left_edge(mouse_x);
    new_y = drag_widget_y1;
    new_width = drag_widget_x2 - new_x;
    new_height = snap_bottom_edge(mouse_y) - new_y;
    if (new_width < MIN_WIDGET_WIDTH) {
      new_width = MIN_WIDGET_WIDTH;
      new_x = drag_widget_x2 - new_width;
    }
    if (new_height < MIN_WIDGET_HEIGHT) {
      new_height = MIN_WIDGET_HEIGHT;
    }
    break;

  case GB_BOTTOM_RIGHT:
    new_x = drag_widget_x1;
    new_y = drag_widget_y1;
    new_width = snap_right_edge(mouse_x) - new_x;
    new_height = snap_bottom_edge(mouse_y) - new_y;
    if (new_width < MIN_WIDGET_WIDTH) {
      new_width = MIN_WIDGET_WIDTH;
    }
    if (new_height < MIN_WIDGET_HEIGHT) {
      new_height = MIN_WIDGET_HEIGHT;
    }
    break;

  case GB_MIDDLE:
    new_x = snap_left_edge(mouse_x - drag_offset_x);
    new_y = snap_top_edge(mouse_y - drag_offset_y);
    new_width = widget->allocation.width;
    new_height = widget->allocation.height;
    if (new_x < 0) new_x = 0;
    if (new_y < 0) new_y = 0;
    break;
  }

  /* Only move/resize widget if values have changed */
  if (new_width != widget->allocation.width
      || new_height != widget->allocation.height) {
    gtk_widget_set_usize(widget, new_width, new_height);
  }
  if (new_x != widget->allocation.x || new_y != widget->allocation.y) {
    /* Do we need both? */
    gtk_fixed_move(GTK_FIXED(widget->parent), widget, new_x, new_y);
    gtk_widget_set_uposition(widget, new_x, new_y);
  }

  gtk_container_unblock_resize(GTK_CONTAINER(widget->parent));

  /* Update properties if widget's properties are currently shown */
  if (GB_IS_HOLDING_WIDGET(widget)) real_widget = GTK_BIN(widget)->child;
  else real_widget = widget;

  if (property_get_widget() == real_widget) {
    property_set_int(GbX, new_x);
    property_set_int(GbY, new_y);
    property_set_int(GbWidth, new_width);
    property_set_int(GbHeight, new_height);
  }

  return TRUE;
}


gint
editor_on_motion_notify(GtkWidget *widget,
			GdkEvent *event,
			gpointer data)
{
  gint x, y, width, height, pos;

  /* Maybe we should remember the current cursor and only set it if changed. */

  if (palette_is_selector_on()) {
    if (widget->parent && GTK_IS_FIXED(widget->parent)) {
      if (drag_action != GB_DRAG_NONE
	  && event->motion.state & GDK_BUTTON1_MASK) {
	return do_drag_action(widget, event);
      } else {
      x = event->motion.x;
      y = event->motion.y;
      width = widget->allocation.width;
      height = widget->allocation.height;

      pos = get_position_in_widget(widget, x, y);
      switch (pos) {
      case GB_TOP_LEFT:
	gdk_window_set_cursor(widget->window, cursor_top_left);
	break;
      case GB_TOP_RIGHT:
	gdk_window_set_cursor(widget->window, cursor_top_right);
	break;
      case GB_BOTTOM_LEFT:
	gdk_window_set_cursor(widget->window, cursor_bottom_left);
	break;
      case GB_BOTTOM_RIGHT:
	gdk_window_set_cursor(widget->window, cursor_bottom_right);
	break;
      case GB_MIDDLE:
	gdk_window_set_cursor(widget->window, cursor_move);
	break;
      }
      }
    } else {
      gdk_window_set_cursor(widget->window, cursor_selector);
    }
  } else {
    if (GTK_IS_FIXED(widget)
	|| (widget->parent && GTK_IS_FIXED(widget->parent)))
      gdk_window_set_cursor(widget->window, cursor_add_to_fixed);
    else
      gdk_window_set_cursor(widget->window, cursor_add_widget);
  }
  return TRUE;
}


/*
 * Adding signals to widgets to allow manipulation, e.g. selecting/drawing
 */

void
editor_add_mouse_signals(GtkWidget *widget)
{
  gtk_widget_set_events(widget, gtk_widget_get_events(widget)
			| GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK
			| GDK_BUTTON_RELEASE_MASK
			| GDK_POINTER_MOTION_MASK | GDK_BUTTON1_MOTION_MASK);
  gtk_signal_connect_after (GTK_OBJECT (widget), "button_press_event",
			    GTK_SIGNAL_FUNC(editor_on_button_press), &widget);
  gtk_signal_connect_after (GTK_OBJECT (widget), "button_release_event",
			    GTK_SIGNAL_FUNC(editor_on_button_release),
			    &widget);
  gtk_signal_connect_after (GTK_OBJECT (widget), "motion_notify_event",
			    GTK_SIGNAL_FUNC(editor_on_motion_notify), &widget);
}

void
editor_add_key_signals(GtkWidget *widget)
{
  gtk_signal_connect_after (GTK_OBJECT (widget), "key_press_event",
			    GTK_SIGNAL_FUNC(editor_on_key_press_event),
			    NULL);
  gtk_signal_connect_after (GTK_OBJECT (widget), "key_release_event",
			    GTK_SIGNAL_FUNC(editor_on_key_release_event),
			    NULL);
}

void
editor_add_draw_signals(GtkWidget *widget)
{
  gtk_signal_connect_after (GTK_OBJECT (widget), "expose_event",
			    GTK_SIGNAL_FUNC(expose_widget), widget);
  gtk_signal_connect_after (GTK_OBJECT (widget), "draw",
			    GTK_SIGNAL_FUNC(draw_widget), widget);

  /* Needed for button, others? */
  gtk_signal_connect_after (GTK_OBJECT (widget), "draw_default",
			    GTK_SIGNAL_FUNC(draw_widget_focus), widget);
  gtk_signal_connect_after (GTK_OBJECT (widget), "draw_focus",
			    GTK_SIGNAL_FUNC(draw_widget_focus), widget);

  /* Needed for scrolled window, clist? & possibly other widgets */
  if (GTK_IS_CONTAINER(widget)) {
    gtk_container_foreach(GTK_CONTAINER(widget),
			  (GtkCallback)editor_add_draw_signals,
			  NULL);
  }
}



gint
expose_widget(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  GtkWidget *holding_widget = widget;
  GbWidgetData *widget_data;

  /* Here we try to make sure that the widget's size & position are displayed
     correctly in the property window when the widget is created. */
  if (property_get_widget() == widget) {
    widget_data = gtk_object_get_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY);
    g_return_val_if_fail(widget_data != NULL, FALSE);
    if (widget_data->width <= 0 && widget_data->height <= 0) {
      if (GTK_WIDGET_NO_WINDOW(widget)) holding_widget = widget->parent;
      property_set_auto_apply(FALSE);
      gb_widget_show_position_properties(widget, holding_widget, widget_data);
      property_set_auto_apply(TRUE);
    }
  }

  paint_widget(widget);
  return FALSE;
}

gint
draw_widget(GtkWidget *widget, GdkRectangle *area, gpointer data)
{
  paint_widget(widget);
  return FALSE;
}

void
draw_widget_focus(GtkWidget *widget, gpointer data)
{
  paint_widget(widget);
}

void
draw_grid(GtkWidget *widget)
{
  GdkGC *gc = widget->style->dark_gc[GTK_STATE_NORMAL];
  gint width = widget->allocation.width;
  gint height = widget->allocation.height;
  gint gridx, gridy;

  if (!editor_show_grid) return;
  if (editor_grid_style == GB_GRID_DOTS) {
    for (gridx = 0; gridx < width; gridx += editor_grid_horz_spacing) {
      for (gridy = 0; gridy < height; gridy += editor_grid_vert_spacing) {
	gdk_draw_point (widget->window, gc, gridx, gridy);
      }
    }
  } else {
    for (gridx = 0; gridx < width; gridx += editor_grid_horz_spacing) {
      gdk_draw_line (widget->window, gc, gridx, 0, gridx, height);
    }
    for (gridy = 0; gridy < height; gridy += editor_grid_vert_spacing) {
      gdk_draw_line (widget->window, gc, 0, gridy, width, gridy);
    }
  }
}

void
paint_selection(GdkWindow *window, GdkGC *gc, gint x, gint y,
		gint width, gint height)
{
  if (width > GB_CORNER_WIDTH && height > GB_CORNER_HEIGHT) {
    gdk_draw_rectangle (window, gc, TRUE,
			x, y,
			GB_CORNER_WIDTH, GB_CORNER_HEIGHT);
    gdk_draw_rectangle (window, gc, TRUE,
			x, y + height - GB_CORNER_HEIGHT, 
			GB_CORNER_WIDTH, GB_CORNER_HEIGHT);
    gdk_draw_rectangle (window, gc, TRUE,
			x + width - GB_CORNER_WIDTH, y,
			GB_CORNER_WIDTH, GB_CORNER_HEIGHT);
    gdk_draw_rectangle (window, gc, TRUE,
			x + width - GB_CORNER_WIDTH,
			y + height - GB_CORNER_HEIGHT,
			GB_CORNER_WIDTH, GB_CORNER_HEIGHT);
  }
  gdk_draw_rectangle (window, gc, FALSE,
		      x, y, width - 1, height - 1);
}


void
paint_widget(GtkWidget *widget)
{
  gint x, y, w, h;
  GtkWidget *ancestor;
  GdkGC *gc = widget->style->fg_gc[GTK_STATE_NORMAL];
  GdkGC *light_gc = widget->style->light_gc[GTK_STATE_NORMAL];
  GdkGC *dark_gc = widget->style->dark_gc[GTK_STATE_NORMAL];
  GdkWindow *window;

  gdk_gc_set_subwindow(gc, GDK_INCLUDE_INFERIORS);

  /* If widget is a placeholder, draw the placeholder pixmap in it */
  if (GB_IS_PLACEHOLDER(widget)) {
    gdk_window_get_size(widget->window, &w, &h);
    /* Do we really have to set this every time its drawn?? */
    gdk_window_set_back_pixmap(widget->window, placeholder_pixmap, FALSE);
    gdk_window_clear(widget->window);
    gdk_draw_line(widget->window, light_gc, 0, 0, w - 1, 0);
    gdk_draw_line(widget->window, light_gc, 0, 0, 0, h - 1);
    gdk_draw_line(widget->window, dark_gc, 0, h - 1, w - 1, h - 1);
    gdk_draw_line(widget->window, dark_gc, w - 1, 0, w - 1, h - 1);
  }

  /* Draw grid for fixed containers */
  if (GTK_IS_FIXED(widget)) draw_grid(widget);

  /* highlight selected widgets */
  ancestor = widget;
  while (ancestor) {
    if (g_list_find(selected_widgets, ancestor)) {
      if (ancestor->parent) {
	window = ancestor->parent->window;
	x = ancestor->allocation.x;
	y = ancestor->allocation.y;
	w = ancestor->allocation.width;
	h = ancestor->allocation.height;
      } else {
	window = ancestor->window;
	x = 0;
	y = 0;
	gdk_window_get_size(window, &w, &h);
      }
      paint_selection(window, gc, x, y, w, h);
    }
    /* Don't try to paint ancestors if it is a viewport */
    if (ancestor != widget && GTK_IS_VIEWPORT(ancestor)) break;

    ancestor = ancestor->parent;
  }
  /* Reset gc - maybe we should remember the current setting */
  gdk_gc_set_subwindow(gc, GDK_CLIP_BY_CHILDREN);
}


/*
 * Redraw the given widget completely, including all space allocated by its
 * parent (since this may be used for drawing the widget's selection).
 * If widget has no parent (i.e. its a toplevel window) just clear
 * it all and redraw.
 */
void
editor_refresh_widget(GtkWidget *widget)
{
  GdkGC *gc = widget->style->bg_gc[GTK_STATE_NORMAL];
  GdkWindow *window;
  gint x, y, w, h;

  if (widget->parent) {
    window = widget->parent->window;
    x = widget->allocation.x;
    y = widget->allocation.y;
    w = widget->allocation.width;
    h = widget->allocation.height;
  } else {
    window = widget->window;
    x = 0;
    y = 0;
    gdk_window_get_size(window, &w, &h);
  }

  /* If widget hasn't been layed out don't try to draw */
  if (x == -1 && y == -1) return;

  /* Clear placeholder's window, so the pixmap is redrawn */
  if (GB_IS_PLACEHOLDER(widget)) {
    gdk_window_clear(widget->window);

  } else {
  /* Use INCLUDE_INFERIORS drawing mode to clear window & all children */
    gdk_gc_set_subwindow(gc, GDK_INCLUDE_INFERIORS);
    gdk_draw_rectangle (window, gc, TRUE, x, y, w, h);
    gdk_gc_set_subwindow(gc, GDK_CLIP_BY_CHILDREN);
  }

  gtk_widget_draw(widget, NULL);
  gtk_widget_draw_children(widget);
}


/*
 * Key events
 */
gint
editor_on_key_press_event(GtkWidget *widget, GdkEventKey *event,
			  gpointer data)
{
  guint key = event->keyval;

  switch (key) {
  case GDK_Delete:
    if (selected_widgets) delete(selected_widgets->data);
    break;
  case GDK_Escape:
    editor_clear_selection(NULL);
    break;
  }
  return FALSE;
}

gint
editor_on_key_release_event(GtkWidget *widget, GdkEventKey *event,
			    gpointer data)
{

  return FALSE;
}


void
editor_on_select_activate(GtkWidget *menuitem, GtkWidget *widget)
{
  /* Note: could start a weird drag? */

  /* If sidget has no window, select parent enent box */
  if (GTK_WIDGET_NO_WINDOW(widget)) widget = widget->parent;
  editor_select_widget(widget, 0, 0, 0);
}


void
editor_on_delete()
{
  if (selected_widgets) delete(selected_widgets->data);
}


void
editor_on_delete_activate(GtkWidget *menuitem, GtkWidget *widget)
{
  delete(widget);
}


void
delete(GtkWidget *widget)
{
  if (GB_IS_PLACEHOLDER(widget)) {
    delete_placeholder(widget);
  } else {
    delete_widget(widget);
  }
}


void
delete_placeholder(GtkWidget *placeholder)
{
  GtkWidget *parent = placeholder->parent;

  /* Remove widget from the selection */
  editor_clear_selection(NULL);

  /* Can't delete children of a paned or a viewport */
  if (GTK_IS_PANED(parent) || GTK_IS_VIEWPORT(parent)) return;

  /* For a Clist, we can delete everything except column title widgets */
  if (GTK_IS_CLIST(parent)) {

    return;
  }

  /* Widgets with these parents can all be deleted OK */
  if (GTK_IS_TOOLBAR(parent) || GTK_IS_TREE(parent) || GTK_IS_LIST(parent)) {
    gtk_widget_destroy(placeholder);
    return;
  }

  /* For these widgets replace the parent with a placeholder, or delete the
     component if parent is a toplevel widget */
  if (GTK_IS_BIN(parent) || GTK_IS_BUTTON(parent)) {
    delete_widget(parent);
    return;
  }

  /* For a box, if the placeholder is the only child then replace the box with
     a placeholder, else just delete the placeholder */
  if (GTK_IS_BOX(parent)) {
    if (g_list_length(GTK_BOX(parent)->children) == 1) delete_widget(parent);
    else {
      gtk_container_remove(GTK_CONTAINER(parent), placeholder);
      /* Shouldn't really need to do this */
      gtk_widget_queue_resize(parent);
    }
    return;
  }

  /* For a notebook, if placeholder is the only page, replace the notebook with
     a placeholder, else delete the current notebook page (i.e. placeholder) */
  if (GTK_IS_NOTEBOOK(parent)) {
    if (g_list_length(GTK_NOTEBOOK(parent)->children) == 1)
      delete_widget(parent);
    else
      gtk_notebook_remove_page(GTK_NOTEBOOK(parent),
			      gtk_notebook_current_page(GTK_NOTEBOOK(parent)));
    return;
  }

  /* For a table, can't delete placeholder, unless there is only 1 row or
     column. In this case delete the placeholder, and move all the other
     children up/left. If the table is 1 x 1 then delete the table. */
  if (GTK_IS_TABLE(parent)) {
    gint nrows, ncols, position = 0, distance_to_move = 0;
    GList *item;
    GtkTableChild *table_child;

    nrows = GTK_TABLE(parent)->nrows;
    ncols = GTK_TABLE(parent)->ncols;
    if (nrows > 1 && ncols > 1) return;
    if (nrows == 1 && ncols == 1) {
      delete_widget(parent);
      return;
    }

    /* Find out where placeholder is */
    item = GTK_TABLE(parent)->children;
    while (item) {
      table_child = (GtkTableChild*) item->data;

      if (table_child->widget == placeholder) {
	/* Calculate how far up/left we will have to move the rest of the
	   children */
	if (nrows == 1) {
	  position = table_child->left_attach;
	  distance_to_move = table_child->right_attach
	    - table_child->left_attach;
	} else {
	  position = table_child->top_attach;
	  distance_to_move = table_child->bottom_attach
	    - table_child->top_attach;
	}
	break;
      }
      item = item->next;
    }
    /* Shouldn't reach the end of the list */
    g_return_if_fail(item != NULL);
    gtk_widget_destroy(placeholder);

    /* Now step through the table again, moving children up or left */
    item = GTK_TABLE(parent)->children;
    while (item) {
      table_child = (GtkTableChild*) item->data;
      if (nrows == 1) {
	if (table_child->left_attach > position) {
	  table_child->left_attach -= distance_to_move;
	  table_child->right_attach -= distance_to_move;
	}
      } else {
	if (table_child->top_attach > position) {
	  table_child->top_attach -= distance_to_move;
	  table_child->bottom_attach -= distance_to_move;
	}
      }
      item = item->next;
    }

    /* Now update the tables nrows & ncols */
    if (nrows == 1) {
      GTK_TABLE(parent)->ncols -= distance_to_move;
    } else {
      GTK_TABLE(parent)->nrows -= distance_to_move;
    }

    return;
  }
}


void
delete_widget(GtkWidget *widget)
{
  GtkWidget *placeholder;
  /*gint w, h;*/

  /* Set properties widget to NULL, in case the widget or parent is deleted */
  property_set_widget(NULL);

  /* Remove widget from the selection */
  editor_clear_selection(NULL);

  /* If widget is a toplevel widget (i.e. a project component) delete the
     component. */
  if (!widget->parent) {
    project_delete_component(widget);
    return;
  }

  /* If widget has no window, delete it's parent event box */
  if (GTK_WIDGET_NO_WINDOW(widget)) widget = widget->parent;

  /* If the widget's parent is a fixed container remove widget completely. */
  if (GTK_IS_FIXED(widget->parent)) {
    gtk_widget_destroy(widget);
    return;
  }

  /* Replace normal widget's with a placeholder with the same size? & select
     it so it can also be deleted easily using the Delete key. */
  placeholder = editor_new_placeholder();
  /*
  w = widget->requisition.width;
  h = widget->requisition.height;
  gtk_widget_set_usize(placeholder, w, h);
  */
  gb_widget_replace_child(widget->parent, widget, placeholder);
  editor_select_widget(placeholder, 0, 0, 0);
}
