/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>

#include "project.h"
#include "property.h"
#include "gbwidget.h"
#include "gbwidgetarray.h"
#include "save.h"
#include "editor.h"


/*
 * The hash table associates a Gtk widget class name with a table of functions
 *  used by the builder, e.g. creating a new widget, saving the widget etc.
 */
GHashTable *gb_widget_table   = NULL;
GHashTable *gb_widget_id_hash = NULL;

#define GB_CELL_EMPTY		1
#define GB_CELL_WIDGET		2
#define GB_CELL_PLACEHOLDER	3

/* Tooltips for the created widgets */
static GtkTooltips     *gb_widget_tooltips;


static GbWidgetData* gb_widget_new_widget_data();
static void on_widget_destroy		(GtkWidget	*widget,
					 gpointer	 data);

static void show_standard_properties	(GtkWidget	*widget,
					GbWidgetShowPropertiesData *data);
static void show_color_properties	(GdkColor	 colors[],
					 gchar		*name);

static void apply_standard_properties	(GtkWidget	*widget,
					 GbWidgetApplyPropertiesData *data);

static void save_standard_widget_data	(GtkWidget	*widget,
					 GbWidgetSaveData *data,
					 gchar		*name,
					 gchar		*class_name);
static void save_widget_style		(GtkWidget	*widget,
					 gpointer	 data);
static void save_widget_accelerators	(GtkWidget	*widget,
					 gpointer	 data);
static void save_widget_signals		(GtkWidget	*widget,
					 gpointer	 data);


void gb_widgets_init() {
  gint index;
  GbWidget *(*init_func)();
  GbWidget *gbwidget;

  gb_widget_table   = g_hash_table_new(g_str_hash, g_str_equal);
  gb_widget_id_hash = g_hash_table_new(g_str_hash, g_str_equal);

  /* Create tooltips */
  gb_widget_tooltips = gtk_tooltips_new();


  /*
   * Add all known gbwidgets to hash.
   */

  /* Skip the first row, the selector, which isn't really a gbwidget */
  index = 2;
  for (;;) {
    if (widgets[index] == NULL) {
      if (widgets[index + 1] == NULL) break;
    } else {
      init_func = widgets[index + 1];
      gbwidget = (*init_func)();
      g_hash_table_insert(gb_widget_table, widgets[index], gbwidget);
    }
    index += 2;
  }
}


void
gb_widget_init_struct(GbWidget *gbwidget)
{
  gbwidget->properties_page_number = GB_PROPERTIES_NOT_CREATED;
  gbwidget->pixmap_struct = NULL;
  gbwidget->tooltip = NULL;

  gbwidget->gb_widget_new		= NULL;
  gbwidget->gb_widget_create_properties	= NULL;
  gbwidget->gb_widget_show_properties	= NULL;
  gbwidget->gb_widget_apply_properties	= NULL;
  gbwidget->gb_widget_create_popup_menu	= NULL;
  gbwidget->gb_widget_load		= NULL;
  gbwidget->gb_widget_save		= NULL;
  gbwidget->gb_widget_write_source	= NULL;
}


GtkWidget*
gb_widget_new(gchar *class_name, GbWidgetNewData *data)
{
  GtkWidget *new_widget;
  GbWidget *gbwidget;

  data->name = NULL;
  data->widget_data = gb_widget_new_widget_data();
  if (data->widget) gtk_widget_ref(data->widget);

  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_val_if_fail(gbwidget != NULL, NULL);

  data->type = gtk_type_from_name(class_name);
  g_return_val_if_fail(data->type != 0, NULL);
  /* Slight kludge to name popup windows 'popup1', 'popup2' ... */
  if (!strcmp(class_name, "GtkWindow") && data->is_a_popup_window)
    data->name = gb_widget_new_name("   Popup");
  else
    data->name = gb_widget_new_name(class_name);

  if (gbwidget->gb_widget_new) new_widget = (gbwidget->gb_widget_new)(data);
  else new_widget = gtk_type_new(data->type);

  if (new_widget) new_widget = gb_widget_initialize(new_widget, data);

  return new_widget;
}


gchar*
gb_widget_new_name(gchar *class_name)
{
  char new_widget_name[128];
  gint widget_id, i;

  widget_id = (gint) g_hash_table_lookup(gb_widget_id_hash, class_name);
  if (widget_id) widget_id++;
  else widget_id = 1;
  g_hash_table_insert(gb_widget_id_hash, class_name, (gpointer) widget_id);

  /* Use 'class_name + 3' to skip 'Gtk' at the start of all class names */
  sprintf(new_widget_name, "%s%i", class_name + 3, widget_id);
  /* convert name to lower case (only normal ASCII chars) */
  for (i = 0; i < strlen(new_widget_name); i++) {
    if ((new_widget_name[i] >= 'A') && (new_widget_name[i] <= 'Z'))
      new_widget_name[i] += 'a' - 'A';
  }

  return g_strdup(new_widget_name);
}

GbWidgetData*
gb_widget_new_widget_data()
{
  GbWidgetData *widget_data = g_new(GbWidgetData, 1);
  widget_data->flags = GB_VISIBLE | GB_STYLE_INHERIT;
  widget_data->x = -1;
  widget_data->y = -1;
  widget_data->width = -1;
  widget_data->height = -1;
  widget_data->events = 0;
  widget_data->tooltip = NULL;
  widget_data->signals = NULL;
  widget_data->accelerators = NULL;
  return widget_data;
}


GtkWidget*
gb_widget_initialize(GtkWidget *widget, GbWidgetNewData	*data)
{
  GtkWidget *no_window_widget;

  gtk_widget_set_name(widget, data->name);

  gtk_object_set_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY,
		      data->widget_data);
  data->widget_data = NULL;
  gtk_signal_connect (GTK_OBJECT (widget), "destroy",
		      GTK_SIGNAL_FUNC(on_widget_destroy), &widget);

  if (GTK_WIDGET_NO_WINDOW(widget)) {
    gtk_widget_show(widget);
    no_window_widget = widget;
    widget = gtk_event_box_new();
    gtk_object_set_data(GTK_OBJECT(widget), GB_HOLDING_WIDGET_KEY, "True");
    gtk_container_add(GTK_CONTAINER (widget), no_window_widget);
  }

  editor_add_mouse_signals(widget);
  editor_add_draw_signals(widget);
  gtk_widget_show(widget);
  return widget;
}

void
gb_widget_free_new_data(GbWidgetNewData *data)
{
  if (data->name) g_free(data->name);
  if (data->widget_data) g_free(data->widget_data);
  if (data->widget) gtk_widget_unref(data->widget);
  g_free(data);
}


void
on_widget_destroy(GtkWidget *widget, gpointer data)
{
  GbWidgetData *widget_data;
  GList *item;
  GbSignal *signal;
  GbAccelerator *accelerator;

  widget_data = gtk_object_get_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY);
  if (widget_data->tooltip) g_free(widget_data->tooltip);
  if (widget_data->signals) {
    item = widget_data->signals;
    while (item) {
      signal = (GbSignal*) item->data;
      g_free(signal->signal_name);
      g_free(signal->handler);
      g_free(signal->data);
      item = item->next;
    }
    g_list_free(widget_data->signals);
  }
  if (widget_data->accelerators) {
    item = widget_data->accelerators;
    while (item) {
      accelerator = (GbAccelerator*) item->data;
      g_free(accelerator->signal);
      item = item->next;
    }
    g_list_free(widget_data->accelerators);
  }
}


/*
 * Loading widgets form project file.
 */
void
gb_widget_load(GtkWidget *widget,
	       GbWidgetLoadData *data)
{
  gchar *class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));
  GbWidget *gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_if_fail(gbwidget != NULL);
  if (gbwidget->gb_widget_load) (gbwidget->gb_widget_load)(widget, data);
}


/*
 * Saving widgets to project file.
 */
void
gb_widget_save(GtkWidget *widget,
	       gpointer user_data)
{
  GtkWidget *holding_widget = widget;
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  GbWidgetSaveData *data = user_data;
  gchar *name, *class_name;

  /* Don't save placeholder if parent is a table - it's only useful for boxes*/
  if (GB_IS_PLACEHOLDER(widget)) {
    if (!GTK_IS_TABLE(widget->parent)) save_tag(data, "Placeholder");
    return;
  }

  /* Sort out the real widget and the holding widget */
  if (GB_IS_HOLDING_WIDGET(widget)) widget = GTK_BIN(widget)->child;
  else if (GTK_WIDGET_NO_WINDOW(widget)) holding_widget = widget->parent;

  name = gtk_widget_get_name(widget);
  class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));

  /* Write out widget data and any child widgets */
  widget_data = gtk_object_get_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY);
  /* NULL is OK since the widget may be a label of a button etc. which
     hasn't been converted into a gbwidget */
  /*g_return_if_fail(widget_data != NULL);*/
  data->widget_data = widget_data;
  data->holding_widget = holding_widget;
  
  save_string(data, "Widget", class_name);
  data->indent++;
  save_standard_widget_data(widget, data, name, class_name);

  /* Call gbwidgets save function for any extra properties */
  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_if_fail(gbwidget != NULL);
  if (gbwidget->gb_widget_save) (gbwidget->gb_widget_save)(widget, data);

  /* If widget is a container, recursively save children.
     But don't save children of file/color selection dialogs */
  if (GTK_IS_CONTAINER(widget) && !GTK_IS_FILE_SELECTION(widget)
      && !GTK_IS_COLOR_SELECTION_DIALOG(widget)) {
    /*save_tag(data, "CHILDREN");*/
    /*data->indent++;*/
    gtk_container_foreach(GTK_CONTAINER(widget), gb_widget_save, data);
    /*data->indent--;*/
    /*save_tag(data, "/CHILDREN");*/
  }
  data->indent--;
  save_tag(data, "/Widget");
}

void
save_position_properties(GtkWidget *widget,
			 GbWidgetSaveData *data)
{
  GbWidgetData *wdata = data->widget_data;
  GtkWidget *holding_widget = data->holding_widget;

  if (GTK_IS_WINDOW(widget)) {
    save_int(data, GbX, wdata->x);
    save_int(data, GbY, wdata->y);
    save_int(data, GbWidth, wdata->width);
    save_int(data, GbHeight, wdata->height);
  } else if (GTK_IS_FIXED(holding_widget->parent)) {
    save_int(data, GbX, holding_widget->allocation.x);
    save_int(data, GbY, holding_widget->allocation.y);
    save_int(data, GbWidth, holding_widget->allocation.width);
    save_int(data, GbHeight, holding_widget->allocation.height);
  } else {
    if (wdata) {
      save_int(data, GbWidth, wdata->width);
      save_int(data, GbHeight, wdata->height);
    }
  }
}


void
save_standard_widget_data(GtkWidget *widget,
			  GbWidgetSaveData *data,
			  gchar *name,
			  gchar *class_name)
{
  gboolean can_default, has_default, can_focus, has_focus;
  GbWidgetData *wdata = data->widget_data;

  /* Save widget's name if it has been set */
  if (strcmp(name, class_name)) save_string(data, GbName, name);

  save_position_properties(widget, data);

  can_default = GTK_WIDGET_CAN_DEFAULT(widget);
  if (can_default) save_bool(data, GbCanDefault, can_default);
  has_default = wdata ? wdata->flags & GB_GRAB_DEFAULT : 0;
  if (has_default) save_bool(data, GbHasDefault, has_default);

  can_focus = GTK_WIDGET_CAN_FOCUS(widget);
  if (can_focus) save_bool(data, GbCanFocus, can_focus);
  has_focus = wdata ? wdata->flags & GB_GRAB_FOCUS : 0;
  if (has_focus) save_bool(data, GbHasFocus, has_focus);

  if (!GTK_WIDGET_NO_WINDOW(widget)) {
    GdkExtensionMode ext_mode = gtk_widget_get_extension_events(widget);
    gint i;

    if (wdata && wdata->events) save_int(data, GbEvents, wdata->events);
    /* Don't save default extension mode ('None') */
    if (ext_mode != GDK_EXTENSION_EVENTS_NONE) {
      for (i = 0; GbExtensionModeChoices[i]; i++) {
	if (GbExtensionModeValues[i] == ext_mode)
	  save_choice(data, GbExtEvents, GbExtensionModeChoices[i]);
      }
    }
  }

  /* If widget is a top-level window or widget's parent is a GtkFixed,
   *  write out widget's x & y coords (Note: parent may be an event box) */

  save_widget_style(widget, data);
  save_widget_accelerators(widget, data);
  save_widget_signals(widget, data);
}


void
save_widget_style(GtkWidget *widget,
		  gpointer data)
{
}

void
save_widget_accelerators(GtkWidget *widget,
			 gpointer data)
{
}

void
save_widget_signals(GtkWidget *widget,
		    gpointer data)
{
}

/*
 * Writing source code for widgets.
 */
void
gb_widget_write_source(GtkWidget *widget,
		       GbWidgetWriteSourceData *data)
{
  GbWidget *gbwidget;
  gchar *class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));
  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_if_fail(gbwidget != NULL);
  if (gbwidget->gb_widget_write_source)
    (gbwidget->gb_widget_write_source)(widget, data);
}



/*
 * Adding UI components for editing properties of a widget.
 */
gint
gb_widget_create_properties(GtkWidget *widget)
{
  GtkWidget *page;
  gint page_number;
  GbWidgetCreatePropertiesData data;
  GbWidget *gbwidget;
  gchar *class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));
  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);

  g_return_val_if_fail(gbwidget != NULL, GB_PROPERTIES_NOT_NEEDED);
  if (gbwidget->gb_widget_create_properties || GTK_IS_CONTAINER(widget)) {
    /* Create skeleton of properties page, so gbwidget just has to add
       properties */
    page = gtk_table_new (10, 3, FALSE);
    gtk_widget_show(page);
    page_number = property_add_gbwidget_page(page);
    property_set_table_position(page, 0);

    /* If widget is a container add a border width property */
    if (GTK_IS_CONTAINER(widget)) {
      gchar buf[128];
      sprintf(buf, "%s::border_width", class_name);
      property_add_int_range(buf, "Border Width:",
			     "The width of the border around the container",
			     0, 1000, 1, 10, 1);
    }

    if (gbwidget->gb_widget_create_properties)
      (gbwidget->gb_widget_create_properties)(widget, &data);
    return page_number;
  } else {
    return GB_PROPERTIES_NOT_NEEDED;
  }
}


/*
 * Showing properties for widget.
 * Note: This function can be passed the widget or the holding widget
 */

void
gb_widget_show_properties(GtkWidget *widget)
{
  GbWidgetShowPropertiesData data;
  gchar *class_name;
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  GtkWidget *holding_widget = widget;
  gint page;

  /* If widget is a placeholder just return */
  if (GB_IS_PLACEHOLDER(widget)) {
    property_set_widget(NULL);
    return;
  }

  /* Sort out the real widget and the holding widget */
  if (GB_IS_HOLDING_WIDGET(widget)) widget = GTK_BIN(widget)->child;
  else if (GTK_WIDGET_NO_WINDOW(widget)) holding_widget = widget->parent;

  class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));
  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_if_fail(gbwidget != NULL);

  /* Turn off auto-apply so we can set properties without the 'changed'
     callbacks calling gb_widget_apply_properties(). */
  property_set_auto_apply(FALSE);

  /* Need this here to make sure properties notebook is sensitive */
  property_set_widget(widget);
  property_refresh();

  page = gbwidget->properties_page_number;
  /* If widget's properties page hasn't been created, create it now */
  if (page == GB_PROPERTIES_NOT_CREATED) {
    page = gb_widget_create_properties(widget);
    gbwidget->properties_page_number = page;
  }

  /* Show the widget's own properties page if it has one.
     Need to show the page before setting properties because of the
     Text widget - it must be realized before setting the text :-( */
  if (page == GB_PROPERTIES_NOT_NEEDED)  property_hide_gbwidget_page();
  else property_show_gbwidget_page(page);

  widget_data = gtk_object_get_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY);
  g_return_if_fail(widget_data != NULL);
  data.widget_data = widget_data;
  data.holding_widget = holding_widget;
  show_standard_properties(widget, &data);

  /* If widget is a container, show the border width */
  if (GTK_IS_CONTAINER(widget)) {
    gchar buf[128];
    sprintf(buf, "%s::border_width", class_name);
    property_set_int(buf, GTK_CONTAINER(widget)->border_width);
  }

  if (gbwidget->gb_widget_show_properties)
    (gbwidget->gb_widget_show_properties)(widget, &data);

  /* Turn auto-apply back on again */
  property_set_auto_apply(TRUE);

  /* GTK bug workaround - make sure property window is displayed correctly */
  /* Would prefer this here, so that any gbwidget properties that have been
     made sensitive are redrawn properly, but Text widget causes a SEGV!! */
  /*property_refresh();*/
}

void
gb_widget_show_position_properties(GtkWidget *widget,
				   GtkWidget *holding_widget,
				   GbWidgetData *wdata)
{
  gboolean sensitive;

  /*g_print("In show_position_properties\n");*/

  if (GTK_IS_WINDOW(widget)) {
    property_set_int(GbX, wdata->x);
    property_set_int(GbY, wdata->y);
    property_set_int(GbWidth, wdata->width);
    property_set_int(GbHeight, wdata->height);
    sensitive = (GTK_WINDOW(widget)->position == GTK_WIN_POS_NONE);
  } else if (GTK_IS_FIXED(holding_widget->parent)) {
    property_set_int(GbX, holding_widget->allocation.x);
    property_set_int(GbY, holding_widget->allocation.y);
    property_set_int(GbWidth, holding_widget->allocation.width);
    property_set_int(GbHeight, holding_widget->allocation.height);
    /* Set these to 1 to mark that the width & height have been shown,
       otherwise this function is repeatedly called in each expose event
       - see expose_widget in editor.c */
    if (holding_widget->allocation.width > 1
	&& holding_widget->allocation.height > 1) {
      wdata->width = 1;
      wdata->height = 1;
    }
    sensitive = TRUE;
  } else {
    property_set_int(GbX, holding_widget->allocation.x);
    property_set_int(GbY, holding_widget->allocation.y);
    if (wdata->width <= 0 && wdata->height <= 0) {
      wdata->width = widget->requisition.width;
      wdata->height = widget->requisition.height;
    }
    property_set_int(GbWidth, widget->requisition.width);
    property_set_int(GbHeight, widget->requisition.height);
    sensitive = FALSE;
  }

  /* Only toplevel widgets and children of fixed containers can use x & y */
  property_set_sensitive(GbX, sensitive);
  property_set_sensitive(GbY, sensitive);

  /*g_print("Out show_position_properties\n");*/
}

GtkTableChild*
find_table_child(GtkWidget *table, GtkWidget *widget)
{
  GList *children;
  GtkTableChild *child;

  children = GTK_TABLE(table)->children;
  while (children) {
    child = children->data;
    if (child->widget == widget) {
      return child;
    }
    children = children->next;
  }
  return NULL;
}

void
show_special_child_properties(GtkWidget *widget, GtkWidget *holding_widget)
{
  GtkWidget *parent = holding_widget->parent;
  gint expand, fill, padding;
  GtkPackType pack_type;

  if (!parent) {
    property_hide_special_child_page();
    return;
  }
  if (GTK_IS_TABLE(parent)) {
    GtkTableChild *table_child = find_table_child(parent, holding_widget);
    g_return_if_fail(table_child != NULL);

    property_set_int(GbXPad,		table_child->xpadding);
    property_set_int(GbYPad,		table_child->ypadding);
    property_set_bool(GbXExpand,	table_child->xexpand);
    property_set_bool(GbYExpand,	table_child->yexpand);
    property_set_bool(GbXShrink,	table_child->xshrink);
    property_set_bool(GbYShrink,	table_child->yshrink);
    property_set_bool(GbXFill,	table_child->xfill);
    property_set_bool(GbYFill,	table_child->yfill);
    property_set_int(GbCellX,	table_child->left_attach);
    property_set_int(GbCellY,	table_child->top_attach);
    property_set_int(GbColSpan,
		     table_child->right_attach - table_child->left_attach);
    property_set_int(GbRowSpan,
		     table_child->bottom_attach - table_child->top_attach);

    property_show_special_child_page(0);
  } else if (GTK_IS_BOX(parent)) {
    gtk_box_query_child_packing(GTK_BOX(parent), holding_widget,
				&expand, &fill, &padding, &pack_type);
    property_set_int(GbPadding, padding);
    property_set_bool(GbExpand, expand);
    property_set_bool(GbFill, fill);
    property_set_bool(GbPack, pack_type == GTK_PACK_START);
    property_show_special_child_page(1);
  } else {
    property_hide_special_child_page();
  }
}

void
show_standard_properties(GtkWidget *widget, GbWidgetShowPropertiesData *data)
{
  GbWidgetData *wdata = data->widget_data;
  GtkWidget *holding_widget = data->holding_widget;
  gchar buffer[128];
  int i;

  /* Properties on widget page */
  property_set_string(GbName, gtk_widget_get_name(widget));

  /* Special child properties page */
  show_special_child_properties(widget, holding_widget);

  /* Properties on standard page */
  gb_widget_show_position_properties(widget, holding_widget, wdata);

  property_set_bool(GbVisible, wdata->flags & GB_VISIBLE);
  property_set_bool(GbSensitive, GTK_WIDGET_IS_SENSITIVE(widget));

  property_set_string(GbTooltip, wdata->tooltip);

  property_set_bool(GbCanDefault, GTK_WIDGET_CAN_DEFAULT(widget));
  property_set_bool(GbHasDefault, wdata->flags & GB_GRAB_DEFAULT);
  property_set_bool(GbCanFocus, GTK_WIDGET_CAN_FOCUS(widget));
  property_set_bool(GbHasFocus, wdata->flags & GB_GRAB_FOCUS);

  if (!GTK_WIDGET_NO_WINDOW(widget)) {
    GdkExtensionMode ext_mode = gtk_widget_get_extension_events(widget);
    property_set_string(GbEvents,
			property_events_value_to_string(wdata->events));
    property_set_sensitive(GbEvents, TRUE);

    for (i = 0; GbExtensionModeChoices[i]; i++) {
      if (GbExtensionModeValues[i] == ext_mode)
	property_set_choice(GbExtEvents, i);
    }
    property_set_sensitive(GbExtEvents, TRUE);
  } else {
    property_set_string(GbEvents, "N/A");
    property_set_sensitive(GbEvents, FALSE);
    property_set_choice(GbExtEvents, 0);
    property_set_sensitive(GbExtEvents, FALSE);
  }


  /* Styles page */
  property_set_bool(GbStyleInherit, wdata->flags & GB_STYLE_INHERIT);

  /* Colors */
  show_color_properties(widget->style->fg,    "fg");
  show_color_properties(widget->style->bg,    "bg");
  show_color_properties(widget->style->light, "light");
  show_color_properties(widget->style->dark,  "dark");
  show_color_properties(widget->style->mid,   "middle");
  show_color_properties(widget->style->text,  "text");
  show_color_properties(widget->style->base,  "base");

  /* Background pixmaps */
  for (i = 0; i < 5; i++) {
    sprintf(buffer, "GtkStyle::%s[%s]", GbBgPixmapName, GbStateNames[i]);
    property_set_pixmap(buffer, widget->style->bg_pixmap[i], NULL, NULL);
  }

  /* Accelerators page */

  /* Signals page */

}


void
show_color_properties(GdkColor colors[], gchar *name)
{
  GdkColor gdkcolor;
  guchar color[4];
  gint state;
  gchar buf[128];

  for (state = 0; state < 5; state++) {
    sprintf(buf, "GtkStyle::%s[%s]", name, GbStateNames[state]);
    gdkcolor = colors[state];
    color[0] = gdkcolor.red   >> 8;
    color[1] = gdkcolor.green >> 8;
    color[2] = gdkcolor.blue  >> 8;
    property_set_color(buf, color);
  }
}


/*
 * Applying properties to widget.
 * Note: This function can be passed the widget or the holding widget
 */
void
gb_widget_apply_properties(GtkWidget *widget, GtkWidget *property)
{
  GtkWidget *holding_widget = widget;
  GbWidgetApplyPropertiesData *data;
  GbWidget *gbwidget;
  GbWidgetData *widget_data;
  gchar *class_name;

  /* Sort out the real widget and the holding widget */
  if (GB_IS_HOLDING_WIDGET(widget)) widget = GTK_BIN(widget)->child;
  else if (GTK_WIDGET_NO_WINDOW(widget)) holding_widget = widget->parent;

  /* Set which property/ies to apply */
  property_set_apply(property);

  class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));
  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_if_fail(gbwidget != NULL);

  widget_data = gtk_object_get_data(GTK_OBJECT(widget), GB_WIDGET_DATA_KEY);
  g_return_if_fail(widget_data != NULL);
  data = g_new(GbWidgetApplyPropertiesData, 1);
  data->widget_data = widget_data;
  data->holding_widget = holding_widget;
  apply_standard_properties(widget, data);

  /* If widget is a container, show the border width */
  if (GTK_IS_CONTAINER(widget)) {
    gchar buf[128];
    gint border_width;
    sprintf(buf, "%s::border_width", class_name);
    border_width = property_get_int(buf);
    if (Apply) gtk_container_border_width(GTK_CONTAINER(widget), border_width);
  }

  if (gbwidget->gb_widget_apply_properties)
    (gbwidget->gb_widget_apply_properties)(widget, data);
  g_free(data);
}


gboolean
apply_color_properties(GtkWidget *widget, GdkColor *colors,
		       GdkColor *new_colors, gchar *name)
{
  gint state;
  gchar buf[128];
  guchar *color;
  GdkColor *old_gdkcolor, *new_gdkcolor;
  gboolean need_redraw = FALSE;

  for (state = 0; state < 5; state++) {
    sprintf(buf, "GtkStyle::%s[%s]", name, GbStateNames[state]);
    old_gdkcolor = &colors[state];
    new_gdkcolor = &new_colors[state];
    new_gdkcolor->red   = old_gdkcolor->red;
    new_gdkcolor->green = old_gdkcolor->green;
    new_gdkcolor->blue  = old_gdkcolor->blue;
    new_gdkcolor->pixel = old_gdkcolor->pixel;

    color = property_get_color(buf);
    if (Apply) {
      new_gdkcolor->red   = color[0] << 8;
      new_gdkcolor->green = color[1] << 8;
      new_gdkcolor->blue  = color[2] << 8;
      if (!gdk_color_equal(new_gdkcolor, old_gdkcolor)) need_redraw = TRUE;
    }
  }
  return need_redraw;
}

void
apply_colors(GtkWidget *widget)
{
  GtkStyle *style;
  GdkColor fg[5];
  GdkColor bg[5];
  GdkColor light[5];
  GdkColor dark[5];
  GdkColor mid[5];
  GdkColor text[5];
  GdkColor base[5];
  GdkPixmap *bg_pixmap[5];
  GdkBitmap *mask;
  gint redraw, i, state;
  gchar buf[128], *filename;

  redraw  = apply_color_properties(widget, widget->style->fg, fg,
				   "fg");
  redraw |= apply_color_properties(widget, widget->style->bg, bg,
				   "bg");
  redraw |= apply_color_properties(widget, widget->style->light, light,
				   "light");
  redraw |= apply_color_properties(widget, widget->style->dark, dark,
				   "dark");
  redraw |= apply_color_properties(widget, widget->style->mid, mid,
				   "middle");
  redraw |= apply_color_properties(widget, widget->style->text, text,
				   "text");
  redraw |= apply_color_properties(widget, widget->style->base, base,
				   "base");

  /* Background pixmaps */
  for (state = 0; state < 5; state++) {
    sprintf(buf, "GtkStyle::%s[%s]", GbBgPixmapName, GbStateNames[state]);
    bg_pixmap[state] = property_get_pixmap(buf, &mask, &filename);
    if (Apply) {
      if (bg_pixmap[state] != widget->style->bg_pixmap[state]) redraw = TRUE;
    }
  }

  if (redraw) {
    style = gtk_style_new();
    for (i = 0; i < 5; i++) {
      style->fg[i]        = fg[i];
      style->bg[i]        = bg[i];
      style->light[i]     = light[i];
      style->dark[i]      = dark[i];
      style->mid[i]       = mid[i];
      style->text[i]      = text[i];
      style->base[i]	  = base[i];
      style->bg_pixmap[i] = bg_pixmap[i];
    }
    gtk_widget_set_style(widget, style);
    editor_refresh_widget(widget);
  }
}

void
split_table_placeholder(GtkWidget *table, GtkWidget *placeholder,
			gint left, gint right, gint top, gint bottom,
			gint skip_row, gint skip_col)
{
  gint row, col;

  gtk_container_remove(GTK_CONTAINER(table), placeholder);
  for (row = top; row < bottom; row++) {
    for (col = left; col < right; col++) {
      if (!(row == skip_row && col == skip_col)) {
	gtk_table_attach_defaults(GTK_TABLE (table),
				  editor_new_placeholder(),
				  col, col + 1, row, row + 1);
      }
    }
  }
}

/* Finds out if cell is occupied by a real widget. If not, returns whether
   the cell is empty or whether a placeholder is in it */
gint
is_cell_occupied(GtkWidget *table, gint row, gint col)
{
  GList *children;
  GtkTableChild *child;
  gint return_val = GB_CELL_EMPTY;

  children = GTK_TABLE(table)->children;
  while (children) {
    child = children->data;
    if (child->top_attach <= row && child->bottom_attach > row
	&& child->left_attach <= col && child->right_attach > col) {
      if (GB_IS_PLACEHOLDER(child->widget))
	return_val = GB_CELL_PLACEHOLDER;
      else 
	return GB_CELL_WIDGET;
    }
    children = children->next;
  }
  return return_val;
}

void
remove_table_placeholders(GtkWidget *table, gint row, gint col)
{
  GList *children, *next;
  GtkTableChild *child;
  gint left, right, top, bottom;

  children = GTK_TABLE(table)->children;
  while (children) {
    next = children->next;
    child = children->data;
    left   = child->left_attach;
    right  = child->right_attach;
    top    = child->top_attach;
    bottom = child->bottom_attach;

    if (top <= row && bottom > row && left <= col && right > col) {
      /* If the widget is a placeholder remove it */
      if (GB_IS_PLACEHOLDER(child->widget)) {
	if (bottom - top > 1 || right - left > 1) {
	  split_table_placeholder(table, child->widget,
				  left, right, top, bottom,
				  row, col);
	} else {
	  gtk_container_remove(GTK_CONTAINER(table), child->widget);
	}
      }
    }
    children = next;
  }
}

void
update_table_placeholders(GtkWidget *table)
{
  gint row, col, cell_contents;

  for (row = 0; row < GTK_TABLE(table)->nrows; row++) {
    for (col = 0; col < GTK_TABLE(table)->ncols; col++) {
      /* Find out what is in the cell */
      cell_contents = is_cell_occupied(table, row, col);
      if (cell_contents == GB_CELL_WIDGET) {
	/* If cell is occupied, delete any placeholders there. If a placeholder
	   occupies the cell but spans multiple rows/cols split it into single
	   cells but without a placeholder in this cell */
	remove_table_placeholders(table, row, col);

      } else if (cell_contents == GB_CELL_EMPTY) {
	/* If the cell is empty, put a placeholder in it */
	gtk_table_attach_defaults(GTK_TABLE (table),
				  editor_new_placeholder(),
				  col, col + 1, row, row + 1);
      }
    }
  }
}


void
apply_special_child_properties(GtkWidget *widget, GtkWidget *holding_widget)
{
  GtkWidget *parent = holding_widget->parent;

  if (!parent) return;
  if (GTK_IS_TABLE(parent)) {
    GtkTableChild *tchild = find_table_child(parent, holding_widget);
    gint xpad, ypad, left_attach, right_attach, top_attach, bottom_attach;
    gint old_left_attach, old_right_attach, old_top_attach, old_bottom_attach;
    gint xexpand, yexpand, xshrink, yshrink, xfill, yfill;
    gboolean myApply;

    g_return_if_fail(tchild != NULL);
    old_left_attach   = tchild->left_attach;
    old_right_attach  = tchild->right_attach;
    old_top_attach    = tchild->top_attach;
    old_bottom_attach = tchild->bottom_attach;
    
    xpad = property_get_int(GbXPad);
    myApply = Apply;
    ypad = property_get_int(GbYPad);
    myApply |= Apply;
    xexpand = property_get_bool(GbXExpand) ? GTK_EXPAND : 0;
    myApply |= Apply;
    yexpand = property_get_bool(GbYExpand) ? GTK_EXPAND : 0;
    myApply |= Apply;
    xshrink = property_get_bool(GbXShrink) ? GTK_SHRINK : 0;
    myApply |= Apply;
    yshrink = property_get_bool(GbYShrink) ? GTK_SHRINK : 0;
    myApply |= Apply;
    xfill = property_get_bool(GbXFill) ? GTK_FILL : 0;
    myApply |= Apply;
    yfill = property_get_bool(GbYFill) ? GTK_FILL : 0;
    myApply |= Apply;
    left_attach = property_get_int(GbCellX);
    myApply |= Apply;
    top_attach = property_get_int(GbCellY);
    myApply |= Apply;
    right_attach = property_get_int(GbColSpan) + left_attach;
    myApply |= Apply;
    bottom_attach = property_get_int(GbRowSpan) + top_attach;
    myApply |= Apply;

    if (myApply) {
      if (xpad != tchild->xpadding || ypad != tchild->ypadding
	  || (xexpand && !tchild->xexpand) || (!xexpand && tchild->xexpand)
	  || (yexpand && !tchild->yexpand) || (!yexpand && tchild->yexpand)
	  || (xshrink && !tchild->xshrink) || (!xshrink && tchild->xshrink)
	  || (yshrink && !tchild->yshrink) || (!yshrink && tchild->yshrink)
	  || (xfill   && !tchild->xfill)   || (!xfill   && tchild->xfill)
	  || (yfill   && !tchild->yfill)   || (!yfill   && tchild->yfill)
	  || left_attach   != tchild->left_attach
	  || right_attach  != tchild->right_attach
	  || top_attach    != tchild->top_attach
	  || bottom_attach != tchild->bottom_attach) {

	gtk_container_block_resize(GTK_CONTAINER(parent));
	gtk_widget_ref(holding_widget);
	gtk_container_remove(GTK_CONTAINER(parent), holding_widget);
	gtk_table_attach(GTK_TABLE(parent), holding_widget,
			 left_attach, right_attach, top_attach, bottom_attach,
			 xexpand | xshrink | xfill, yexpand | yshrink | yfill,
			 xpad, ypad);
	gtk_widget_unref(holding_widget);
	update_table_placeholders(parent);
	gtk_container_unblock_resize(GTK_CONTAINER(parent));
      }
    }
  } else if (GTK_IS_BOX(parent)) {
    gint padding;
    gboolean expand, fill, pack, myApply;

    padding = property_get_int(GbPadding);
    myApply = Apply;
    expand = property_get_bool(GbExpand);
    myApply |= Apply;
    fill = property_get_bool(GbFill);
    myApply |= Apply;
    pack = property_get_bool(GbPack);
    myApply |= Apply;
    if (myApply) gtk_box_set_child_packing(GTK_BOX(parent), holding_widget,
					   expand, fill, padding,
					   pack ? GTK_PACK_START : GTK_PACK_END);
  }
}


void
apply_position_properties(GtkWidget *widget,
			  GtkWidget *holding_widget,
			  GbWidgetData *wdata)
{
  gint x, y, w, h;
  gboolean applyX, applyY, applyWidth, applyHeight;

  x = property_get_int(GbX);
  applyX = Apply;
  y = property_get_int(GbY);
  applyY = Apply;
  w = property_get_int(GbWidth);
  applyWidth = Apply;
  h = property_get_int(GbHeight);
  applyHeight = Apply;

  /*  g_print("In apply_position_properties x:%i y:%i w:%i h:%i\n", x, y, w, h);*/

  if (GTK_IS_WINDOW(widget)) {
    wdata->x = x;
    wdata->y = y;
    wdata->width = w;
    wdata->height = h;
  } else if (GTK_IS_FIXED(holding_widget->parent)) {
    if (applyWidth || applyHeight) {
      if (w != holding_widget->allocation.width
	  || h != holding_widget->allocation.height) {
	gtk_widget_set_usize(holding_widget, w, h);
      }
    }
    if (applyX || applyY) {
      if (x != holding_widget->allocation.x
	  || y != holding_widget->allocation.y) {
	/* Do we need both? */
	gtk_fixed_move(GTK_FIXED(holding_widget->parent), holding_widget,
		       x, y);
	gtk_widget_set_uposition(holding_widget, x, y);
      }
    }
  } else {
    wdata->width = w;
    wdata->height = h;
    gtk_widget_set_usize(holding_widget, w, h);
  }
  /*g_print("Out apply_position_properties\n");*/
}


void
apply_standard_properties(GtkWidget *widget, GbWidgetApplyPropertiesData *data)
{
  GbWidgetData *wdata = data->widget_data;
  GtkWidget *holding_widget = data->holding_widget;
  gchar *name, *tooltip;
  gboolean visible, sensitive, can_default, has_default, can_focus, has_focus;

  /* Properties on widget page */
  name = property_get_string(GbName);
  if (Apply) gtk_widget_set_name(widget, name);

  /* Special child properties page */
  apply_special_child_properties(widget, holding_widget);


  /* Properties on standard page */
  apply_position_properties(widget, holding_widget, wdata);

  visible = property_get_bool(GbVisible);
  if (Apply) {
    if (visible) wdata->flags |= GB_VISIBLE;
    else  wdata->flags &= ~GB_VISIBLE;
  }
  sensitive = property_get_bool(GbSensitive);
  if (Apply) gtk_widget_set_sensitive(widget, sensitive);
  tooltip = property_get_string(GbTooltip);
  if (Apply) {
    if (strlen(tooltip) > 0)
      gtk_tooltips_set_tip(gb_widget_tooltips, widget, tooltip, NULL);
    else
      gtk_tooltips_set_tip(gb_widget_tooltips, widget, NULL, NULL);
  }
  can_default = property_get_bool(GbCanDefault);
  if (Apply) {
    if (can_default) GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_DEFAULT);
    else GTK_WIDGET_UNSET_FLAGS (widget, GTK_CAN_DEFAULT);
    editor_refresh_widget(widget);
  }
  has_default = property_get_bool(GbHasDefault);
  if (Apply) {
    if (has_default) wdata->flags |= GB_GRAB_DEFAULT;
    else  wdata->flags &= ~GB_GRAB_DEFAULT;
  }
  can_focus = property_get_bool(GbCanFocus);
  if (Apply) {
    if (can_focus) GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);
    else GTK_WIDGET_UNSET_FLAGS (widget, GTK_CAN_FOCUS);
  }
  has_focus = property_get_bool(GbHasFocus);
  if (Apply) {
    if (has_focus) wdata->flags |= GB_GRAB_FOCUS;
    else  wdata->flags &= ~GB_GRAB_FOCUS;
  }

  /* Styles page */

  /* Colors */
  apply_colors(widget);

  /* Accelerators page */

  /* Signals page */

}


/*
 * Creating context-menu for widget.
 */
void
gb_widget_show_popup_menu	(GtkWidget		    *widget,
				 GdkEventButton		    *event)
{
  static GtkWidget *menu = NULL;

  GbWidget *gbwidget;
  gchar *class_name, *name;
  GtkWidget *menuitem, *ancestor, *submenu;
  GbWidgetCreatePopupMenuData *data;

  class_name = gtk_type_name(GTK_OBJECT_TYPE(widget));
  gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
  g_return_if_fail(gbwidget != NULL);

  name = gtk_widget_get_name(widget);
  if (GB_IS_HOLDING_WIDGET(widget)) {
    widget = GTK_BIN(widget)->child;
    name = gtk_widget_get_name(widget);
  } else if (GB_IS_PLACEHOLDER(widget)) {
    name = "Placeholder";
  }
  data = g_new(GbWidgetCreatePopupMenuData, 1);

  /* bug workaround - delete last popup menu */
  if (menu) gtk_widget_destroy(menu);

  menu = gtk_menu_new();
  /* doesn't call menuitem's 'activate' handler if we do this
    gtk_signal_connect_object (GTK_OBJECT (menu), "deactivate",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(menu));
			     */
  menuitem = gtk_menu_item_new_with_label (name);
  gtk_widget_show (menuitem);
  gtk_widget_set_sensitive(menuitem, FALSE);
  gtk_menu_append (GTK_MENU (menu), menuitem);

  menuitem = gtk_menu_item_new_with_label ("Select");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(editor_on_select_activate), widget);
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (menu), menuitem);

  menuitem = gtk_menu_item_new_with_label ("Delete");
  gtk_widget_show (menuitem);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(editor_on_delete_activate), widget);

  if (gbwidget->gb_widget_create_popup_menu)
    (gbwidget->gb_widget_create_popup_menu)(widget, data);

  ancestor = widget->parent;
  while (ancestor) {
    name = gtk_widget_get_name(ancestor);
    if (GB_IS_PLACEHOLDER(ancestor)) name = "Placeholder";

    /* Skip special event boxes */
    if (!GB_IS_HOLDING_WIDGET(ancestor)) {
      /* Add a separator */
      menuitem = gtk_menu_item_new ();
      gtk_menu_append (GTK_MENU (menu), menuitem);
      gtk_widget_show (menuitem);

      menuitem = gtk_menu_item_new_with_label (name);
      gtk_widget_show (menuitem);
      gtk_menu_append (GTK_MENU (menu), menuitem);

      /* Create submenu */
      submenu = gtk_menu_new();
      gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), submenu);
      /*gtk_widget_show (submenu);*/

      menuitem = gtk_menu_item_new_with_label ("Select");
      gtk_widget_show (menuitem);
      gtk_menu_append (GTK_MENU (submenu), menuitem);
      gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			  GTK_SIGNAL_FUNC(editor_on_select_activate),
			  ancestor);

      menuitem = gtk_menu_item_new_with_label ("Delete");
      gtk_widget_show (menuitem);
      gtk_menu_append (GTK_MENU (submenu), menuitem);
      gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			  GTK_SIGNAL_FUNC(editor_on_delete_activate),
			  ancestor);

      /* Call ancestors function to add any menu items */
      class_name = gtk_type_name(GTK_OBJECT_TYPE(ancestor));
      gbwidget = (GbWidget*) g_hash_table_lookup(gb_widget_table, class_name);
      if (gbwidget != NULL && gbwidget->gb_widget_create_popup_menu)
	(gbwidget->gb_widget_create_popup_menu)(ancestor, data);
    }
    ancestor = ancestor->parent;
  }
  g_free(data);

  gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL,
		 event->button, event->time);
}


static gint find_child_current_index;
static gint find_child_found_index;

void
find_child_index_callback(GtkWidget *widget, gpointer child)
{
  if (widget == child) find_child_found_index = find_child_current_index;
  find_child_current_index++;
}

gint
find_child_index(GtkWidget *widget, GtkWidget *child)
{
  find_child_current_index = 0;
  find_child_found_index = -1;
  gtk_container_foreach(GTK_CONTAINER(widget), find_child_index_callback,
			child);
  return find_child_found_index;
}

GtkNotebookPage*
find_notebook_child_data(GtkWidget *widget, GtkWidget *current_child,
			 gint *pos)
{
  GList *children;
  GtkNotebookPage *page;
  gint page_num = 0;

  children = GTK_NOTEBOOK(widget)->children;
  while (children) {
    page = children->data;
    if (page->child == current_child || page->tab_label == current_child) {
      *pos       = page_num;
      return page;
    }
    children = children->next;
    page_num++;
  }
  return NULL;
}

gint
find_table_child_data(GtkWidget *widget, GtkWidget *current_child,
		      gint *left, gint *right, gint *top, gint *bottom)
{
  GList *children;
  GtkTableChild *child;

  children = GTK_TABLE(widget)->children;
  while (children) {
    child = children->data;
    if (child->widget == current_child) {
      *left   = child->left_attach;
      *right  = child->right_attach;
      *top    = child->top_attach;
      *bottom = child->bottom_attach;
      return TRUE;
    }
    children = children->next;
  }
  return FALSE;
}

/*
 * Replace a child with a new child. Used to replace placeholders with
 * a widget when adding widgets, and also to replace widgets with
 * placeholders when deleting.
 * NOTE: gbwidgets do not currently have their own function for this,
 * as there are only a few special cases (I hope!)
 */
void gb_widget_replace_child    (GtkWidget *widget,
				 GtkWidget *current_child,
				 GtkWidget *new_child)
{
  /* Stop resizes until child is replaced */
  gtk_container_block_resize(GTK_CONTAINER(widget));

  if (GTK_IS_BIN(widget)) {
    /* For a bin, we just delete the existing child and add the new one
     */
    gtk_container_remove(GTK_CONTAINER(widget), current_child);
    gtk_container_add(GTK_CONTAINER(widget), new_child);

  } else if (GTK_IS_BOX(widget)) {
    /* For a box, we find out the position of the current child, delete it,
       add the new one, and move it into position with reorder_child()
       */
    gint pos = find_child_index(widget, current_child);
    g_return_if_fail(pos != -1);
    gtk_container_remove(GTK_CONTAINER(widget), current_child);
    gtk_container_add(GTK_CONTAINER(widget), new_child);
    gtk_box_reorder_child(GTK_BOX(widget), new_child, pos);

  } else if (GTK_IS_LIST(widget)) {
    /* For a list, we find out the position of the current child, delete it,
       and add the new one at the same position.
       */
    gint pos = gtk_list_child_position(GTK_LIST(widget), current_child);
    GList glist = { current_child, NULL, NULL };
    gtk_list_remove_items(GTK_LIST(widget), &glist);
    glist.data = new_child;
    gtk_list_insert_items(GTK_LIST(widget), &glist, pos);

  } else if (GTK_IS_NOTEBOOK(widget)) {
    /* For a notebook, we find out the position of the current child, delete
       it, and add the new one at the same position. If the current_child is
       a notebook tab, just replace it.
       */
    GtkWidget *tab_label, *notebook_page;
    gint pos;
    GtkNotebookPage *page = find_notebook_child_data(widget, current_child,
						     &pos);
    g_return_if_fail(page != NULL);
    if (page->child == current_child) {
      tab_label = page->tab_label;
      gtk_widget_ref (tab_label);
      gtk_notebook_remove_page(GTK_NOTEBOOK(widget), pos);
      gtk_notebook_insert_page(GTK_NOTEBOOK(widget), new_child, tab_label,
			       pos);
      gtk_notebook_set_page(GTK_NOTEBOOK(widget), pos);
      gtk_widget_unref (tab_label);
    } else {
      notebook_page = page->child;
      gtk_widget_ref (notebook_page);
      gtk_notebook_remove_page(GTK_NOTEBOOK(widget), pos);
      gtk_notebook_insert_page(GTK_NOTEBOOK(widget), notebook_page, new_child,
			       pos);
      gtk_notebook_set_page(GTK_NOTEBOOK(widget), pos);
      gtk_widget_unref (notebook_page);
    }

  } else if (GTK_IS_PANED(widget)) {
    /* For paned, we find out the position of the current child, delete it,
       and add the new one at the same position.
       */
    gint pos = (GTK_PANED(widget)->child1 == current_child) ? 1 : 2;
    gtk_container_remove(GTK_CONTAINER(widget), current_child);
    if (pos == 1) gtk_paned_add1(GTK_PANED(widget), new_child);
    else gtk_paned_add2(GTK_PANED(widget), new_child);

  } else if (GTK_IS_TABLE(widget)) {
    /* For a table, we find out the position & size of the current child,
       delete it, and add the new one in the same place.
       */
    gint left, right, top, bottom;
    gint found = find_table_child_data(widget, current_child,
				       &left, &right, &top, &bottom);
    g_return_if_fail(found);
    gtk_container_remove(GTK_CONTAINER(widget), current_child);
    gtk_table_attach_defaults(GTK_TABLE(widget), new_child, 
			      left, right, top, bottom);

  } else if (GTK_IS_TREE(widget)) {
    /* For a tree, we find out the position of the current child, delete it,
       and add the new one at the same position.
       */
    gint pos = gtk_tree_child_position(GTK_TREE(widget), current_child);
    GList glist = { current_child, NULL, NULL };
    gtk_tree_remove_items(GTK_TREE(widget), &glist);
    gtk_tree_insert(GTK_TREE(widget), new_child, pos);

  } else if (GTK_IS_CLIST(widget)) {
    /* For a clist, we check if the widget is a column title and replace if so.
       If not, we can ?? */
    g_warning("replacing child of clist - not implemented yet\n");

  } else if (GTK_IS_BUTTON(widget)) {
    /* For a button, we just delete the existing child and add the new one
     */
    gtk_container_remove(GTK_CONTAINER(widget), current_child);
    gtk_container_add(GTK_CONTAINER(widget), new_child);

  } else if (GTK_IS_CONTAINER(widget)) {
    /* General code for container - has to remove all children and add back
       NOTE: this may not work for specialised containers.
       NOTE: need to ref widgets?
       */
    g_warning("replacing child of container - not implemented yet\n");
  }

  /* Now allow a resize */
  gtk_container_unblock_resize(GTK_CONTAINER(widget));
}



/*
 * Tooltip settings
 */
gboolean
gb_widget_get_show_tooltips()
{
  return GTK_TOOLTIPS(gb_widget_tooltips)->enabled;
}


void
gb_widget_toggle_show_tooltips(GtkWidget* widget, gpointer data)
{
  if (GTK_CHECK_MENU_ITEM(widget)->active)
    gtk_tooltips_enable(gb_widget_tooltips);
  else
    gtk_tooltips_disable(gb_widget_tooltips);
}
