/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#define GB_CLASS_KEY		"GbClass"
#define GB_WIDGET_DATA_KEY	"GbWidgetData"

#define GB_LINE_BUFFER_SIZE	1024

#define GB_PROPERTIES_NOT_CREATED	-1
#define GB_PROPERTIES_NOT_NEEDED	-2

/* These keys have non-null values in a widget if it is a placeholder/holding
   widget */
#define GB_PLACEHOLDER_KEY	"GB_PLACEHOLDER"
#define GB_HOLDING_WIDGET_KEY	"GB_HOLDING_WIDGET"
/* Macros to see if a widget is a placeholder/holding widget */
#define GB_IS_PLACEHOLDER(w)	(gtk_object_get_data(GTK_OBJECT(w), GB_PLACEHOLDER_KEY) != NULL)
#define GB_IS_HOLDING_WIDGET(w) (gtk_object_get_data(GTK_OBJECT(w), GB_HOLDING_WIDGET_KEY) != NULL)

/*
 * The hash table associates a Gtk widget class name with a table of functions
 *  used by the builder, e.g. creating a new widget, saving the widget etc.
 */
extern GHashTable *gb_widget_table;
extern GHashTable *gb_widget_id_hash;


/* Flags in GbWidgetData */
enum
{
  GB_VISIBLE		= 1 << 0, /* If the widget is initially visible */
  GB_GRAB_DEFAULT	= 1 << 1, /* If it grabs the default */
  GB_GRAB_FOCUS		= 1 << 2, /* If it grabs the focus */
  GB_STYLE_INHERIT	= 1 << 3, /* If it inherit's its parent's style */
  GB_STYLE_ALLOCATED	= 1 << 4, /* If its style has been allocated (colors)*/
  GB_IS_POPUP		= 1 << 5, /* If it is a popup window */
  GB_ACTIVE		= 1 << 6, /* If it is initially active (toggles) */
};

typedef struct _GbWidgetData {
  guint16	 flags;
  gint16	 x;
  gint16	 y;
  gint16	 width;
  gint16	 height;
  gint		 events;
  gchar		*tooltip;
  GList		*signals;
  GList		*accelerators;
} GbWidgetData;

typedef struct _GbSignal {
  gchar		*signal_name;
  gchar		*handler;
  gchar		*object;
  gboolean	 after;
  gchar		*data;
} GbSignal;

typedef struct _GbAccelerator {
  guint8	 modifiers;
  guchar	 key;
  gchar		*signal;
} GbAccelerator;


/* Data structures passed to the GbWidget functions */

/* GbWidgetNewData - used when new widgets are created.
   name is the name of the new widget. type is its Gtk type id.
   callback is the function to be called when the widget has been created.
   It is used after a dialog has been shown to let the user configure a
   widget (e.g. setting the number of rows in a new vbox).
   widget is the placeholder to be replaced or the fixed container to add
   the widget to. x, y, width & height specify the position and size of the
   new widget, when adding to a fixed container.
   data1 & data2 are used to hold pointers to entry fields in the dialog boxes,
   so we can get at the values easily.
   */
typedef struct _GbWidgetNewData {
  gchar		*name;
  guint		 type;
  void		(*callback)(GtkWidget *widget, struct _GbWidgetNewData *data);
  GtkWidget	*widget;
  gint		 x, y, width, height;
  gboolean	 is_a_popup_window; /* Only used for GtkWindow */
  GbWidgetData	*widget_data;
} GbWidgetNewData;

typedef struct _GbWidgetCreatePropertiesData {

} GbWidgetCreatePropertiesData;

typedef struct _GbWidgetShowPropertiesData {
  GbWidgetData	*widget_data;
  GtkWidget	*holding_widget;
} GbWidgetShowPropertiesData;

typedef struct _GbWidgetApplyPropertiesData {
  GbWidgetData	*widget_data;
  GtkWidget	*holding_widget;
} GbWidgetApplyPropertiesData;

typedef struct _GbWidgetCreatePopupMenuData {

} GbWidgetCreatePopupMenuData;

typedef struct _GbWidgetLoadData {
  gchar		*filename;
  FILE		*fp;
  gint		 line_number;
  char		 line_buffer[GB_LINE_BUFFER_SIZE];
  char		*tag_name;  /* points into line_buffer */
  char		*tag_value; /* points into line_buffer */
  gint		 status;
} GbWidgetLoadData;

typedef struct _GbWidgetSaveData {
  GbWidgetData	*widget_data;
  GtkWidget	*holding_widget;
  gchar		*filename;
  FILE		*fp;
  gint		 indent;
  gint		 status;
} GbWidgetSaveData;

typedef struct _GbWidgetWriteSourceData {

} GbWidgetWriteSourceData;


/* The GbWidget struct, which consists of pointers to the functions
   implemented by each GbWidget. The properties_page_number is the page in
   the widget properties notebook corresponding to the GbWidget */
typedef struct _GbWidget {
  gchar		**pixmap_struct;
  GdkPixmap	 *gdkpixmap;
  GdkBitmap	 *mask;
  gchar		 *tooltip;
  gint		  properties_page_number;	/* Used internally */

  GtkWidget* (* gb_widget_new)		(GbWidgetNewData             *data);

  void (* gb_widget_create_properties)	(GtkWidget		     *widget,
					 GbWidgetCreatePropertiesData *data);
  void (* gb_widget_show_properties)	(GtkWidget                   *widget,
					 GbWidgetShowPropertiesData  *data);
  void (* gb_widget_apply_properties)	(GtkWidget                   *widget,
					 GbWidgetApplyPropertiesData *data);
  void (* gb_widget_create_popup_menu)	(GtkWidget                   *widget,
					 GbWidgetCreatePopupMenuData *data);
  void (* gb_widget_load)		(GtkWidget                   *widget,
					 GbWidgetLoadData            *data);
  void (* gb_widget_save)		(GtkWidget                   *widget,
					 GbWidgetSaveData            *data);
  void (* gb_widget_write_source)	(GtkWidget                   *widget,
					 GbWidgetWriteSourceData     *data);
} GbWidget;



/* Call this first, to initialize all GtkWidget & GbWidget types */
void	    gb_widgets_init();

void	    gb_widget_init_struct	(GbWidget		    *gbwidget);


GtkWidget*  gb_widget_new		(gchar			    *class_name,
					 GbWidgetNewData	    *data);
gchar*	    gb_widget_new_name		(gchar			    *class_name);
GtkWidget*  gb_widget_initialize	(GtkWidget		    *widget,
					 GbWidgetNewData	    *data);
void	    gb_widget_free_new_data	(GbWidgetNewData	    *data);


gint	    gb_widget_create_properties	(GtkWidget		    *widget);

void	    gb_widget_show_properties	(GtkWidget		    *widget);
void	    gb_widget_show_position_properties (GtkWidget	    *widget,
						GtkWidget	    *holding_widget,
						GbWidgetData	    *wdata);

void	    gb_widget_apply_properties	(GtkWidget		    *widget,
					 GtkWidget		    *property);

void	    gb_widget_show_popup_menu	(GtkWidget		    *widget,
					 GdkEventButton		    *event);
void	    gb_widget_load		(GtkWidget		    *widget,
					 GbWidgetLoadData	    *data);
void	    gb_widget_save		(GtkWidget		    *widget,
					 gpointer		     data);
void	    gb_widget_write_source	(GtkWidget		    *widget,
					 GbWidgetWriteSourceData    *data);

void	    gb_widget_replace_child	(GtkWidget		    *widget,
					 GtkWidget		    *current_child,
					 GtkWidget		    *new_child);

gboolean    gb_widget_get_show_tooltips	(void);
void	    gb_widget_toggle_show_tooltips (GtkWidget		    *widget,
					    gpointer		     data);
