/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <string.h>

#include <gtk/gtk.h>
#include "../gb.h"

/* Include the 21x21 icon pixmap for this widget, to be used in the palette */
#include "../graphics/combo.xpm"

/*
 * This is the GbWidget struct for this widget (see ../gbwidget.h).
 * It is initialized in the init() function at the end of this file
 */
static GbWidget gbwidget;

static gchar *Case	= "GtkCombo::case_sensitive";
static gchar *Arrows	= "GtkCombo::use_arrows";
static gchar *Always	= "GtkCombo::use_arrows_always";
static gchar *Items	= "GtkCombo::items";

#define	ITEMS_BUFFER_SIZE	8192
#define	ITEMS_INVALID		"GB_INVALID"

static void add_label(GtkWidget *widget, gchar *items);

/******
 * NOTE: To use these functions you need to uncomment them AND add a pointer
 * to the function in the GbWidget struct at the end of this file.
 ******/

/*
 * Creates a new GtkWidget of class GtkCombo, performing any specialized
 * initialization needed for the widget to work correctly in this environment.
 * If a dialog box is used to initialize the widget, return NULL from this
 * function, and call data->callback with your new widget when it is done.
 * If the widget needs a special destroy handler, add a signal here.
 */
/*
GtkWidget*
gb_combo_new(GbWidgetNewData *data)
{

}
*/



/*
 * Creates the components needed to edit the extra properties of this widget.
 */
void
gb_combo_create_properties(GtkWidget *widget, GbWidgetCreatePropertiesData *data)
{
  property_add_bool(Case, "Case Sens.:",
		    "If the searching is case sensitive");
  property_add_bool(Arrows, "Use Arrows:",
		    "If arrows can be used to change the value");
  property_add_bool(Always, "Use Always:",
		    "If arrows work even if the value is not in the list");
  property_add_text(Items, "Items:",
		    "The items in the combo list, one per line", 5);
}



/*
 * Shows the extra properties of the widget.
 */
void
gb_combo_show_properties(GtkWidget *widget, GbWidgetShowPropertiesData *data)
{
  gchar items[ITEMS_BUFFER_SIZE];

  property_set_bool(Case, GTK_COMBO(widget)->case_sensitive);
  property_set_bool(Arrows, GTK_COMBO(widget)->use_arrows);
  property_set_bool(Always, GTK_COMBO(widget)->use_arrows_always);

  /* Clear the buffer and make sure it's seen as valid */
  items[0] = '\0';
  items[1] = '0';
  gtk_container_foreach(GTK_CONTAINER(GTK_COMBO(widget)->list),
			(GtkCallback) add_label, &items);
  if (!strcmp(&items[1], ITEMS_INVALID)) {
    property_set_text(Items, "");
    property_set_sensitive(Items, FALSE);
  } else {
    property_set_text(Items, items);
    property_set_sensitive(Items, TRUE);
  }
}

void
add_label(GtkWidget *widget, gchar *items)
{
  gchar *label_text;

  /* Return if list has already been marked invalid (i.e. not all items are
     labels or the list is too big*/
  if (items[0] == '\0' && !strcmp(&items[1], ITEMS_INVALID)) return;

  if (GTK_IS_LABEL(GTK_BIN(widget)->child)) {
    gtk_label_get(GTK_LABEL(GTK_BIN(widget)->child), &label_text);
    if (strlen(items) + strlen(label_text) + 2 < ITEMS_BUFFER_SIZE) {
      strcat(items, label_text);
      strcat(items, "\n");
      return;
    }
  }

  /* If we reach here, the list item is not a label or the list is too big */
  items[0] = '\0';
  strcpy(&items[1], ITEMS_INVALID);
}


/*
 * Applies the properties which have just been edited.
 */
void
gb_combo_apply_properties(GtkWidget *widget, GbWidgetApplyPropertiesData *data)
{
  gboolean case_sensitive, arrows, arrows_always;
  gchar *items;
  GList *item_list = NULL;

  case_sensitive = property_get_bool(Case);
  if (Apply) gtk_combo_set_case_sensitive(GTK_COMBO(widget), case_sensitive);
  arrows = property_get_bool(Arrows);
  if (Apply) gtk_combo_set_use_arrows(GTK_COMBO(widget), arrows);
  arrows_always = property_get_bool(Always);
  if (Apply) gtk_combo_set_use_arrows_always(GTK_COMBO(widget), arrows_always);

  items = property_get_text(Items);
  if (Apply) {
    gchar *pos = items;
    gchar *items_end = &items[strlen(items)];

    while (pos < items_end) {
      gchar *item_end = strchr(pos, '\n');
      if (item_end == NULL) item_end = items_end;
      *item_end = '\0';
      item_list = g_list_append(item_list, pos);
      pos = item_end + 1;
    }
    if (item_list) gtk_combo_set_popdown_strings(GTK_COMBO(widget), item_list);
    else gtk_list_clear_items(GTK_LIST(GTK_COMBO(widget)->list), 0, -1);
    g_list_free(item_list);
  }
  g_free(items);
}



/*
 * Adds menu items to a context menu which is just about to appear!
 * Add commands to aid in editing a GtkCombo, with signals pointing to
 * other functions in this file.
 */
/*
void
gb_combo_create_popup_menu(GtkWidget *widget, GbWidgetCreatePopupMenuData *data)
{

}
*/



/*
 * Loads any attribute/values for this GtkCombo. 
 * Note: most of the widgets attributes will have already been read in by
 * the standard code.
 */
/*
void
gb_combo_load(GtkWidget *widget, GbWidgetLoadData *data)
{

}
*/


/*
 * Saves any special attribute/value pairs for this GtkCombo.
 * Note: as with load(), most work will already have been done.
 */
void
gb_combo_save(GtkWidget *widget, GbWidgetSaveData *data)
{
  gchar items[ITEMS_BUFFER_SIZE];

  save_bool(data, Case, GTK_COMBO(widget)->case_sensitive);
  save_bool(data, Arrows, GTK_COMBO(widget)->use_arrows);
  save_bool(data, Always, GTK_COMBO(widget)->use_arrows_always);

  /* Clear the buffer and make sure it's seen as valid */
  items[0] = '\0';
  items[1] = '0';
  gtk_container_foreach(GTK_CONTAINER(GTK_COMBO(widget)->list),
			(GtkCallback) add_label, &items);
  if (!strcmp(&items[1], ITEMS_INVALID)) {
    save_text(data, Items, "");
  } else {
    save_text(data, Items, items);
  }
}



/*
 * Writes the source code needed to create this widget.
 * You have to output everything necessary to create the widget here, though
 * there are some convenience functions to help.
 */
/*
void
gb_combo_write_source(GtkWidget *widget, GbWidgetWriteSourceData *data)
{

}
*/



/*
 * Initializes the GbWidget structure.
 * I've placed this at the end of the file so we don't have to include
 * declarations of all the functions.
 */
GbWidget*
gb_combo_init()
{
  /* Initialise the GTK type */
  gtk_combo_get_type();

  /* Initialize the GbWidget structure */
  gb_widget_init_struct(&gbwidget);

  /* Fill in the pixmap struct & tooltip */
  gbwidget.pixmap_struct = combo_xpm;
  gbwidget.tooltip = "Combo Box";

  /* Fill in any functions that this GbWidget has */
  gbwidget.gb_widget_create_properties	= gb_combo_create_properties;
  gbwidget.gb_widget_show_properties	= gb_combo_show_properties;
  gbwidget.gb_widget_apply_properties	= gb_combo_apply_properties;
/*
  gbwidget.gb_widget_new		= gb_combo_new;
  gbwidget.gb_widget_create_popup_menu	= gb_combo_create_popup_menu;
  gbwidget.gb_widget_load		= gb_combo_load;
*/
  gbwidget.gb_widget_save		= gb_combo_save;
/*
  gbwidget.gb_widget_write_source	= gb_combo_write_source;
*/

  return &gbwidget;
}

