/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include "../gb.h"

/* Include the 21x21 icon pixmap for this widget, to be used in the palette */
#include "../graphics/label.xpm"

/*
 * This is the GbWidget struct for this widget (see ../gbwidget.h).
 * It is initialized in the init() function at the end of this file
 */
static GbWidget gbwidget;

static gchar *Label	= "GtkLabel::label";
static gchar *Justify	= "GtkLabel::justify";
static gchar *XAlign	= "GtkLabel::xalign";
static gchar *YAlign	= "GtkLabel::yalign";
static gchar *XPad	= "GtkLabel::xpad";
static gchar *YPad	= "GtkLabel::ypad";

/* Note: Fill doesn't currently work with labels */
static gchar* GbJustifyChoices[] = { "Left", "Right", "Center",
				     /* "Fill", */ NULL};
static gint GbJustifyValues[] = {
  GTK_JUSTIFY_LEFT,
  GTK_JUSTIFY_RIGHT,
  GTK_JUSTIFY_CENTER,
  /*GTK_JUSTIFY_FILL*/
};

/******
 * NOTE: To use these functions you need to uncomment them AND add a pointer
 * to the function in the GbWidget struct at the end of this file.
 ******/

/*
 * Creates a new GtkWidget of class GtkLabel, performing any specialized
 * initialization needed for the widget to work correctly in this environment.
 * If a dialog box is used to initialize the widget, return NULL from this
 * function, and call data->callback with your new widget when it is done.
 * If the widget needs a special destroy handler, add a signal here.
 */
GtkWidget*
gb_label_new(GbWidgetNewData *data)
{
  return gtk_label_new(data->name);
}



/*
 * Creates the components needed to edit the extra properties of this widget.
 */
void
gb_label_create_properties(GtkWidget *widget, GbWidgetCreatePropertiesData *data)
{
  property_add_text(Label, "Label:", "The text to display", 2);
  property_add_choice(Justify, "Justify:",
		      "The justification of the lines of the label",
		      GbJustifyChoices);
  property_add_float_range(XAlign, "X Align:",
			   "The horizontal alignment of the entire label",
			   0, 1, 0.01, 1, 0, 2);
  property_add_float_range(YAlign, "Y Align:",
			   "The vertical alignment of the entire label",
			   0, 1, 0.01, 1, 0, 2);
  property_add_int_range(XPad, "X Pad:", "The horizontal padding",
			 0, 1000, 1, 10, 1);
  property_add_int_range(YPad, "Y Pad:", "The vertical padding",
			 0, 1000, 1, 10, 1);
}



/*
 * Shows the extra properties of the widget.
 */
void
gb_label_show_properties(GtkWidget *widget, GbWidgetShowPropertiesData *data)
{
  gchar *label_text;
  gint i;

  gtk_label_get(GTK_LABEL(widget), &label_text);
  property_set_text(Label, label_text);
  for (i = 0; i < sizeof(GbJustifyValues) / sizeof(GbJustifyValues[0]); i++) {
    if (GbJustifyValues[i] == GTK_LABEL(widget)->jtype)
      property_set_choice(Justify, i);
  }
  property_set_float(XAlign, GTK_MISC(widget)->xalign);
  property_set_float(YAlign, GTK_MISC(widget)->yalign);
  property_set_int(XPad, GTK_MISC(widget)->xpad);
  property_set_int(YPad, GTK_MISC(widget)->ypad);
}



/*
 * Applies the properties which have just been edited.
 */
void
gb_label_apply_properties(GtkWidget *widget, GbWidgetApplyPropertiesData *data)
{
  gchar *label, *justify;
  gfloat xalign, yalign;
  gint xpad, ypad, i;
  gboolean myApply;

  label = property_get_text(Label);
  if (Apply) gtk_label_set(GTK_LABEL(widget), label);
  g_free(label);

  justify = property_get_choice(Justify);
  if (Apply) {
    for (i = 0; i < sizeof(GbJustifyValues) / sizeof(GbJustifyValues[0]);i++) {
      if (!strcmp(justify, GbJustifyChoices[i])) {
	gtk_label_set_justify(GTK_LABEL(widget), GbJustifyValues[i]);
	break;
      }
    }
  }

  xalign = property_get_float(XAlign);
  myApply = Apply; /* save Apply flag */
  yalign = property_get_float(YAlign);
  if (Apply || myApply)
    gtk_misc_set_alignment(GTK_MISC(widget), xalign, yalign);

  xpad = property_get_int(XPad);
  myApply = Apply; /* save Apply flag */
  ypad = property_get_int(YPad);
  if (Apply || myApply)
    gtk_misc_set_padding(GTK_MISC(widget), xpad, ypad);
}



/*
 * Adds menu items to a context menu which is just about to appear!
 * Add commands to aid in editing a GtkLabel, with signals pointing to
 * other functions in this file.
 */
/*
void
gb_label_create_popup_menu(GtkWidget *widget, GbWidgetCreatePopupMenuData *data)
{

}
*/



/*
 * Loads any attribute/values for this GtkLabel. 
 * Note: most of the widgets attributes will have already been read in by
 * the standard code.
 */
/*
void
gb_label_load(GtkWidget *widget, GbWidgetLoadData *data)
{

}
*/


/*
 * Saves any special attribute/value pairs for this GtkLabel.
 * Note: as with load(), most work will already have been done.
 */
void
gb_label_save(GtkWidget *widget, GbWidgetSaveData *data)
{
  gchar *label_text;
  gint i;

  gtk_label_get(GTK_LABEL(widget), &label_text);
  save_text(data, Label, label_text);
  for (i = 0; i < sizeof(GbJustifyValues) / sizeof(GbJustifyValues[0]); i++) {
    if (GbJustifyValues[i] == GTK_LABEL(widget)->jtype)
      save_choice(data, Justify, GbJustifyChoices[i]);
  }
  save_float(data, XAlign, GTK_MISC(widget)->xalign);
  save_float(data, YAlign, GTK_MISC(widget)->yalign);
  save_int(data, XPad, GTK_MISC(widget)->xpad);
  save_int(data, YPad, GTK_MISC(widget)->ypad);
}



/*
 * Writes the source code needed to create this widget.
 * You have to output everything necessary to create the widget here, though
 * there are some convenience functions to help.
 */
/*
void
gb_label_write_source(GtkWidget *widget, GbWidgetWriteSourceData *data)
{

}
*/



/*
 * Initializes the GbWidget structure.
 * I've placed this at the end of the file so we don't have to include
 * declarations of all the functions.
 */
GbWidget*
gb_label_init()
{
  /* Initialise the GTK type */
  gtk_label_get_type();

  /* Initialize the GbWidget structure */
  gb_widget_init_struct(&gbwidget);

  /* Fill in the pixmap struct & tooltip */
  gbwidget.pixmap_struct = label_xpm;
  gbwidget.tooltip = "Label";

  /* Fill in any functions that this GbWidget has */
  gbwidget.gb_widget_new		= gb_label_new;
  gbwidget.gb_widget_create_properties	= gb_label_create_properties;
  gbwidget.gb_widget_show_properties	= gb_label_show_properties;
  gbwidget.gb_widget_apply_properties	= gb_label_apply_properties;
/*
  gbwidget.gb_widget_create_popup_menu	= gb_label_create_popup_menu;
  gbwidget.gb_widget_load		= gb_label_load;
*/
  gbwidget.gb_widget_save		= gb_label_save;
/*
  gbwidget.gb_widget_write_source	= gb_label_write_source;
*/

  return &gbwidget;
}

