/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include "../gb.h"

/* Include the 21x21 icon pixmap for this widget, to be used in the palette */
#include "../graphics/spinbutton.xpm"

/*
 * This is the GbWidget struct for this widget (see ../gbwidget.h).
 * It is initialized in the init() function at the end of this file
 */
static GbWidget gbwidget;

static gchar *ClimbRate	= "GtkSpinButton::climb_rate";
static gchar *Digits	= "GtkSpinButton::digits";
static gchar *Numeric	= "GtkSpinButton::numeric";
static gchar *Policy	= "GtkSpinButton::policy";
static gchar *Snap	= "GtkSpinButton::snap";


static gchar *Values[]	= {
  "GtkSpinButton::hvalue",
  "GtkSpinButton::hlower",
  "GtkSpinButton::hupper",
  "GtkSpinButton::hstep",
  "GtkSpinButton::hpage",
  "GtkSpinButton::hpage_size",
};

static gchar* GbPolicyChoices[] = { "Always", "If Valid", NULL};
static gint GbPolicyValues[] = {
  GTK_UPDATE_ALWAYS,
  GTK_UPDATE_IF_VALID
};


/******
 * NOTE: To use these functions you need to uncomment them AND add a pointer
 * to the function in the GbWidget struct at the end of this file.
 ******/

/*
 * Creates a new GtkWidget of class GtkSpinButton, performing any specialized
 * initialization needed for the widget to work correctly in this environment.
 * If a dialog box is used to initialize the widget, return NULL from this
 * function, and call data->callback with your new widget when it is done.
 * If the widget needs a special destroy handler, add a signal here.
 */
GtkWidget*
gb_spin_button_new(GbWidgetNewData *data)
{
  return gtk_spin_button_new(NULL, 0.0, 0);
}



/*
 * Creates the components needed to edit the extra properties of this widget.
 */
void
gb_spin_button_create_properties(GtkWidget *widget, GbWidgetCreatePropertiesData *data)
{
  property_add_float(ClimbRate, "Climb Rate:",
		     "The climb rate of the spinbutton, used in conjunction with the Page Increment");
  property_add_int_range(Digits, "Digits:",
			 "The number of decimal digits to show",
			 0, 5, 1, 1, 0);
  property_add_bool(Numeric, "Numeric:",
		    "If local conversion should be done to the value");
  property_add_choice(Policy, "Update Policy:",
		      "When value_changed signals are emitted",
		      GbPolicyChoices);
  property_add_bool(Snap, "Snap:",
		    "If the value is snapped to multiples of the step increment");
  property_add_adjustment(Values, GB_ADJUST_DEFAULT_LABELS);
}



/*
 * Shows the extra properties of the widget.
 */
void
gb_spin_button_show_properties(GtkWidget *widget, GbWidgetShowPropertiesData *data)
{
  gint i, update_policy;
  property_set_float(ClimbRate, GTK_SPIN_BUTTON(widget)->climb_rate);
  property_set_int(Digits, GTK_SPIN_BUTTON(widget)->digits);
  property_set_bool(Numeric, GTK_SPIN_BUTTON(widget)->numeric);

  /* This is a slight kludge since the spin_button's update policy is
     a set of flags rather than integer values */
  update_policy = GTK_SPIN_BUTTON(widget)->update_policy
    & (GTK_UPDATE_ALWAYS | GTK_UPDATE_IF_VALID);
  for (i = 0; i < sizeof(GbPolicyValues) / sizeof(GbPolicyValues[0]); i++) {
    if (GbPolicyValues[i] == update_policy)
      property_set_choice(Policy, i);
  }
  property_set_bool(Snap, GTK_SPIN_BUTTON(widget)->update_policy
		    & GTK_UPDATE_SNAP_TO_TICKS);
  property_show_adjustment(Values, GTK_SPIN_BUTTON(widget)->adjustment);
}



/*
 * Applies the properties which have just been edited.
 */
void
gb_spin_button_apply_properties(GtkWidget *widget, GbWidgetApplyPropertiesData *data)
{
  gfloat climb_rate;
  gint digits, snap_value, i;
  gchar *policy;
  gboolean numeric, snap, myApply;

  climb_rate = property_get_float(ClimbRate);
  /* No set function for this */
  if (Apply) GTK_SPIN_BUTTON(widget)->climb_rate = climb_rate;

  digits = property_get_int(Digits);
  if (Apply) gtk_spin_button_set_digits(GTK_SPIN_BUTTON(widget), digits);

  numeric = property_get_bool(Numeric);
  if (Apply) gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(widget), numeric);

  snap = property_get_bool(Snap);
  snap_value = snap ? GTK_UPDATE_SNAP_TO_TICKS : 0;
  myApply = Apply;

  policy = property_get_choice(Policy);
  if (Apply || myApply) {
    for (i = 0; i < sizeof(GbPolicyValues) / sizeof(GbPolicyValues[0]);i++) {
      if (!strcmp(policy, GbPolicyChoices[i])) {
	gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(widget),
					  GbPolicyValues[i] | snap_value);
	break;
      }
    }
  }

  if (property_apply_adjustment(Values, GTK_SPIN_BUTTON(widget)->adjustment))
    gtk_signal_emit_by_name (GTK_OBJECT(GTK_SPIN_BUTTON(widget)->adjustment),
			     "value_changed");
}



/*
 * Adds menu items to a context menu which is just about to appear!
 * Add commands to aid in editing a GtkSpinButton, with signals pointing to
 * other functions in this file.
 */
/*
void
gb_spin_button_create_popup_menu(GtkWidget *widget, GbWidgetCreatePopupMenuData *data)
{

}
*/



/*
 * Loads any attribute/values for this GtkSpinButton. 
 * Note: most of the widgets attributes will have already been read in by
 * the standard code.
 */
/*
void
gb_spin_button_load(GtkWidget *widget, GbWidgetLoadData *data)
{

}
*/


/*
 * Saves any special attribute/value pairs for this GtkSpinButton.
 * Note: as with load(), most work will already have been done.
 */
void
gb_spin_button_save(GtkWidget *widget, GbWidgetSaveData *data)
{
  gint i, update_policy;
  save_float(data, ClimbRate, GTK_SPIN_BUTTON(widget)->climb_rate);
  save_int(data, Digits, GTK_SPIN_BUTTON(widget)->digits);
  save_bool(data, Numeric, GTK_SPIN_BUTTON(widget)->numeric);

  /* This is a slight kludge since the spin_button's update policy is
     a set of flags rather than integer values */
  update_policy = GTK_SPIN_BUTTON(widget)->update_policy
    & (GTK_UPDATE_ALWAYS | GTK_UPDATE_IF_VALID);
  for (i = 0; i < sizeof(GbPolicyValues) / sizeof(GbPolicyValues[0]); i++) {
    if (GbPolicyValues[i] == update_policy)
      save_choice(data, Policy, GbPolicyChoices[i]);
  }
  save_bool(data, Snap,
	    GTK_SPIN_BUTTON(widget)->update_policy & GTK_UPDATE_SNAP_TO_TICKS);
  save_adjustment(data, Values, GTK_SPIN_BUTTON(widget)->adjustment);
}



/*
 * Writes the source code needed to create this widget.
 * You have to output everything necessary to create the widget here, though
 * there are some convenience functions to help.
 */
/*
void
gb_spin_button_write_source(GtkWidget *widget, GbWidgetWriteSourceData *data)
{

}
*/



/*
 * Initializes the GbWidget structure.
 * I've placed this at the end of the file so we don't have to include
 * declarations of all the functions.
 */
GbWidget*
gb_spin_button_init()
{
  /* Initialise the GTK type */
  gtk_spin_button_get_type();

  /* Initialize the GbWidget structure */
  gb_widget_init_struct(&gbwidget);

  /* Fill in the pixmap struct & tooltip */
  gbwidget.pixmap_struct = spinbutton_xpm;
  gbwidget.tooltip = "Spin Button";

  /* Fill in any functions that this GbWidget has */
  gbwidget.gb_widget_new		= gb_spin_button_new;
  gbwidget.gb_widget_create_properties	= gb_spin_button_create_properties;
  gbwidget.gb_widget_show_properties	= gb_spin_button_show_properties;
  gbwidget.gb_widget_apply_properties	= gb_spin_button_apply_properties;
/*
  gbwidget.gb_widget_create_popup_menu	= gb_spin_button_create_popup_menu;
  gbwidget.gb_widget_load		= gb_spin_button_load;
*/
  gbwidget.gb_widget_save		= gb_spin_button_save;
/*
  gbwidget.gb_widget_write_source	= gb_spin_button_write_source;
*/

  return &gbwidget;
}

