/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "gbwidget.h"
#include "editor.h"
#include "palette.h"
#include "property.h"
#include "load.h"
#include "save.h"
#include "utils.h"
#include "project.h"

/* Include pixmaps */
#include "graphics/window.xpm"
#include "graphics/dialog.xpm"
#include "graphics/fileseldialog.xpm"
#include "graphics/colorseldialog.xpm"
#include "graphics/popup.xpm"
#include "graphics/popupmenu.xpm"

enum
{
  GB_COMPONENT_WINDOW		= 0,
  GB_COMPONENT_DIALOG		= 1,
  GB_COMPONENT_POPUP		= 2,
  GB_COMPONENT_FILESEL		= 3,
  GB_COMPONENT_COLORSEL		= 4,
  GB_COMPONENT_POPUPMENU	= 5
};

GtkWidget	*win_main;
GtkWidget	*component_list;
GtkTooltips	*tooltips;
GtkWidget	*component_pixmaps[6];

gchar *current_directory = NULL;
gchar *project_filename = NULL;

static void create_main_window			(void);
static void destroy_window			(GtkWidget	    *widget,
						 gpointer	     data);

static void on_quit				(GtkWidget	    *widget,
						 gpointer	     data);

static void on_component_list_select		(GtkWidget	    *clist,
						 gint		     row,
						 gint		     column,
						 GdkEventButton	    *bevent,
						 gpointer	     data);
static void new_component			(GtkWidget	    *widget,
						 gpointer	     data);
static void show_component			(GtkWidget	    *widget,
						 gpointer	     data);
static gint on_key_press_event			(GtkWidget	    *widget,
						 GdkEventKey	    *event,
						 gpointer	     data);
static void on_delete				(GtkWidget	    *widget,
						 gpointer	     data);

static GtkWidget* new_pixmap			(GtkWidget	    *widget,
						 char		   **data);

static void show_about_dialog			(GtkWidget	    *widget,
						 gpointer	     data);

static void show_open_project_dialog		(GtkWidget	    *widget,
						 gpointer	     data);
static void real_open_project			(GtkWidget	    *widget,
						 GtkFileSelection   *filesel);
static void show_save_project_as_dialog		(GtkWidget	    *widget,
						 gpointer	     data);
static void save_project			(GtkWidget	    *widget,
						 gpointer	     data);
static void save_project_as			(GtkWidget	    *widget,
						 GtkFileSelection   *filesel);
static void real_save_project			(gint		     warn_before_overwrite);


int
main(int argc, char *argv[])
{
  gtk_init(&argc, &argv);

  gb_widgets_init();

  tooltips = gtk_tooltips_new();

  create_main_window();
  gtk_widget_show(win_main);

  editor_init(win_main);

  palette_show(NULL, NULL);
  property_show(NULL, NULL);

  gtk_main();

  return 0;
}


void
create_main_window()
{
  GtkWidget *vbox_main;

  GtkWidget *menubar;
  GtkWidget *menu;
  GtkWidget *menuitem;
  GtkAcceleratorTable *accelerator_table;
  GtkWidget *toolbar;

  GtkWidget *button;

  win_main = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(win_main, 160, 260);
  gtk_widget_set_uposition(win_main, 0, 0);

  gtk_widget_realize(win_main);

  gtk_signal_connect (GTK_OBJECT (win_main), "destroy",
		      GTK_SIGNAL_FUNC(destroy_window), NULL);
  gtk_signal_connect (GTK_OBJECT (win_main), "delete_event",
		      GTK_SIGNAL_FUNC(on_quit), NULL);
  gtk_signal_connect_after (GTK_OBJECT (win_main), "key_press_event",
			    GTK_SIGNAL_FUNC(on_key_press_event), NULL);

  gtk_window_set_title (GTK_WINDOW (win_main), "Gtk+ Builder");
  gtk_container_border_width (GTK_CONTAINER (win_main), 0);

  vbox_main = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (win_main), vbox_main);
  gtk_widget_show (vbox_main);


  /* Create accelerator table */
  accelerator_table = gtk_accelerator_table_new();
  gtk_window_add_accelerator_table(GTK_WINDOW(win_main), accelerator_table);


  /* create menu bar */
  menubar = gtk_menu_bar_new ();
  gtk_box_pack_start (GTK_BOX (vbox_main), menubar, FALSE, TRUE, 0);
  gtk_widget_show (menubar);


  /* create File menu */
  menuitem = gtk_menu_item_new_with_label ("File");
  gtk_menu_bar_append (GTK_MENU_BAR (menubar), menuitem);
  gtk_widget_show (menuitem);

  menu = gtk_menu_new ();
  gtk_menu_set_accelerator_table(GTK_MENU(menu), accelerator_table);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);

  menuitem = gtk_menu_item_new_with_label ("New");
  gtk_menu_append (GTK_MENU (menu), menuitem);

  gtk_tooltips_set_tip(tooltips, menuitem, "Create a new project", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Open...");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(show_open_project_dialog), NULL);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", 'O', GDK_CONTROL_MASK);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Open an existing project", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Save");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(save_project), NULL);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", 'S', GDK_CONTROL_MASK);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Save project", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Save As...");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(show_save_project_as_dialog), NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Save project to a new file", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new ();
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Quit");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(on_quit), NULL);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", 'Q', GDK_CONTROL_MASK);
  gtk_tooltips_set_tip(tooltips, menuitem, "Quit Builder", NULL);
  gtk_widget_show (menuitem);


  /* Create Edit menu */
  menuitem = gtk_menu_item_new_with_label ("Edit");
  gtk_menu_bar_append (GTK_MENU_BAR (menubar), menuitem);
  gtk_widget_show (menuitem);

  menu = gtk_menu_new ();
  gtk_menu_set_accelerator_table(GTK_MENU(menu), accelerator_table);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);

  menuitem = gtk_menu_item_new_with_label ("Cut");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", 'X', GDK_CONTROL_MASK);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Copy");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", 'C', GDK_CONTROL_MASK);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Paste");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", 'V', GDK_CONTROL_MASK);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Delete");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(on_delete), NULL);
  gtk_widget_install_accelerator(menuitem, accelerator_table,
				"activate", GDK_DELETE, GDK_CONTROL_MASK);
  gtk_widget_show (menuitem);


  /* Create View menu */
  menuitem = gtk_menu_item_new_with_label ("View");
  gtk_menu_bar_append (GTK_MENU_BAR (menubar), menuitem);
  gtk_widget_show (menuitem);

  menu = gtk_menu_new ();
  gtk_menu_set_accelerator_table(GTK_MENU(menu), accelerator_table);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);

  menuitem = gtk_menu_item_new_with_label ("Show Palette");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(palette_show),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Show the palette of widgets", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Show Properties");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(property_show),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Show the properties of the selected widget", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_check_menu_item_new_with_label ("Show Widget Tooltips");
  gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem),
				gb_widget_get_show_tooltips());
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(gb_widget_toggle_show_tooltips),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Shows the tooltips of created widgets", NULL);
  gtk_widget_show (menuitem);


  menuitem = gtk_menu_item_new_with_label ("Grid");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_show (menuitem);

  /* Create Grid submenu */
  menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);

  menuitem = gtk_check_menu_item_new_with_label ("Show Grid");
  gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem),
				editor_get_show_grid());
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(editor_toggle_show_grid),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Shows the grid (in fixed containers only)", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Set Grid Options...");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(editor_show_grid_settings_dialog),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Set the spacing between grid lines", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_check_menu_item_new_with_label ("Snap to Grid");
  gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(menuitem),
				editor_get_snap_to_grid());
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(editor_toggle_snap_to_grid),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Snap widgets to the grid (in fixed containers only)", NULL);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Set Snap Options...");
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(editor_show_snap_settings_dialog),
		      NULL);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_tooltips_set_tip(tooltips, menuitem, "Set which parts of a widget snap to the grid", NULL);
  gtk_widget_show (menuitem);

  /* Create Help menu */
  menuitem = gtk_menu_item_new_with_label ("Help");
  gtk_menu_bar_append (GTK_MENU_BAR (menubar), menuitem);
  gtk_widget_show (menuitem);

  menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);

  menuitem = gtk_menu_item_new_with_label ("Contents");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("Index");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new ();
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_show (menuitem);

  menuitem = gtk_menu_item_new_with_label ("About...");
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
		      GTK_SIGNAL_FUNC(show_about_dialog),
		      NULL);
  gtk_widget_show (menuitem);


  /* Create component pixmaps */
  component_pixmaps[GB_COMPONENT_WINDOW]    = new_pixmap(win_main,
							 window_xpm);
  component_pixmaps[GB_COMPONENT_DIALOG]    = new_pixmap(win_main,
							 dialog_xpm);
  component_pixmaps[GB_COMPONENT_POPUP]     = new_pixmap(win_main,
							 popup_xpm);
  component_pixmaps[GB_COMPONENT_FILESEL]   = new_pixmap(win_main,
							 fileseldialog_xpm);
  component_pixmaps[GB_COMPONENT_COLORSEL]  = new_pixmap(win_main,
							 colorseldialog_xpm);
  component_pixmaps[GB_COMPONENT_POPUPMENU] = new_pixmap(win_main,
							 popupmenu_xpm);


  /* Create toolbar */
  toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
  gtk_box_pack_start (GTK_BOX (vbox_main), toolbar, FALSE, TRUE, 0);
  gtk_widget_show (toolbar);

  gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
			   "New Window", "New window", "",
			   component_pixmaps[GB_COMPONENT_WINDOW],
			   (GtkSignalFunc) new_component,
			   (gpointer) GB_COMPONENT_WINDOW);
  gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
			   "New Dialog", "New dialog", "",
			   component_pixmaps[GB_COMPONENT_DIALOG],
			   (GtkSignalFunc) new_component,
			   (gpointer) GB_COMPONENT_DIALOG);
  gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
			   "New File Sel.",
			   "New file selection dialog", "",
			   component_pixmaps[GB_COMPONENT_FILESEL],
			   (GtkSignalFunc) new_component,
			   (gpointer) GB_COMPONENT_FILESEL);
  gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
			   "New Color Sel.",
			   "New color selection dialog", "",
			   component_pixmaps[GB_COMPONENT_COLORSEL],
			   (GtkSignalFunc) new_component,
			   (gpointer) GB_COMPONENT_COLORSEL);

  gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
			   "New Popup", "New pop-up window", "",
			   component_pixmaps[GB_COMPONENT_POPUP],
			   (GtkSignalFunc) new_component,
			   (gpointer) GB_COMPONENT_POPUP);
  gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
			   "New Popup Menu",
			   "New pop-up menu", "",
			   component_pixmaps[GB_COMPONENT_POPUPMENU],
			   (GtkSignalFunc) new_component,
			   (gpointer) GB_COMPONENT_POPUPMENU);
  /*gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));*/


  /* Create list of components */
  component_list = gtk_clist_new (1);
  gtk_clist_column_titles_hide(GTK_CLIST(component_list));
  gtk_clist_set_row_height(GTK_CLIST(component_list), 20);
  gtk_clist_set_column_width(GTK_CLIST(component_list), 0, 200);
  gtk_clist_set_policy(GTK_CLIST(component_list), GTK_POLICY_AUTOMATIC,
		       GTK_POLICY_AUTOMATIC);
  gtk_signal_connect (GTK_OBJECT (component_list), "select_row",
		      GTK_SIGNAL_FUNC(on_component_list_select), NULL);
  gtk_box_pack_start (GTK_BOX (vbox_main), component_list, TRUE, TRUE, 0);
  gtk_widget_show (component_list);


  /* Create button to view windows */
  button = gtk_button_new_with_label ("Show");
  gtk_box_pack_start (GTK_BOX (vbox_main), button, FALSE, FALSE, 5);
  gtk_widget_show (button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(show_component), NULL);
}


void
on_component_list_select (GtkWidget *clist,
			  gint row,
			  gint column,
			  GdkEventButton *bevent,
			  gpointer data)
{
  if (bevent && bevent->type == GDK_2BUTTON_PRESS) {
    show_component(NULL, NULL);
  }
}


static void
on_quit	(GtkWidget	    *widget,
	 gpointer	     data)
{
  /* Ask to save changes */

  gtk_widget_destroy(win_main);
}


void
destroy_window (GtkWidget  *widget,
		gpointer    data)
{
  /* Could the main window be destroyed without the delete event being
     sent? Need to tidy this up a bit. */
  if (widget == win_main) gtk_exit(0);
}


GtkWidget *
new_pixmap (GtkWidget *widget,
	    char **data)
{
  GdkPixmap *gdkpixmap;
  GdkBitmap *mask;
  GtkWidget *pixmap;

  gdkpixmap = gdk_pixmap_create_from_xpm_d(widget->window, &mask,
				   &widget->style->bg[GTK_STATE_NORMAL],
				   data);
  pixmap = gtk_pixmap_new (gdkpixmap, mask);
  return pixmap;
}

void
new_component (GtkWidget *widget,
	       gpointer   data)
{
  GtkWidget *window = NULL;
  gchar *name;
  GdkPixmap *gdkpixmap;
  GdkBitmap *mask;
  gint row;
  gint component = (gint) data;

  switch (component) {
  case GB_COMPONENT_WINDOW:
    window = editor_new_window();
    break;
  case GB_COMPONENT_DIALOG:
    window = editor_new_dialog();
    break;
  case GB_COMPONENT_POPUP:
    window = editor_new_popup();
    break;
  case GB_COMPONENT_FILESEL:
    window = editor_new_filesel();
    break;
  case GB_COMPONENT_COLORSEL:
    window = editor_new_colorsel();
    break;
  case GB_COMPONENT_POPUPMENU:
    window = editor_new_popupmenu();
    break;
  default:
    g_warning("Invalid component\n");
    return;
  }

  g_return_if_fail(window != NULL);
  name = gtk_widget_get_name(window);
  row = gtk_clist_append(GTK_CLIST (component_list), NULL);
  gtk_clist_set_row_data(GTK_CLIST (component_list), row, window);

  gtk_pixmap_get(GTK_PIXMAP(component_pixmaps[component]), &gdkpixmap, &mask);
  gtk_clist_set_pixtext(GTK_CLIST (component_list), row, 0, name, 3,
			gdkpixmap, mask);
  
  gb_widget_show_properties(window);
}


void
show_component (GtkWidget *widget,
		gpointer   data)
{
  GtkWidget *component;
  GList *selection = GTK_CLIST(component_list)->selection;
  if (!selection) return;
  component = gtk_clist_get_row_data(GTK_CLIST(component_list),
				     (gint) selection->data);
  g_return_if_fail (component != NULL);
  gtk_widget_show(GTK_WIDGET(component));
  gdk_window_raise(GTK_WIDGET(component)->window);
}

gint
on_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
  guint key = event->keyval;
  switch (key) {
  case GDK_Delete:
    on_delete(widget, NULL);
    break;
  }
  return TRUE;
}

void
on_delete (GtkWidget *widget,
	   gpointer   data)
{
  GtkWidget *window;
  GList *selection = GTK_CLIST(component_list)->selection;

  if (selection) {
    window = gtk_clist_get_row_data(GTK_CLIST(component_list),
				    (gint) selection->data);
    project_delete_component(window);
  } else {
    editor_on_delete();
  }
}

void
project_delete_component (GtkWidget *widget)
{
  gint row = gtk_clist_find_row_from_data(GTK_CLIST(component_list), widget);
  g_return_if_fail(row != -1);
  gtk_clist_remove(GTK_CLIST(component_list), row);
  gtk_widget_destroy(GTK_WIDGET(widget));
}


void
project_foreach_component(GtkCallback   callback,
			  gpointer      callback_data)
{
  gint row;
  GtkWidget *window;
  for (row = 0; row < GTK_CLIST(component_list)->rows; row++) {
    window = gtk_clist_get_row_data(GTK_CLIST(component_list), row);
    if (window) (*callback)(window, callback_data);
  }
}


void
project_clear_component_selection()
{
  GList *selection = GTK_CLIST(component_list)->selection;
  while (selection) {
    GList *next = selection->next;
    gtk_clist_unselect_row (GTK_CLIST(component_list),
			    (gint) selection->data, 0);
    selection = next;
  }
}


void
show_about_dialog (GtkWidget *widget,
		   gpointer   data)
{
  gchar buf[256];
  /* VERSION comes from configure.in - the only place it should be defined */
  sprintf(buf, "Gtk+ Interface Builder Version %s\n\nBy Damon Chaplin",
	  VERSION);
  show_message_box(buf);
}



/*
 * Loading/Saving projects
 */

void
show_open_project_dialog (GtkWidget *widget,
	      gpointer   data)
{
  GtkWidget *filesel;

  filesel = gtk_file_selection_new("Open Project");
  gtk_window_position (GTK_WINDOW (filesel), GTK_WIN_POS_MOUSE);

  gtk_signal_connect_object (GTK_OBJECT (filesel), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (filesel));
  gtk_signal_connect_object (GTK_OBJECT (filesel), "delete_event",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(filesel));
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		      "clicked", GTK_SIGNAL_FUNC(real_open_project), filesel);
  gtk_signal_connect_object (GTK_OBJECT
			     (GTK_FILE_SELECTION (filesel)->cancel_button),
			     "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (filesel));
  gtk_widget_show(filesel);
}

void
real_open_project (GtkWidget        *widget,
		   GtkFileSelection *filesel)
{
  g_print ("%s\n", gtk_file_selection_get_filename (GTK_FILE_SELECTION (filesel)));
  gtk_widget_destroy (GTK_WIDGET (filesel));
}


void
show_save_project_as_dialog (GtkWidget *widget,
		 gpointer   data)
{
  GtkWidget *filesel;

  filesel = gtk_file_selection_new("Save Project");
  gtk_window_position (GTK_WINDOW (filesel), GTK_WIN_POS_MOUSE);

  gtk_signal_connect_object (GTK_OBJECT (filesel), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (filesel));
  gtk_signal_connect_object (GTK_OBJECT (filesel), "delete_event",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(filesel));
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
		      "clicked", GTK_SIGNAL_FUNC(save_project_as), filesel);
  gtk_signal_connect_object (GTK_OBJECT
			     (GTK_FILE_SELECTION (filesel)->cancel_button),
			     "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (filesel));
  gtk_widget_show(filesel);
}


void
save_project_as (GtkWidget        *widget,
		 GtkFileSelection *filesel)
{
  if (project_filename) g_free(project_filename);
  project_filename = g_strdup(gtk_file_selection_get_filename (GTK_FILE_SELECTION(filesel)));
  real_save_project(TRUE);
  gtk_widget_destroy (GTK_WIDGET (filesel));
}


void
save_project (GtkWidget *widget,
	      gpointer   data)
{
  if (project_filename == NULL) show_save_project_as_dialog(widget, data);
  else real_save_project(FALSE);
}


void
real_save_project(gint warn_before_overwrite)
{
  GbWidgetSaveData *data = g_new(GbWidgetSaveData, 1);
  data->filename = project_filename;
  data->status = GB_OK;

  g_print ("Saving: %s\n", data->filename);
  save_project_file(data);

  if (data->status != GB_OK && data->status != GB_EOF) {
    g_print ("Error saving file.\n");
  }
  g_free(data);
}
