/*  Gtk+ User Interface Builder
 *  Copyright (C) 1998  Damon Chaplin
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <string.h>

/* These were used to try to get the real window positions, but it didn't work
#include <X11/Xlib.h>
#include <gdk/gdkprivate.h>
*/
#include <gtk/gtk.h>

#include "editor.h"
#include "project.h"
#include "keys.h"
#include "utils.h"
#include "gbwidget.h"
#include "property.h"

/* These are the standard widget attribute names */
gchar*	GbName		= "GtkWidget::name";
gchar*	GbX		= "GtkWidget::x";
gchar*	GbY		= "GtkWidget::y";
gchar*	GbWidth		= "GtkWidget::width";
gchar*	GbHeight	= "GtkWidget::height";
gchar*	GbVisible	= "GtkWidget::visible";
gchar*	GbSensitive	= "GtkWidget::sensitive";
gchar*	GbTooltip	= "GtkWidget::tooltip";
gchar*	GbCanDefault	= "GtkWidget::can_default";
gchar*	GbHasDefault	= "GtkWidget::has_default";
gchar*	GbCanFocus	= "GtkWidget::can_focus";
gchar*	GbHasFocus	= "GtkWidget::has_focus";
gchar*	GbEvents	= "GtkWidget::events";
gchar*	GbExtEvents	= "GtkWidget::extension_events";

gchar*	GbStyleInherit	= "GtkWidget::style_inherit";
gchar*	GbStyleName	= "GtkWidget::style_name";
gchar*	GbStyleFont	= "GtkWidget::style_font";

/* For children of a table */
gchar*	GbCellX		= "GtkWidget::cell_x";
gchar*	GbCellY		= "GtkWidget::cell_y";
gchar*	GbColSpan	= "GtkWidget::col_span";
gchar*	GbRowSpan	= "GtkWidget::row_span";
gchar*	GbXPad		= "GtkWidget::xpad";
gchar*	GbYPad		= "GtkWidget::ypad";
gchar*	GbXExpand	= "GtkWidget::xexpand";
gchar*	GbYExpand	= "GtkWidget::yexpand";
gchar*	GbXShrink	= "GtkWidget::xshrink";
gchar*	GbYShrink	= "GtkWidget::yshrink";
gchar*	GbXFill		= "GtkWidget::xfill";
gchar*	GbYFill		= "GtkWidget::yfill";

/* For children of a box */
gchar*	GbPadding	= "GtkWidget::padding";
gchar*	GbExpand	= "GtkWidget::expand";
gchar*	GbFill		= "GtkWidget::fill";
gchar*	GbPack		= "GtkWidget::pack";

/* Signals page */
gchar*	GbSignalName	= "GtkWidget::signal_name";
gchar*	GbSignalHandler	= "GtkWidget::signal_handler";
gchar*	GbSignalObject	= "GtkWidget::signal_object";
gchar*	GbSignalAfter	= "GtkWidget::signal_after";
gchar*	GbSignalData	= "GtkWidget::signal_data";

/* Accelerators page */
gchar*	GbAccelKey	= "GtkWidget::accelerator_key";
gchar*	GbAccelSignal	= "GtkWidget::accelerator_signal";



gchar*  GbStateNames[] =	{ "NORMAL", "ACTIVE", "PRELIGHT",
				  "SELECTED", "INSENSITIVE" };
gchar*  GbStateTabLabels[] =	{ "Normal", "Active", "Prelight",
				  "Select", "Insens" };
gchar*  GbColorNames[] =	{ "fg", "bg", "light", "dark",
				  "middle", "text", "base" };
gchar*  GbBgPixmapName =	"bg_pixmap";

/* Keys in object data hash used to store color, bg pixmap & filename */
gchar*	GbColorKey		= "GbColor";
gchar*	GbBgPixmapKey		= "GbBgPixmap";
gchar*	GbBgMaskKey		= "GbBgMask";
gchar*	GbBgFilenameKey		= "GbBgFilename";
gchar*	GbValueKey		= "GbValue";
gchar*	GbDialogValueKey	= "GbValue";

/* Hashtables of all the label & value widgets on the various property pages */
GHashTable *gb_property_labels  = NULL;
GHashTable *gb_property_values  = NULL;
GHashTable *gb_property_buttons = NULL;

/* Major components of the property window */
#define GB_PAGE_SIGNALS	5
GtkWidget *win_property = NULL;
GtkWidget *main_notebook = NULL;
GtkWidget *property_widget_notebook;
GtkWidget *special_child_property_notebook;
GtkWidget *apply_button = NULL;
GtkWidget *styles_notebook;

/* Extension mode choices data */
gchar* GbExtensionModeChoices[] = { "None", "All", "Cursor", NULL };
gint GbExtensionModeValues[] = {
  GDK_EXTENSION_EVENTS_NONE,
  GDK_EXTENSION_EVENTS_ALL,
  GDK_EXTENSION_EVENTS_CURSOR
};


/*
 * Private variables
 */

/* The widget whose properties are currently shown (or NULL) */
static GtkWidget *property_widget = NULL;

/* The current table & row used when creating properties */
static GtkWidget *property_table;
static gint property_table_row;

/* This is used when automatically applying properties as they are changed.
   It is the property widget (entry/spinbutton etc.) which has just changed,
   or NULL if all properties are to be applied. */
static gboolean auto_apply;
static GtkWidget *property_to_apply = NULL;

/* Color selection dialog */
static GtkColorSelectionDialog *colorsel = NULL;
static GtkWidget *color_value = NULL;

/* File selection dialog */
static GtkFileSelection *filesel = NULL;
static GtkWidget *filename_value = NULL;

/* Labels for boolean property - shown in the togglebuttons */
static gchar* GbBoolTrueString = "Yes";
static gchar* GbBoolFalseString = "No";

/* Checkbuttons on Accelerators page */
static GtkWidget *accel_control_button;
static GtkWidget *accel_shift_button;
static GtkWidget *accel_alt_button;

/* Clipboard used for copying/pasting colors */
#define GB_COLOR_CLIPBD_EMPTY    1
#define GB_COLOR_CLIPBD_COLOR    2
#define GB_COLOR_CLIPBD_BGPIXMAP 3
#define GB_COLOR_CLIPBD_STATE    4
static gint clipbd_state = GB_COLOR_CLIPBD_EMPTY;
static guchar clipbd_colors[7][4];
static GdkPixmap *clipbd_bgpixmap;
static GdkBitmap *clipbd_bgmask;
static gchar *clipbd_bgfilename;
static GtkWidget *selected_style_item = NULL;


/*
 * Private functions
 */

static GtkWidget* create_widget_property_page();
static GtkWidget* create_standard_property_page();
static GtkWidget* create_special_child_properties();
static GtkWidget* create_style_property_page();
static GtkWidget* create_style_page(gchar *state);
static GtkWidget* create_signals_property_page();
static GtkWidget* create_accelerators_property_page();

static void property_add(gchar *property_name, gchar *label_string,
			 GtkWidget *value, GtkWidget *dialog_button,
			 gchar *tooltip);
  
static GtkWidget* create_color_preview();
static void show_color_in_preview(GtkWidget *preview, gpointer data);
static void show_pixmap_in_drawing_area(GtkWidget *drawing_area, 
					GdkPixmap *gdkpixmap);


static gint expose_pixmap(GtkWidget *drawing_area, GdkEventExpose *event,
			  gpointer data);
static void draw_pixmap(GtkWidget *widget, GdkRectangle *area, gpointer data);
static void draw_pixmap_focus(GtkWidget *widget, gpointer data);
static void real_draw_pixmap(GtkWidget *drawing_area);

static void show_events_dialog(GtkWidget *widget, gpointer value);
static void on_events_dialog_ok(GtkWidget *widget, GtkWidget *clist);

static void on_keys_dialog_ok(GtkWidget *widget, GtkWidget *clist);
static void show_keys_dialog(GtkWidget *widget, gpointer value);
static void on_keys_clist_select(GtkWidget *widget, gint row, gint column,
				 GdkEventButton *bevent, gpointer data);

static void show_signals_dialog(GtkWidget *widget, gpointer value);
static void on_signals_dialog_ok(GtkWidget *widget, GtkWidget *clist);
static void on_signals_clist_select(GtkWidget *widget, gint row, gint column,
				    GdkEventButton *bevent, gpointer data);


static void on_accelerator_add(GtkWidget *widget, GtkWidget *clist);
static void on_accelerator_update(GtkWidget *widget, GtkWidget *clist);
static void on_accelerator_delete(GtkWidget *widget, GtkWidget *clist);
static void on_accelerator_clear(GtkWidget *widget, GtkWidget *clist);
static void on_accelerator_select (GtkWidget *clist, gint row, gint column,
				   GdkEventButton *event, gpointer user_data);

static void on_signal_add(GtkWidget *widget, GtkWidget *clist);
static void on_signal_update(GtkWidget *widget, GtkWidget *clist);
static void on_signal_delete(GtkWidget *widget, GtkWidget *clist);
static void on_signal_clear(GtkWidget *widget, GtkWidget *clist);
static void on_signal_select (GtkWidget *clist, gint row, gint column,
			      GdkEventButton *event, gpointer user_data);

static void on_style_copy(GtkWidget *widget, gpointer data);
static void on_style_copy_all(GtkWidget *widget, gpointer date);
static void on_style_paste(GtkWidget *widget, gpointer data);

/* Currently unused - its all auto-applied
static void on_apply(GtkWidget *widget, gpointer data);
*/
static void on_property_changed(GtkWidget *widget, gpointer data);

static void set_pixmap(GtkWidget *drawing_area, GdkPixmap *gdkpixmap,
		       GdkBitmap *mask, gchar *filename);

#define GB_EVENT_MASKS_COUNT 19

gchar *GbEventMasks[GB_EVENT_MASKS_COUNT * 2] = {
  "GDK_EXPOSURE_MASK",		    "When the window needs redrawing",
  "GDK_POINTER_MOTION_MASK",	    "When the mouse moves",
  "GDK_POINTER_MOTION_HINT_MASK",   "Mouse movement hints",
  "GDK_BUTTON_MOTION_MASK",	    "Mouse movement with any button pressed",
  "GDK_BUTTON1_MOTION_MASK",	    "Mouse movement with button 1 pressed",
  "GDK_BUTTON2_MOTION_MASK",	    "Mouse movement with button 2 pressed",
  "GDK_BUTTON3_MOTION_MASK",	    "Mouse movement with button 3 pressed",
  "GDK_BUTTON_PRESS_MASK",	    "Any mouse button pressed",
  "GDK_BUTTON_RELEASE_MASK",	    "Any mouse button released",
  "GDK_KEY_PRESS_MASK",		    "Any key pressed",
  "GDK_KEY_RELEASE_MASK",	    "Any key released",
  "GDK_ENTER_NOTIFY_MASK",	    "When the mouse enters the window",
  "GDK_LEAVE_NOTIFY_MASK",	    "When the mouse leaves the window",
  "GDK_FOCUS_CHANGE_MASK",	    "Any change in input focus",
  "GDK_STRUCTURE_MASK",		    "Any change in window structure",
  "GDK_PROPERTY_CHANGE_MASK",	    "Any change in X Windows property",
  "GDK_VISIBILITY_NOTIFY_MASK",     "Any change in visibility",
  "GDK_PROXIMITY_IN_MASK",	    "For cursors in XInput-aware programs",
  "GDK_PROXIMITY_OUT_MASK",	    "For cursors in XInput-aware programs"
};


gint GbEventMaskValues[GB_EVENT_MASKS_COUNT] = {
  GDK_EXPOSURE_MASK,
  GDK_POINTER_MOTION_MASK,
  GDK_POINTER_MOTION_HINT_MASK,
  GDK_BUTTON_MOTION_MASK,
  GDK_BUTTON1_MOTION_MASK,
  GDK_BUTTON2_MOTION_MASK,
  GDK_BUTTON3_MOTION_MASK,
  GDK_BUTTON_PRESS_MASK,
  GDK_BUTTON_RELEASE_MASK,
  GDK_KEY_PRESS_MASK,
  GDK_KEY_RELEASE_MASK,
  GDK_ENTER_NOTIFY_MASK,
  GDK_LEAVE_NOTIFY_MASK,
  GDK_FOCUS_CHANGE_MASK,
  GDK_STRUCTURE_MASK,
  GDK_PROPERTY_CHANGE_MASK,
  GDK_VISIBILITY_NOTIFY_MASK,
  GDK_PROXIMITY_IN_MASK,
  GDK_PROXIMITY_OUT_MASK
};


/*
 * Create the properties window
 */

void
property_create() {
  GtkWidget *vbox1, *label, *child;

  /* Create hash table for widgets containing property values */
  gb_property_labels  = g_hash_table_new(g_str_hash, g_str_equal);
  gb_property_values  = g_hash_table_new(g_str_hash, g_str_equal);
  gb_property_buttons = g_hash_table_new(g_str_hash, g_str_equal);

  /* Create property window */
  win_property = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_uposition(win_property, 180, 0);

  gtk_signal_connect (GTK_OBJECT (win_property), "delete_event",
		      GTK_SIGNAL_FUNC(property_hide), NULL);

  gtk_window_set_title (GTK_WINDOW (win_property), "Properties");
  gtk_container_border_width (GTK_CONTAINER (win_property), 0);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (win_property), vbox1);
  gtk_widget_show(vbox1);
 
  /* Create main notebook for different categories of widget properties */
  main_notebook = gtk_notebook_new ();
  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (main_notebook), GTK_POS_TOP);
  /*gtk_notebook_set_tab_border (GTK_NOTEBOOK (main_notebook), 2);*/
  gtk_box_pack_start (GTK_BOX (vbox1), main_notebook, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (main_notebook), 2);
  gtk_widget_show(main_notebook);

  /* Create the pages of the main notebook */
  /* NOTE: If you add/remove pages you need to change the GB_PAGE_SIGNALS
     value at the top of this file */
  label = gtk_label_new ("Widget");
  gtk_widget_show(label);
  child = create_widget_property_page();
  gtk_notebook_append_page (GTK_NOTEBOOK (main_notebook), child, label);

  label = gtk_label_new ("Place");
  gtk_widget_show(label);
  child = create_special_child_properties();
  gtk_notebook_append_page (GTK_NOTEBOOK (main_notebook), child, label);

  label = gtk_label_new ("Basic");
  gtk_widget_show(label);
  child = create_standard_property_page();
  gtk_notebook_append_page (GTK_NOTEBOOK (main_notebook), child, label);

  label = gtk_label_new ("Style");
  gtk_widget_show(label);
  child = create_style_property_page();
  gtk_notebook_append_page (GTK_NOTEBOOK (main_notebook), child, label);

  label = gtk_label_new ("Accel.");
  gtk_widget_show(label);
  child = create_accelerators_property_page();
  gtk_notebook_append_page (GTK_NOTEBOOK (main_notebook), child, label);

  label = gtk_label_new ("Sig.");
  gtk_widget_show(label);
  child = create_signals_property_page();
  gtk_notebook_append_page (GTK_NOTEBOOK (main_notebook), child, label);

  /* Apply/Close buttons at bottom */
  /*
  hbox = gtk_hbox_new(TRUE, 5);
  gtk_box_pack_end(GTK_BOX(vbox1), hbox, FALSE, TRUE, 5);
  gtk_widget_show(hbox);

  apply_button = gtk_button_new_with_label("Apply");
  gtk_box_pack_start(GTK_BOX(hbox), apply_button, TRUE, TRUE, 0);
  gtk_widget_show(apply_button);
  gtk_signal_connect (GTK_OBJECT (apply_button), "clicked",
		      GTK_SIGNAL_FUNC(on_apply), NULL);

  button = gtk_button_new_with_label("Close");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(property_hide), NULL);
		      */

  property_set_widget(NULL);
}


void
property_show (GtkWidget *widget,
	       gpointer   data)
{
  if (!win_property) property_create();
  gtk_widget_show(win_property);
  gdk_window_raise(GTK_WIDGET(win_property)->window);
}

gint
property_hide (GtkWidget *widget,
	       gpointer   data)
{
  gtk_widget_hide(win_property);
  return TRUE;
}


void
property_refresh()
{
  /* bug workaround - labels aren't redrawn correctly */
  if (win_property->window) editor_refresh_widget(win_property);
}


GtkWidget*
property_get_widget()
{
  return property_widget;
}


void
property_set_widget(GtkWidget *widget)
{
  gchar buffer[128];
  gchar *name;

  if (widget) {
    strcpy(buffer, "Properties: ");
    name = gtk_widget_get_name(widget);
    strncat(buffer, name, 120 - strlen(buffer));
    gtk_window_set_title(GTK_WINDOW(win_property), buffer);
    gtk_widget_set_sensitive(main_notebook, TRUE);
    if (apply_button) gtk_widget_set_sensitive(apply_button, TRUE);
    property_widget = widget;
  } else {
    gtk_window_set_title(GTK_WINDOW(win_property),"Properties: <none>");
    property_hide_gbwidget_page();
    gtk_widget_set_sensitive(main_notebook, FALSE);
    if (apply_button) gtk_widget_set_sensitive(apply_button, FALSE);
    /* Need to clear the property widget before clearing the name property,
       since auto-apply may be on, and hence we may clear the name of the
       currently shown widget. */
    property_widget = NULL;
    property_set_string(GbName, NULL);
  }
}

GtkWidget*
create_widget_property_page()
{
  GtkWidget *page, *table, *vbox;
  page = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(page),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_viewport_set_shadow_type(GTK_VIEWPORT(GTK_SCROLLED_WINDOW(page)
					    ->viewport), GTK_SHADOW_NONE);
  gtk_widget_show(page);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(page), vbox);

  table = gtk_table_new (9, 3, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, TRUE, 2);
  gtk_widget_show(table);

  property_set_table_position(table, 0);
  property_add_string(GbName, "Name:",
		      "The name of the widget");

  property_widget_notebook = gtk_notebook_new ();
  gtk_widget_show(property_widget_notebook);
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK (property_widget_notebook),
			      FALSE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (property_widget_notebook),
				FALSE);
  gtk_box_pack_start(GTK_BOX (vbox), property_widget_notebook, FALSE, TRUE, 0);
  /*gtk_container_border_width (GTK_CONTAINER (property_widget_notebook), 2);*/

  return page;
}

GtkWidget*
create_standard_property_page()
{
  GtkWidget *page;

  page = gtk_table_new (9, 3, FALSE);
  gtk_widget_show(page);

  property_set_table_position(page, 0);
  property_add_int_range(GbX, "X:",
			 "The left edge of the widget, relative to its parent",
			 -1, 10000, 1, 10, 1);
  property_add_int_range(GbY, "Y:",
			 "The top edge of the widget, relative to its parent",
			 -1, 10000, 1, 10, 1);
  property_add_int_range(GbWidth, "Width:",
			 "The width of the widget",
			 -1, 10000, 1, 10, 1);
  property_add_int_range(GbHeight, "Height:",
			 "The height of the widget",
			 -1, 10000, 1, 10, 1);

  property_add_bool(GbVisible, "Visible:",
		    "If the widget is initially visible");
  property_add_bool(GbSensitive, "Sensitive:",
		    "If the widget responds to input");
  property_add_string(GbTooltip, "Tooltip:",
		      "The tooltip to display if the mouse lingers over the widget");

  property_add_bool(GbCanDefault, "Can Default:",
		    "If the widget can be the default");
  property_add_bool(GbHasDefault, "Has Default:",
		    "If the widget is the default");
  property_add_bool(GbCanFocus, "Can Focus:",
		    "If the widget can accept the input focus");
  property_add_bool(GbHasFocus, "Has Focus:",
		    "If the widget has the input focus");

  property_add_dialog(GbEvents, "Events:",
		      "The X events that the widget receives",
		      FALSE, show_events_dialog);
  property_add_choice(GbExtEvents, "Ext.Events:",
		      "The X Extension events mode",
		      GbExtensionModeChoices);
  return page;
}

GtkWidget*
create_style_property_page()
{
  GtkWidget *page, *table, *label, *child, *hbox, *button;
  int i;

  page = gtk_vbox_new (FALSE, 0);
  gtk_widget_show(page);

  table = gtk_table_new (3, 3, FALSE);
  gtk_box_pack_start(GTK_BOX(page), table, FALSE, TRUE, 0);
  gtk_widget_show(table);

  property_set_table_position(table, 0);
  property_add_bool(GbStyleInherit, "Inherit:",
		    "Set True to use the same style as the parent widget");
  property_add_combo(GbStyleName, "Style Name:",
		     "The name of the style, which can be shared by several widgets", NULL);
  property_add_string(GbStyleFont, "Font:",
		      "The font to use for any text in the widget");

  /* Create notebook for 5 states */
  styles_notebook = gtk_notebook_new ();
  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (styles_notebook), GTK_POS_TOP);
  /*gtk_notebook_set_tab_border (GTK_NOTEBOOK (styles_notebook), 2);*/
  gtk_box_pack_start (GTK_BOX (page), styles_notebook, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (styles_notebook), 0);

  for (i = 0; i < 5; i++) {
    label = gtk_label_new (GbStateTabLabels[i]);
    gtk_widget_show(label);
    child = create_style_page(GbStateNames[i]);
    gtk_notebook_append_page (GTK_NOTEBOOK (styles_notebook), child, label);
  }
  gtk_widget_show(styles_notebook);

  /* Add/Update/Delete buttons at bottom */
  hbox = gtk_hbox_new(TRUE, 5);
  button = gtk_button_new_with_label("Copy");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_style_copy), NULL);

  button = gtk_button_new_with_label("Copy All");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_style_copy_all), NULL);

  button = gtk_button_new_with_label("Paste");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_style_paste), NULL);

  gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, TRUE, 3);
  gtk_widget_show(hbox);

  return page;
}

GtkWidget*
create_style_page(gchar *state)
{
  GtkWidget *page, *table;
  int i;
  gchar buffer[128];
  gchar *labels[]   = { "Foreground:", "Background:", "Text:", "Light:",
			"Middle:", "Dark:", "Base:" };
  gchar *tooltips[] = { "Foreground color", "Background color",	"Text color",
			"Light color", "Middle color", "Dark color",
			"Base color" };

  page = gtk_vbox_new (FALSE, 0);
  gtk_widget_show(page);

  /* Add the seven colours for this state */
  table = gtk_table_new (8, 3, FALSE);
  property_set_table_position(table, 0);
  for (i = 0; i < 7; i++) {
    sprintf(buffer, "GtkStyle::%s[%s]", GbColorNames[i], state);
    property_add_color(buffer, labels[i], tooltips[i]);
  }

  /* Add the background pixmap */
  sprintf(buffer, "GtkStyle::%s[%s]", GbBgPixmapName, state);
  property_add_pixmap(buffer, "Back. Pixmap:",
		    "The graphic to use as the background of the widget");

  gtk_box_pack_start(GTK_BOX(page), table, TRUE, TRUE, 0);
  gtk_widget_show(table);
  return page;
}


GtkWidget*
create_special_child_properties()
{
  GtkWidget *page, *table;

  page = gtk_vbox_new (FALSE, 0);
  gtk_widget_show(page);

  special_child_property_notebook = gtk_notebook_new ();
  gtk_widget_show(special_child_property_notebook);
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK (special_child_property_notebook),
			      FALSE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (special_child_property_notebook),
				FALSE);
  gtk_box_pack_start(GTK_BOX(page), special_child_property_notebook,
		     TRUE, TRUE, 0);

  /* Create table for special properties of widgets in a table */
  table = gtk_table_new (3, 3, FALSE);
  gtk_widget_show(table);
  gtk_notebook_append_page (GTK_NOTEBOOK (special_child_property_notebook),
			    table, NULL);

  property_set_table_position(table, 0);
  property_add_int_range(GbCellX, "Cell X:",
			 "The left edge of the widget in the table",
			 0, 10000, 1, 10, 1);
  property_add_int_range(GbCellY, "Cell Y:",
			 "The top edge of the widget in the table",
			 0, 10000, 1, 10, 1);

  property_add_int_range(GbColSpan, "Col Span:",
			 "The number of columns spanned by the widget in the table",
			 1, 10000, 1, 10, 1);
  property_add_int_range(GbRowSpan, "Row Span:",
			 "The number of rows spanned by the widget in the table",
			 1, 10000, 1, 10, 1);
  property_add_int_range(GbXPad, "H Padding:",
			 "The horizontal padding",
			 0, 10000, 1, 10, 1);
  property_add_int_range(GbYPad, "V Padding:",
			 "The vertical padding",
			 0, 10000, 1, 10, 1);
  property_add_bool(GbXExpand, "X Expand:",
		    "Set True to let the widget expand horizontally");
  property_add_bool(GbYExpand, "Y Expand:",
		    "Set True to let the widget expand vertically");
  property_add_bool(GbXShrink, "X Shrink:",
		    "Set True to let the widget shrink horizontally");
  property_add_bool(GbYShrink, "Y Shrink:",
		    "Set True to let the widget shrink vertically");
  property_add_bool(GbXFill, "X Fill:",
		    "Set True to let the widget fill its horizontal allocated area");
  property_add_bool(GbYFill, "Y Fill:",
		    "Set True to let the widget fill its vertical allocated area");

  /* Create table for special properties of widgets in a box */
  table = gtk_table_new (3, 3, FALSE);
  gtk_widget_show(table);
  gtk_notebook_append_page (GTK_NOTEBOOK (special_child_property_notebook),
			    table, NULL);

  property_set_table_position(table, 0);
  property_add_int_range(GbPadding, "Padding:",
			 "The widget's padding",
			 0, 10000, 1, 10, 1);
  property_add_bool(GbExpand, "Expand:",
		    "Set True to let the widget expand");
  property_add_bool(GbFill, "Fill:",
		    "Set True to let the widget fill its allocated area");
  property_add_bool(GbPack, "Pack Start:",
		    "Set True to pack the widget at the start of the box");

  return page;
}



gint
property_add_gbwidget_page(GtkWidget *page)
{
  gtk_notebook_append_page(GTK_NOTEBOOK(property_widget_notebook), page, NULL);
  return g_list_length(GTK_NOTEBOOK(property_widget_notebook)->children) - 1;
}

void
property_hide_gbwidget_page()
{
  gtk_widget_hide(property_widget_notebook);
}

void
property_show_gbwidget_page(gint page)
{
  gtk_notebook_set_page(GTK_NOTEBOOK(property_widget_notebook), page);
  gtk_widget_show(property_widget_notebook);
}

void
property_hide_special_child_page()
{
  gtk_widget_hide(special_child_property_notebook);
}

void
property_show_special_child_page(gint page)
{
  gtk_notebook_set_page(GTK_NOTEBOOK(special_child_property_notebook), page);
  gtk_widget_show(special_child_property_notebook);
}

void
on_style_copy(GtkWidget *widget, gpointer data)
{
  guchar *color;
  if (!selected_style_item) {
    show_message_box("You need to select a color or background to copy");
    return;
  }
  if (GTK_IS_PREVIEW(selected_style_item)) {
    color = gtk_object_get_data(GTK_OBJECT(selected_style_item), GbColorKey);
    g_return_if_fail(color != NULL);
    clipbd_colors[0][0] = color[0];
    clipbd_colors[0][1] = color[1];
    clipbd_colors[0][2] = color[2];
    clipbd_state = GB_COLOR_CLIPBD_COLOR;
  } else if (GTK_IS_DRAWING_AREA(selected_style_item)) {
    clipbd_bgpixmap = gtk_object_get_data(GTK_OBJECT(selected_style_item),
					  GbBgPixmapKey);
    clipbd_bgmask = gtk_object_get_data(GTK_OBJECT(selected_style_item),
					GbBgMaskKey);
    clipbd_bgfilename = gtk_object_get_data(GTK_OBJECT(selected_style_item),
					    GbBgFilenameKey);
    clipbd_state = GB_COLOR_CLIPBD_BGPIXMAP;
  } else {
    g_warning("Invalid selection in on_style_copy()");
  }
}

void
on_style_copy_all(GtkWidget *widget, gpointer data)
{
  gint page, i;
  guchar *color;
  gchar buffer[128];

  /* Find out which state is currently displayed */
  page = gtk_notebook_current_page(GTK_NOTEBOOK(styles_notebook));

  for (i = 0; i < 7; i++) {
    sprintf(buffer, "GtkStyle::%s[%s]", GbColorNames[i], GbStateNames[page]);
    color = property_get_color(buffer);
    clipbd_colors[i][0] = color[0];
    clipbd_colors[i][1] = color[1];
    clipbd_colors[i][2] = color[2];
  }
  sprintf(buffer, "GtkStyle::%s[%s]", GbBgPixmapName, GbStateNames[page]);
  clipbd_bgpixmap = property_get_pixmap(buffer, &clipbd_bgmask,
					&clipbd_bgfilename);

  clipbd_state = GB_COLOR_CLIPBD_STATE;
}

void
on_style_paste(GtkWidget *widget, gpointer data)
{
  guchar *color;
  gchar buffer[128];
  gint page, i;

  switch(clipbd_state) {
  case GB_COLOR_CLIPBD_EMPTY:
    show_message_box("You need to copy a color or background pixmap first");
    break;
  case GB_COLOR_CLIPBD_COLOR:
    if (!selected_style_item || !GTK_IS_PREVIEW(selected_style_item)) {
      show_message_box("You need to select a color to paste into");
      return;
    }
    color = gtk_object_get_data(GTK_OBJECT(selected_style_item), GbColorKey);
    g_return_if_fail(color != NULL);
    color[0] = clipbd_colors[0][0];
    color[1] = clipbd_colors[0][1];
    color[2] = clipbd_colors[0][2];
    show_color_in_preview(selected_style_item, NULL);
    break;
  case GB_COLOR_CLIPBD_BGPIXMAP:
    if (!selected_style_item || !GTK_IS_DRAWING_AREA(selected_style_item)) {
      show_message_box("You need to select a background pixmap to paste into");
      return;
    }
    set_pixmap(selected_style_item, clipbd_bgpixmap, clipbd_bgmask,
	       clipbd_bgfilename);
    break;
  case GB_COLOR_CLIPBD_STATE:
    page = gtk_notebook_current_page(GTK_NOTEBOOK(styles_notebook));

    for (i = 0; i < 7; i++) {
      sprintf(buffer, "GtkStyle::%s[%s]", GbColorNames[i], GbStateNames[page]);
      property_set_color(buffer, clipbd_colors[i]);
    }

    sprintf(buffer, "GtkStyle::%s[%s]", GbBgPixmapName, GbStateNames[page]);
    property_set_pixmap(buffer, clipbd_bgpixmap, clipbd_bgmask,
			clipbd_bgfilename);

    break;
  }
}

/*
 * Color selection dialog
 */

gint
close_colorsel_dialog(GtkWidget *widget, gpointer data)
{
  gint x, y;

  /*
  Window root, parent, *children;
  unsigned int nchildren;
  XWindowAttributes window_attrs;
  GdkWindowPrivate *gdkwindow_private = (GdkWindowPrivate*)
    (GTK_WIDGET(colorsel)->window);
  XQueryTree(gdkwindow_private->xdisplay, gdkwindow_private->xwindow,
	     &root, &parent, &children, &nchildren);
  XGetWindowAttributes(gdkwindow_private->xdisplay, parent, &window_attrs);
  g_print(" window x:%i y:%i\n", window_attrs.x, window_attrs.y);
  */

  /* remember position - doesn't work properly yet - moves down each time */
  gdk_window_get_origin(GTK_WIDGET(colorsel)->window, &x, &y);
  /*gdk_window_get_position(GTK_WIDGET(colorsel)->window, &x, &y);*/
  gtk_widget_set_uposition(GTK_WIDGET(colorsel), x, y);

  /*gdk_window_set_hints(GTK_WIDGET(colorsel)->window, x, y, 0, 0, 0, 0,
		       GDK_HINT_POS);*/
  gtk_widget_hide(GTK_WIDGET(colorsel));

  /*gtk_widget_unmap(GTK_WIDGET(colorsel));*/
  return TRUE;
}

void
update_color(GtkWidget *widget, gpointer data)
{
  guchar *color;
  gdouble rgb[4];
  GtkWidget *preview = GTK_BIN(color_value)->child;
  int i;
  
  gtk_color_selection_get_color(GTK_COLOR_SELECTION(colorsel->colorsel), rgb);
  color = gtk_object_get_data(GTK_OBJECT(preview), GbColorKey);
  g_return_if_fail(color != NULL);
  for (i = 0; i < 3; i++) color[i] = (guchar) 255.0 * rgb[i];
  show_color_in_preview(preview, NULL);
  close_colorsel_dialog(widget, NULL);
  on_property_changed(color_value, color_value);
}

void
show_colorsel_dialog(GtkWidget *widget, gpointer data)
{
  guchar *color;
  gdouble rgb[4];
  int i;

  /* Create the dialog if it doesn't exist yet */
  if (!colorsel) {
    colorsel = GTK_COLOR_SELECTION_DIALOG(gtk_color_selection_dialog_new("Color Selection Dialog"));
    /* Hide the Help button since it is not used */
    gtk_widget_hide(GTK_WIDGET(colorsel->help_button));
    /*gtk_window_position(GTK_WINDOW(colorsel), GTK_WIN_POS_CENTER);*/
    gtk_signal_connect_object (GTK_OBJECT (colorsel), "destroy",
			       GTK_SIGNAL_FUNC(gtk_widget_destroy),
			       GTK_OBJECT (colorsel));
    gtk_signal_connect (GTK_OBJECT (colorsel), "delete_event",
			GTK_SIGNAL_FUNC(close_colorsel_dialog), NULL);
    gtk_signal_connect (GTK_OBJECT(colorsel->cancel_button), "clicked",
			GTK_SIGNAL_FUNC(close_colorsel_dialog), NULL);
    gtk_signal_connect (GTK_OBJECT (colorsel->ok_button), "clicked",
			GTK_SIGNAL_FUNC(update_color), NULL);
  }

  color_value = GTK_WIDGET(data);
  g_return_if_fail(GTK_IS_FRAME(color_value));

  color = gtk_object_get_data(GTK_OBJECT(GTK_BIN(color_value)->child),
			      GbColorKey);
  g_return_if_fail(color != NULL);
  for (i = 0; i < 3; i++) rgb[i] = ((gdouble)color[i]) / 255;

  gtk_color_selection_set_color(GTK_COLOR_SELECTION(colorsel->colorsel), rgb);
  gtk_widget_map(GTK_WIDGET(colorsel));
  gtk_widget_show(GTK_WIDGET(colorsel));
  gdk_window_raise(GTK_WIDGET(colorsel)->window);
}


/*
 * File selection dialog
 */

gint
close_filesel_dialog(GtkWidget *widget, gpointer data)
{
  gint x, y;
  /* remember position - doesn't work properly yet - moves down each time */
  gdk_window_get_origin(GTK_WIDGET(filesel)->window, &x, &y);
  /*gdk_window_get_position(GTK_WIDGET(filesel)->window, &x, &y);*/
  gtk_widget_set_uposition(GTK_WIDGET(filesel), x, y);

  /*gdk_window_set_hints(GTK_WIDGET(filesel)->window, x, y, 0, 0, 0, 0,
		       GDK_HINT_POS);*/
  gtk_widget_hide(GTK_WIDGET(filesel));
  return TRUE;
}


void
update_file(GtkWidget *widget, gpointer data)
{
  GdkPixmap *gdkpixmap, *old_gdkpixmap;
  GdkBitmap *mask, *old_mask;
  gchar *filename, *old_filename, *filename_copy;
  GtkWidget *drawing_area;

  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));

  if (GTK_IS_ENTRY(filename_value)) {
    gtk_entry_set_text(GTK_ENTRY(filename_value), filename);
    close_filesel_dialog(widget, NULL);
    on_property_changed(filename_value, filename_value);
    return;
  }

  gdkpixmap = gdk_pixmap_create_from_xpm (widget->window, &mask,
					  &widget->style->bg[GTK_STATE_NORMAL],
					  filename);
  if (!gdkpixmap) {
    show_message_box("Couldn't create pixmap from file\n");
    return;
  }

  drawing_area = GTK_BIN(filename_value)->child;
  old_gdkpixmap = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgPixmapKey);
  if (old_gdkpixmap) gdk_pixmap_unref(old_gdkpixmap);
  old_mask = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgMaskKey);
  if (old_mask) gdk_bitmap_unref(old_mask);
  old_filename = gtk_object_get_data(GTK_OBJECT(drawing_area),GbBgFilenameKey);
  if (old_filename) g_free(old_filename);

  gtk_object_set_data(GTK_OBJECT(drawing_area), GbBgPixmapKey, gdkpixmap);
  gtk_object_set_data(GTK_OBJECT(drawing_area), GbBgMaskKey, mask);
  filename_copy = filename ? g_strdup(filename) : NULL;
  gtk_object_set_data(GTK_OBJECT(drawing_area), GbBgFilenameKey,filename_copy);

  if (gdkpixmap) gdk_pixmap_ref(gdkpixmap);
  if (mask) gdk_bitmap_ref(mask);

  close_filesel_dialog(widget, NULL);
  show_pixmap_in_drawing_area(drawing_area, gdkpixmap);
  gtk_widget_queue_draw(drawing_area);
  on_property_changed(filename_value, filename_value);
}

void
show_filesel_dialog(GtkWidget *widget, gpointer data)
{
  gchar *filename;
  /* value can be an Entry (for filename properties) or a Frame (for
     background pixmaps) */
  filename_value = GTK_WIDGET(data);

  /* Create the dialog if it doesn't exist yet */
  if (!filesel) {
    filesel = GTK_FILE_SELECTION(gtk_file_selection_new("Select File"));
    gtk_signal_connect_object (GTK_OBJECT (filesel), "destroy",
			       GTK_SIGNAL_FUNC(gtk_widget_destroy),
			       GTK_OBJECT (filesel));
    gtk_signal_connect (GTK_OBJECT (filesel), "delete_event",
			GTK_SIGNAL_FUNC(close_filesel_dialog), NULL);
    gtk_signal_connect (GTK_OBJECT(filesel->cancel_button), "clicked",
			GTK_SIGNAL_FUNC(close_filesel_dialog), NULL);
    gtk_signal_connect (GTK_OBJECT (filesel->ok_button), "clicked",
			GTK_SIGNAL_FUNC(update_file), NULL);
  }

  /* set to current file, if there is one */
  if (GTK_IS_ENTRY(filename_value)) {
    filename = gtk_entry_get_text(GTK_ENTRY(filename_value));
  } else {
    filename = gtk_object_get_data(GTK_OBJECT(GTK_BIN(filename_value)->child),
				   GbBgFilenameKey);
  }
  if (filename != NULL) {
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), filename);
  }
  gtk_widget_show(GTK_WIDGET(filesel));
  gdk_window_raise(GTK_WIDGET(filesel)->window);
}


/*
 *
 */


GtkWidget*
create_signals_property_page()
{
  GtkWidget *page, *clist, *table, *hbox, *button;
  gchar *titles[] = { "Signal", "Handler", "Data", "After", "Object" };

  page = gtk_vbox_new (FALSE, 0);
  gtk_widget_show(page);

  /* List of current signal handlers - Signal/Handler/Data/Options */
  clist = gtk_clist_new_with_titles(5, titles);
  gtk_clist_set_column_width(GTK_CLIST(clist), 0, 150);
  gtk_clist_set_column_width(GTK_CLIST(clist), 1, 150);
  gtk_clist_set_column_width(GTK_CLIST(clist), 2, 80);
  gtk_clist_set_column_width(GTK_CLIST(clist), 3, 30);
  gtk_clist_set_column_width(GTK_CLIST(clist), 4, 80);
  gtk_widget_show(clist);
  gtk_box_pack_start(GTK_BOX(page), clist, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (clist), "select_row",
		      GTK_SIGNAL_FUNC(on_signal_select), NULL);

  /* Mod, Key & Signal fields */
  table = gtk_table_new (3, 3, FALSE);
  gtk_widget_show(table);
  property_set_table_position(table, 0);
  property_add_dialog(GbSignalName, "Signal:",
		      "The signal to add a handler for", TRUE,
		      show_signals_dialog);
  property_add_string(GbSignalHandler, "Handler:",
		      "The function to handle the signal");
  property_add_string(GbSignalObject, "Object:",
		      "The object which receives the signal");
  property_add_bool(GbSignalAfter, "After:",
		    "If the handler runs after the class function");
  property_add_string(GbSignalData, "Data:",
		      "The data passed to the handler");
  gtk_box_pack_start(GTK_BOX(page), table, FALSE, TRUE, 5);

  /* Add/Update/Delete buttons at bottom */
  hbox = gtk_hbox_new(TRUE, 2);
  button = gtk_button_new_with_label("Add");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_signal_add), clist);

  button = gtk_button_new_with_label("Update");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_signal_update), clist);

  button = gtk_button_new_with_label("Delete");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_signal_delete), clist);

  button = gtk_button_new_with_label("Clear");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_signal_clear), clist);

  gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, TRUE, 5);
  gtk_widget_show(hbox);

  return page;
}

GtkWidget*
create_accelerators_property_page()
{
  GtkWidget *page, *clist, *hbox, *label, *button, *table;
  int row;
  gchar *titles[] = { "Mod", "Key", "Signal to emit" };

  page = gtk_vbox_new (FALSE, 0);
  gtk_widget_show(page);

  /* List of current accelerators - Mods/Keys/Signals */
  clist = gtk_clist_new_with_titles(3, titles);
  gtk_box_pack_start(GTK_BOX(page), clist, TRUE, TRUE, 0);
  gtk_clist_set_column_width(GTK_CLIST(clist), 0, 30);
  gtk_clist_set_column_width(GTK_CLIST(clist), 1, 120);
  gtk_clist_set_column_width(GTK_CLIST(clist), 2, 200);
  gtk_widget_show(clist);
  gtk_signal_connect (GTK_OBJECT (clist), "select_row",
		      GTK_SIGNAL_FUNC(on_accelerator_select), NULL);

  /* Mod, Key & Signal fields */
  table = gtk_table_new (3, 3, FALSE);
  gtk_widget_show(table);
  row = 0;
  label = gtk_label_new ("Modifiers:");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_widget_show(label);
  gtk_table_attach(GTK_TABLE (table), label, 0, 1, row, row + 1,
		   GTK_FILL, 0, 1, 1);

  hbox = gtk_hbox_new(FALSE, 2);
  gtk_widget_show(hbox);
  accel_control_button = gtk_check_button_new_with_label("Ctrl");
  gtk_widget_show(accel_control_button);
  gtk_box_pack_start(GTK_BOX(hbox), accel_control_button, TRUE, TRUE, 0);
  accel_shift_button = gtk_check_button_new_with_label("Shift");
  gtk_widget_show(accel_shift_button);
  gtk_box_pack_start(GTK_BOX(hbox), accel_shift_button, TRUE, TRUE, 0);
  accel_alt_button = gtk_check_button_new_with_label("Alt");
  gtk_widget_show(accel_alt_button);
  gtk_box_pack_start(GTK_BOX(hbox), accel_alt_button, TRUE, TRUE, 0);
  gtk_table_attach(GTK_TABLE (table), hbox, 1, 2, row, row + 1,
		   GTK_EXPAND | GTK_FILL, 0, 1, 1);

  property_set_table_position(table, ++row);
  property_add_dialog(GbAccelKey, "Key:",
		      "The accelerator key", TRUE, show_keys_dialog);
  property_add_dialog(GbAccelSignal, "Signal:",
		      "The signal to emit when the accelerator is pressed",
		      TRUE, show_signals_dialog);
  gtk_box_pack_start(GTK_BOX(page), table, FALSE, TRUE, 5);

  /* Add/Update/Delete buttons at bottom */
  hbox = gtk_hbox_new(TRUE, 2);
  button = gtk_button_new_with_label("Add");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_accelerator_add), clist);

  button = gtk_button_new_with_label("Update");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_accelerator_update), clist);

  button = gtk_button_new_with_label("Delete");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_accelerator_delete), clist);

  button = gtk_button_new_with_label("Clear");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  gtk_widget_show(button);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_accelerator_clear), clist);

  gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, TRUE, 5);
  gtk_widget_show(hbox);

  return page;
}


void
on_color_draw(GtkWidget *widget, gpointer data)
{
  if (widget == selected_style_item) {
    gdk_draw_rectangle (widget->window, widget->style->black_gc, FALSE, 0, 0,
			widget->allocation.width  - 1,
			widget->allocation.height - 1);
    gdk_draw_rectangle (widget->window, widget->style->white_gc, FALSE,	1, 1,
			widget->allocation.width  - 3,
			widget->allocation.height - 3);
    gdk_draw_rectangle (widget->window,	widget->style->black_gc, FALSE,	2, 2,
			widget->allocation.width  - 5,
			widget->allocation.height - 5);
  }
}


void
on_color_expose_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  on_color_draw(widget, data);
}


void
on_color_select(GtkWidget *widget, gpointer data)
{
  if (selected_style_item == widget) return;
  if (selected_style_item && selected_style_item != widget)
    gtk_widget_queue_draw(selected_style_item);
  selected_style_item = widget;
  gtk_widget_queue_draw(widget);
}


/*
 * Creating property widgets
 */

void
property_set_table_position(GtkWidget *table, gint row)
{
  property_table = table;
  property_table_row = row;
}

void
property_add_string		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value = gtk_entry_new();
  gtk_widget_set_usize(value, 80, -1);
  property_add(property_name, label, value, NULL, tooltip);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
}

void
property_add_text		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip,
				 gint		     visible_lines)
{
  GtkWidget *value = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(value), TRUE);
  gtk_widget_set_usize(value, 80, 34);
  property_add(property_name, label, value, NULL, tooltip);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
}

void
property_add_int		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value = gtk_entry_new();
  gtk_widget_set_usize(value, 80, -1);
  property_add(property_name, label, value, NULL, tooltip);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
}

void
property_add_int_range		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip,
				 gint		     min,
				 gint		     max,
				 gint		     step_increment,
				 gint		     page_increment,
				 gint		     climb_rate)
{
  GtkObject *adjustment = gtk_adjustment_new(min, min, max, step_increment,
					     page_increment, page_increment);
  GtkWidget *value = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),
					 climb_rate, 0);
  gtk_widget_set_usize(value, 80, -1);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
  property_add(property_name, label, value, NULL, tooltip);
}

void
property_add_float		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value = gtk_entry_new();
  gtk_widget_set_usize(value, 80, -1);
  property_add(property_name, label, value, NULL, tooltip);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
}

void
property_add_float_range	(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip,
				 gfloat		     min,
				 gfloat		     max,
				 gfloat		     step_increment,
				 gfloat		     page_increment,
				 gfloat		     climb_rate,
				 gint		     decimal_digits)
{
  GtkObject *adjustment = gtk_adjustment_new(min, min, max, step_increment,
					     page_increment, page_increment);
  GtkWidget *value = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment),
					 climb_rate, decimal_digits);
  gtk_widget_set_usize(value, 80, -1);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
  property_add(property_name, label, value, NULL, tooltip);
}

void
on_bool_property_toggle(GtkWidget *value, gpointer data)
{
  guint active = GTK_TOGGLE_BUTTON(value)->active;
  GtkWidget *label = GTK_BUTTON(value)->child;
  gtk_label_set(GTK_LABEL(label), active ? GbBoolTrueString
		: GbBoolFalseString);
  on_property_changed(value, value);
}

void
property_add_bool		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value = gtk_toggle_button_new_with_label(GbBoolFalseString);
  gtk_signal_connect (GTK_OBJECT (value), "toggled",
		      GTK_SIGNAL_FUNC(on_bool_property_toggle), NULL);
  property_add(property_name, label, value, NULL, tooltip);
}

void
property_add_choice		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip,
				 gchar		   **choices)
{
  GtkWidget *value = gtk_option_menu_new();
  GtkWidget *menu = gtk_menu_new ();

  gint i = 0;
  while (choices[i]) {
    GtkWidget *menuitem = gtk_menu_item_new_with_label ((gchar*) (choices[i]));
    gtk_menu_append (GTK_MENU (menu), menuitem);
    gtk_widget_show (menuitem);
    gtk_signal_connect (GTK_OBJECT (menuitem), "activate",
			GTK_SIGNAL_FUNC(on_property_changed), value);
    i++;
  }
  gtk_option_menu_set_menu (GTK_OPTION_MENU (value), menu);
  property_add(property_name, label, value, NULL, tooltip);
}

void
property_add_combo		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip,
				 GList		    *choices)
{
  GtkWidget *value = gtk_combo_new();
  gtk_widget_set_usize(GTK_COMBO(value)->entry, 60, -1);
  gtk_widget_set_usize(value, 80, -1);
  if (choices) gtk_combo_set_popdown_strings(GTK_COMBO(value), choices);
  gtk_signal_connect (GTK_OBJECT (GTK_COMBO(value)->entry), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
  property_add(property_name, label, value, NULL, tooltip);
}

void
property_add_color		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value, *preview, *dialog_button;
  value = gtk_frame_new(NULL);
  preview = create_color_preview();
  gtk_widget_set_events(preview, gtk_widget_get_events(preview)
			| GDK_BUTTON_PRESS_MASK);
  gtk_signal_connect_after (GTK_OBJECT (preview), "expose_event",
			    GTK_SIGNAL_FUNC(on_color_expose_event), value);
  gtk_signal_connect_after (GTK_OBJECT (preview), "draw",
			    GTK_SIGNAL_FUNC(on_color_draw), value);
  gtk_signal_connect (GTK_OBJECT (preview), "button_press_event",
		      GTK_SIGNAL_FUNC(on_color_select), value);
  
  gtk_container_add(GTK_CONTAINER(value), preview);
  dialog_button = gtk_button_new_with_label("...");
  gtk_signal_connect (GTK_OBJECT (dialog_button), "clicked",
		      GTK_SIGNAL_FUNC(show_colorsel_dialog), value);
  property_add(property_name, label, value, dialog_button, tooltip);
}

void
property_add_pixmap		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value, *drawing_area, *dialog_button;
  value = gtk_frame_new(NULL);
  drawing_area = gtk_drawing_area_new();
  gtk_widget_set_events(drawing_area, gtk_widget_get_events(drawing_area)
			| GDK_BUTTON_PRESS_MASK | GDK_EXPOSURE_MASK);
  gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 100, 20);
  gtk_widget_show(drawing_area);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
		      GTK_SIGNAL_FUNC(expose_pixmap), NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "draw",
		      GTK_SIGNAL_FUNC(draw_pixmap), NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "draw_default",
		      GTK_SIGNAL_FUNC(draw_pixmap_focus), NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "draw_focus",
		      GTK_SIGNAL_FUNC(draw_pixmap_focus), NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
		      GTK_SIGNAL_FUNC(on_color_select), value);
  gtk_container_add(GTK_CONTAINER(value), drawing_area);
  dialog_button = gtk_button_new_with_label("...");
  gtk_signal_connect (GTK_OBJECT (dialog_button), "clicked",
		      GTK_SIGNAL_FUNC(show_filesel_dialog), value);
  property_add(property_name, label, value, dialog_button, tooltip);
}


void
property_add_dialog		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip,
				 gboolean	     editable,
				 GtkCallback	     callback)
{
  GtkWidget *value, *dialog_button;
  value = gtk_entry_new();
  gtk_widget_set_usize(value, 80, -1);
  gtk_entry_set_editable(GTK_ENTRY(value), editable);
  dialog_button = gtk_button_new_with_label("...");
  gtk_signal_connect (GTK_OBJECT (dialog_button), "clicked",
		      GTK_SIGNAL_FUNC(callback), value);
  gtk_signal_connect (GTK_OBJECT (value), "changed",
		      GTK_SIGNAL_FUNC(on_property_changed), value);
  property_add(property_name, label, value, dialog_button, tooltip);
}


void
property_add_filename		(gchar		    *property_name,
				 gchar		    *label,
				 gchar		    *tooltip)
{
  GtkWidget *value, *dialog_button;
  value = gtk_entry_new();
  dialog_button = gtk_button_new_with_label("...");
  gtk_signal_connect (GTK_OBJECT (dialog_button), "clicked",
		      GTK_SIGNAL_FUNC(show_filesel_dialog), value);
  property_add(property_name, label, value, dialog_button, tooltip);
}


void
property_add(gchar		*property_name,
	     gchar		*label_string,
	     GtkWidget		*value,
	     GtkWidget		*dialog_button,
	     gchar		*tooltip)
{
  GtkWidget *label, *eventbox;
  gchar *property_name_copy;

  label = gtk_label_new (label_string);
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.3);
  /*gtk_misc_set_padding(GTK_MISC(label), 0, 5);*/
  gtk_widget_show(label);

  eventbox = gtk_event_box_new();
  gtk_widget_set_usize(eventbox, 80, -1);
  gtk_widget_show(eventbox);
  gtk_container_add(GTK_CONTAINER(eventbox), label);

  gtk_widget_show(value);
  /* Test */
  gtk_widget_set_usize(value, 80, -1);

  if (tooltip) gtk_tooltips_set_tip(tooltips, eventbox, tooltip, NULL);

  gtk_table_attach(GTK_TABLE (property_table), eventbox, 0, 1,
		   property_table_row, property_table_row + 1,
		   GTK_FILL, GTK_FILL, 1, 1);
  if (dialog_button) {
    gtk_table_attach(GTK_TABLE (property_table), value, 1, 2,
		     property_table_row, property_table_row + 1,
		     GTK_EXPAND | GTK_FILL, 0, 1, 1);
    gtk_widget_show(dialog_button);
    gtk_table_attach(GTK_TABLE (property_table), dialog_button, 2, 3,
		     property_table_row, property_table_row + 1,
		     0, 0, 1, 1);
  } else {
    gtk_table_attach(GTK_TABLE (property_table), value, 1, 3,
		     property_table_row, property_table_row + 1,
		     GTK_EXPAND | GTK_FILL, 0, 1, 1);
  }
  property_table_row++;

  /* Insert property label & value widgets into hash tables */
  property_name_copy = g_strdup(property_name);
  g_hash_table_insert(gb_property_labels, property_name_copy, label);
  g_hash_table_insert(gb_property_values, property_name_copy, value);
  if (dialog_button) g_hash_table_insert(gb_property_buttons, property_name_copy,
					 dialog_button);
}


/*
 * Functions for getting/setting properties.
 * NOTE: should also specify whether set properties are copied and
 *       if values returned by get should be freed.
 *       For widgets which use a GtkEntry: string value set is copied,
 *       & don't free string returned by get (it is the actual entry text).
 *
 * NOTE: property_get() functions also set the very special global variable
 *       Apply. This will be used to enable the automatic applying of
 *       properties as they are changed in the property editor. It is set
 *       to TRUE if the property should be applied.
 */

void
property_set_apply		(GtkWidget	    *property)
{
  property_to_apply = property;
}



gchar*
property_get_string		(gchar		    *property_name)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, "");
  return (gtk_entry_get_text(GTK_ENTRY(widget)));
}

void
property_set_string		(gchar		    *property_name,
				 gchar		    *value)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_entry_set_text(GTK_ENTRY(widget), value ? value : "");
}

/* Note: returned string must be freed with g_free() */
gchar*
property_get_text		(gchar		    *property_name)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, "");
  return (gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1));
}

void
property_set_text		(gchar		    *property_name,
				 gchar		    *value)
{
  gchar *text;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_editable_delete_text(GTK_EDITABLE(widget), 0, -1);
  text = value ? value : "";
  gtk_text_insert(GTK_TEXT(widget), NULL, NULL, NULL, text, strlen(text));
}

gint
property_get_int		(gchar		    *property_name)
{
  gchar *text;
  gint value = 0;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, 0);
  if (GTK_IS_SPIN_BUTTON(widget)) {
    return gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
  } else {
    text = gtk_entry_get_text(GTK_ENTRY(widget));
    sscanf(text, "%i", &value);
    return value;
  }
}

void
property_set_int		(gchar		    *property_name,
				 gint		     value)
{
  gchar buffer[128];
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  if (GTK_IS_SPIN_BUTTON(widget)) {
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), value);
  } else {
    sprintf(buffer, "%i", value);
    gtk_entry_set_text(GTK_ENTRY(widget), buffer);
  }
}

gfloat
property_get_float		(gchar		    *property_name)
{
  gchar *text;
  gfloat value = 0;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, 0);
  if (GTK_IS_SPIN_BUTTON(widget)) {
    return gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(widget));
  } else {
    text = gtk_entry_get_text(GTK_ENTRY(widget));
    sscanf(text, "%f", &value);
    return value;
  }
}

void
property_set_float		(gchar		    *property_name,
				 gfloat		     value)
{
  gchar buffer[128];
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  if (GTK_IS_SPIN_BUTTON(widget)) {
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), value);
  } else {
    sprintf(buffer, "%f", value);
    gtk_entry_set_text(GTK_ENTRY(widget), buffer);
  }
}

gboolean
property_get_bool		(gchar		    *property_name)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, FALSE);
  return (GTK_TOGGLE_BUTTON(widget)->active ? TRUE : FALSE);
}


void
property_set_bool		(gchar		    *property_name,
				 gint		     value)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(widget), value);
}

gchar*
property_get_choice		(gchar		    *property_name)
{
  GtkWidget *label;
  gchar *label_text;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, NULL);
  label = GTK_BUTTON(widget)->child;
  gtk_label_get(GTK_LABEL(label), &label_text);
  return label_text;
}

void
property_set_choice		(gchar		    *property_name,
				 gint		     value)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_option_menu_set_history(GTK_OPTION_MENU(widget), value);
}

gchar*
property_get_combo		(gchar		    *property_name)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, "");
  return (gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(widget)->entry)));
}

void
property_set_combo		(gchar		    *property_name,
				 gchar		    *value)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(widget)->entry), value);
}

guchar*
property_get_color		(gchar		    *property_name)
{
  guchar *color;
  GtkWidget *preview;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, NULL);
  preview = GTK_BIN(widget)->child;
  g_return_val_if_fail(GTK_IS_PREVIEW(preview), NULL);
  color = gtk_object_get_data(GTK_OBJECT(preview), GbColorKey);
  g_return_val_if_fail(color != NULL, NULL);
  return color;
}

void
property_set_color		(gchar		    *property_name,
				 guchar		    *value)
{
  GtkWidget *preview;
  guchar *color;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  preview = GTK_BIN(widget)->child;
  g_return_if_fail(GTK_IS_PREVIEW(preview));
  color = gtk_object_get_data(GTK_OBJECT(preview), GbColorKey);
  g_return_if_fail(color != NULL);
  color[0] = value[0];
  color[1] = value[1];
  color[2] = value[2];
  show_color_in_preview(preview, NULL);
}


GdkPixmap*
property_get_pixmap		(gchar		    *property_name,
				 GdkBitmap	   **mask,
				 gchar		   **filename)
{
  GtkWidget *drawing_area;
  GdkPixmap *gdkpixmap;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, NULL);
  drawing_area = GTK_BIN(widget)->child;
  g_return_val_if_fail(GTK_IS_DRAWING_AREA(drawing_area), NULL);

  gdkpixmap = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgPixmapKey);
  *mask = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgMaskKey);
  *filename = gtk_object_get_data(GTK_OBJECT(drawing_area),GbBgFilenameKey);
  return gdkpixmap;
}


void
property_set_pixmap		(gchar		    *property_name,
				 GdkPixmap	    *gdkpixmap,
				 GdkBitmap	    *mask,
				 gchar		    *filename)
{
  GtkWidget *drawing_area;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  drawing_area = GTK_BIN(widget)->child;
  g_return_if_fail(GTK_IS_DRAWING_AREA(drawing_area));

  set_pixmap(drawing_area, gdkpixmap, mask, filename);
}

void
set_pixmap(GtkWidget *drawing_area, GdkPixmap *gdkpixmap, GdkBitmap *mask,
	   gchar *filename)
{
  GdkPixmap *old_gdkpixmap;
  GdkBitmap *old_mask;
  gchar *old_filename, *filename_copy;

  /* free/unref any existing values */
  old_gdkpixmap = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgPixmapKey);
  if (old_gdkpixmap) gdk_pixmap_unref(old_gdkpixmap);
  old_mask = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgMaskKey);
  if (old_mask) gdk_bitmap_unref(old_mask);
  old_filename = gtk_object_get_data(GTK_OBJECT(drawing_area),GbBgFilenameKey);
  if (old_filename) g_free(old_filename);

  gtk_object_set_data(GTK_OBJECT(drawing_area), GbBgPixmapKey, gdkpixmap);
  gtk_object_set_data(GTK_OBJECT(drawing_area), GbBgMaskKey, mask);
  filename_copy = filename ? g_strdup(filename) : NULL;
  gtk_object_set_data(GTK_OBJECT(drawing_area), GbBgFilenameKey,filename_copy);

  if (gdkpixmap) gdk_pixmap_ref(gdkpixmap);
  if (mask) gdk_bitmap_ref(mask);

  show_pixmap_in_drawing_area(drawing_area, gdkpixmap);
}


gpointer
property_get_dialog		(gchar		    *property_name)
{
  gpointer dialog_value;
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, NULL);
  dialog_value = gtk_object_get_data(GTK_OBJECT(widget), GbDialogValueKey);
  if (dialog_value) return dialog_value;
  else return (gtk_entry_get_text(GTK_ENTRY(widget)));
}

void
property_set_dialog		(gchar		    *property_name,
				 gchar		    *string,
				 gpointer	     value)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  if (value) gtk_object_set_data(GTK_OBJECT(widget), GbDialogValueKey, value);
  gtk_entry_set_text(GTK_ENTRY(widget), value ? value : "");
}


gchar*
property_get_filename	(gchar		    *property_name)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  Apply = (!property_to_apply || property_to_apply == widget) ? TRUE : FALSE;
  g_return_val_if_fail(widget != NULL, "");
  return (gtk_entry_get_text(GTK_ENTRY(widget)));
}

void
property_set_filename	(gchar		    *property_name,
			 gchar		    *value)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_entry_set_text(GTK_ENTRY(widget), value ? value : "");
}


/*
 * Setting properties sensitive/insensitive
 */
void
property_set_sensitive(gchar *property_name, gboolean sensitive)
{
  GtkWidget *widget = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						       property_name);
  g_return_if_fail(widget != NULL);
  gtk_widget_set_sensitive(widget, sensitive);

  widget = (GtkWidget*) g_hash_table_lookup(gb_property_labels,
					    property_name);
  if (widget) gtk_widget_set_sensitive(widget, sensitive);

  widget = (GtkWidget*) g_hash_table_lookup(gb_property_buttons,
					    property_name);
  if (widget) gtk_widget_set_sensitive(widget, sensitive);
}


/*
 * Color previews - for showing style colors
 */

GtkWidget*
create_color_preview()
{
  GtkWidget *preview;
  guchar *color;

  gtk_widget_push_visual (gtk_preview_get_visual ());
  gtk_widget_push_colormap (gtk_preview_get_cmap ());
  preview = gtk_preview_new (GTK_PREVIEW_COLOR);
  gtk_preview_set_expand (GTK_PREVIEW (preview), TRUE);
  gtk_preview_size (GTK_PREVIEW (preview), 100, 20);
  gtk_widget_show (preview);
  gtk_signal_connect_after (GTK_OBJECT (preview), "size_allocate",
			    GTK_SIGNAL_FUNC (show_color_in_preview), NULL);

  color = g_new(guchar, 4);
  color[0] = color[1] = color[2] = color[3] = 255;
  gtk_object_set_data(GTK_OBJECT(preview), GbColorKey, color);
  return preview;
}


void
show_color_in_preview(GtkWidget *preview, gpointer data)
{
  static guchar *buf = NULL;
  static gint buf_width = 0;

  gint i, j, k;
  gint width  = preview->allocation.width;
  gint height = preview->allocation.height;
  guchar red, green, blue;
  guchar *color;

  g_return_if_fail(GTK_IS_PREVIEW(preview));
  color = gtk_object_get_data(GTK_OBJECT(preview), GbColorKey);
  g_return_if_fail(color != NULL);
  red   = color[0];
  green = color[1];
  blue  = color[2];

  if (buf == NULL || (width > buf_width)) {
    if (buf) g_free(buf);
    buf_width = width;
    buf = g_new(guchar, 3 * buf_width);
    g_return_if_fail(buf != NULL);
  }

  for (i = 0; i < height; i++) {
    for (j = 0, k = 0; j < width; j++) {
	  buf[k]   = red;
	  buf[k+1] = green;
	  buf[k+2] = blue;
	  k += 3;
    }
    gtk_preview_draw_row (GTK_PREVIEW (preview), buf, 0, i, width);
  }
  gtk_widget_queue_draw(preview);
}


/*
 * Pixmap values (displayed as background of drawing area)
 */
void
show_pixmap_in_drawing_area(GtkWidget *drawing_area, GdkPixmap *gdkpixmap)
{
  g_return_if_fail(GTK_IS_DRAWING_AREA(drawing_area));
  /* The drawing area doesn't get a window until the notebook page is shown!*/
  if (drawing_area->window == NULL) return;
  if (gdkpixmap) {
    gdk_window_set_back_pixmap(drawing_area->window, gdkpixmap, FALSE);
  } else {
    gdk_window_set_background(drawing_area->window,
			      &drawing_area->style->bg[GTK_STATE_NORMAL]);
  }
  gdk_window_clear(drawing_area->window);
}


gint
expose_pixmap(GtkWidget *drawing_area, GdkEventExpose *event, gpointer data)
{
  real_draw_pixmap(drawing_area);
  return FALSE;
}

void
draw_pixmap(GtkWidget *drawing_area, GdkRectangle *area, gpointer data)
{
  real_draw_pixmap(drawing_area);
}

void
draw_pixmap_focus(GtkWidget *drawing_area, gpointer data)
{
  real_draw_pixmap(drawing_area);
}

void
real_draw_pixmap(GtkWidget *drawing_area)
{
  GdkPixmap *gdkpixmap;
  g_return_if_fail(GTK_IS_DRAWING_AREA(drawing_area));
  gdkpixmap = gtk_object_get_data(GTK_OBJECT(drawing_area), GbBgPixmapKey);
  show_pixmap_in_drawing_area(drawing_area, gdkpixmap);
  on_color_draw(drawing_area, NULL);
}


/*
 * Events dialog
 */

void
show_events_dialog(GtkWidget *widget, gpointer value)
{
  GtkDialog *dialog;
  GtkWidget *clist, *button;
  gchar *titles[] = { "Event Mask", "Description" };
  int i;
  gchar *event_mask_string;
  gint event_mask_value;

  dialog = GTK_DIALOG(gtk_dialog_new());
  gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);

  gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));

  gtk_window_set_title (GTK_WINDOW (dialog), "Select X Events");

  clist = gtk_clist_new_with_titles(2, titles);
  gtk_clist_column_titles_passive (GTK_CLIST(clist));
  gtk_clist_set_column_width(GTK_CLIST(clist), 0, 230);
  gtk_clist_set_column_width(GTK_CLIST(clist), 1, 100);
  gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_MULTIPLE);
  gtk_clist_set_policy(GTK_CLIST(clist), GTK_POLICY_AUTOMATIC,
		       GTK_POLICY_AUTOMATIC);
  gtk_widget_set_usize(clist, 500, 350);
  gtk_widget_show(clist);
  gtk_box_pack_start(GTK_BOX(dialog->vbox), clist, TRUE, TRUE, 0);

  /* Insert events & descriptions */
  gtk_clist_freeze (GTK_CLIST(clist));
  for (i = 0; i < GB_EVENT_MASKS_COUNT; i++) {
    gtk_clist_append(GTK_CLIST(clist), &GbEventMasks[i * 2]);
  }

  /* Select rows according to current mask setting */
  event_mask_string = gtk_entry_get_text(GTK_ENTRY(value));
  event_mask_value = property_events_string_to_value(event_mask_string);
  for (i = 0; i < GB_EVENT_MASKS_COUNT; i++) {
    if (event_mask_value & GbEventMaskValues[i]) {
      gtk_clist_select_row(GTK_CLIST(clist), i, 0);
    }
  }
  gtk_clist_thaw (GTK_CLIST(clist));

  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(dialog->action_area), button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);
  /* Save pointer to value to use when OK pressed */
  gtk_object_set_data(GTK_OBJECT(clist), GbValueKey, value);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_events_dialog_ok), clist);

  button = gtk_button_new_with_label("Cancel");
  gtk_box_pack_start(GTK_BOX(dialog->action_area), button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_widget_show(GTK_WIDGET(dialog));
}


void
on_events_dialog_ok(GtkWidget *widget, GtkWidget *clist)
{
  gint row, mask_value = 0;
  GtkWidget *value;
  GList *selection = GTK_CLIST(clist)->selection;

  while (selection) {
    row = (gint) selection->data;
    mask_value |= GbEventMaskValues[row];
    selection = selection->next;
  }

  value = gtk_object_get_data(GTK_OBJECT(clist), GbValueKey);
  g_return_if_fail(value != NULL);
  gtk_entry_set_text(GTK_ENTRY(value),
		     property_events_value_to_string(mask_value));
  gtk_widget_destroy(gtk_widget_get_toplevel(clist));
}


gchar*
property_events_value_to_string(gint event_mask)
{
  static gchar buf[GB_EVENT_MASKS_COUNT + 2];
  int i;

  for (i = 0; i < GB_EVENT_MASKS_COUNT; i++) {
    if (event_mask & GbEventMaskValues[i])
      buf[GB_EVENT_MASKS_COUNT - i - 1] = '1';
    else buf[GB_EVENT_MASKS_COUNT - i - 1] = '0';
  }
  buf[GB_EVENT_MASKS_COUNT] = '0';
  buf[GB_EVENT_MASKS_COUNT + 1] = '\0';
  return buf;
}


gint
property_events_string_to_value(gchar *event_string)
{
  gint i, value = 0;

  if (strlen(event_string) < GB_EVENT_MASKS_COUNT) return 0;
  for (i = 0; i < GB_EVENT_MASKS_COUNT; i++) {
    if (event_string[GB_EVENT_MASKS_COUNT - i - 1] == '1')
      value |= GbEventMaskValues[i];
  }
  return value;
}


/*
 * Accelerator Keys dialog
 */

void
show_keys_dialog(GtkWidget *widget, gpointer value)
{
  GtkDialog *dialog;
  GtkWidget *clist, *button;
  int i;
  gchar *current_key;
  gchar *titles[] = { "Keys" };

  dialog = GTK_DIALOG(gtk_dialog_new());
  gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);

  gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));

  gtk_window_set_title (GTK_WINDOW (dialog), "Select Accelerator Key");

  clist = gtk_clist_new_with_titles(1, titles);
  gtk_clist_column_titles_passive (GTK_CLIST(clist));
  gtk_clist_set_policy(GTK_CLIST(clist), GTK_POLICY_AUTOMATIC,
		       GTK_POLICY_AUTOMATIC);
  gtk_widget_set_usize(clist, 280, 350);
  gtk_widget_show(clist);
  gtk_box_pack_start(GTK_BOX(dialog->vbox), clist, TRUE, TRUE, 0);

  /* Insert events & descriptions */
  gtk_clist_freeze (GTK_CLIST(clist));

  i = 0;
  current_key = gtk_entry_get_text(GTK_ENTRY(value));
  while (GbKeys[i]) {
    gtk_clist_append(GTK_CLIST(clist), &GbKeys[i]);
    if (!strcmp(current_key, GbKeys[i])) {
      gtk_clist_select_row(GTK_CLIST(clist), i, 0);
    }
    i++;
  }
  gtk_clist_thaw (GTK_CLIST(clist));
  gtk_signal_connect (GTK_OBJECT (clist), "select_row",
		      GTK_SIGNAL_FUNC(on_keys_clist_select), clist);

  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(dialog->action_area), button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);
  /* Save pointer to value to use when OK pressed */
  gtk_object_set_data(GTK_OBJECT(clist), GbValueKey, value);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_keys_dialog_ok), clist);

  button = gtk_button_new_with_label("Cancel");
  gtk_box_pack_start(GTK_BOX(dialog->action_area), button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_widget_show(GTK_WIDGET(dialog));
}


void
on_keys_dialog_ok(GtkWidget *widget, GtkWidget *clist)
{
  gint row;
  GtkWidget *value;
  GList *selection = GTK_CLIST(clist)->selection;

  if (selection) {
    row = (gint) selection->data;
    value = gtk_object_get_data(GTK_OBJECT(clist), GbValueKey);
    g_return_if_fail(value != NULL);
    gtk_entry_set_text(GTK_ENTRY(value), GbKeys[row]);
  }
  gtk_widget_destroy(gtk_widget_get_toplevel(clist));
}


void
on_keys_clist_select(GtkWidget *widget, gint row, gint column,
		     GdkEventButton *bevent, gpointer data)
{
  if (bevent && bevent->type == GDK_2BUTTON_PRESS) {
    on_keys_dialog_ok(widget, widget);
  }
}


/*
 * Signals dialog
 */

void
show_signals_dialog(GtkWidget *widget, gpointer value)
{
  GtkDialog *dialog;
  GtkWidget *clist, *button;
  gint row, i;
  gchar *current_signal, *name;
  gchar *titles[] = { "Signals" };
  guint *signals;
  guint nsignals;
  GtkSignalQuery *query_info;
  guint type;
  gpointer class;
  GdkColor *inactive_fg, *inactive_bg;

  if (!property_widget) return;

  dialog = GTK_DIALOG(gtk_dialog_new());
  gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);

  gtk_signal_connect_object (GTK_OBJECT (dialog), "destroy",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));

  gtk_window_set_title (GTK_WINDOW (dialog), "Select Signal");

  clist = gtk_clist_new_with_titles(1, titles);
  gtk_clist_column_titles_passive (GTK_CLIST(clist));
  gtk_clist_set_policy(GTK_CLIST(clist), GTK_POLICY_AUTOMATIC,
		       GTK_POLICY_AUTOMATIC);
  gtk_widget_set_usize(clist, 240, 260);
  gtk_widget_show(clist);
  gtk_box_pack_start(GTK_BOX(dialog->vbox), clist, TRUE, TRUE, 0);

  /* Insert events & descriptions */
  gtk_clist_freeze (GTK_CLIST(clist));

  current_signal = gtk_entry_get_text(GTK_ENTRY(value));
  type = GTK_OBJECT_TYPE(property_widget);
  row = 0;
  inactive_fg = &widget->style->fg[GTK_STATE_INSENSITIVE];
  inactive_bg = &widget->style->bg[GTK_STATE_INSENSITIVE];
  while (type) {
    class = gtk_type_class(type);
    signals = GTK_OBJECT_CLASS(class)->signals;
    nsignals = GTK_OBJECT_CLASS(class)->nsignals;

    name = gtk_type_name(type);
    gtk_clist_append(GTK_CLIST(clist), &name); 
    gtk_clist_set_foreground(GTK_CLIST(clist), row, inactive_fg);
    gtk_clist_set_background(GTK_CLIST(clist), row, inactive_bg);
    gtk_clist_set_row_data(GTK_CLIST(clist), row, "ClassName");
    row++;

    for (i = 0; i < nsignals; i++) {
      query_info = gtk_signal_query(signals[i]);
      name = (gchar*) query_info->signal_name;
      gtk_clist_append(GTK_CLIST(clist), &name);
      gtk_clist_set_shift(GTK_CLIST(clist), row, 0, 0, 10);
      if (!strcmp(current_signal, name)) {
	gtk_clist_select_row(GTK_CLIST(clist), row, 0);
      }
      row++;
      g_free(query_info);
    }
    type = gtk_type_parent(type);
  }

  gtk_clist_thaw (GTK_CLIST(clist));
  gtk_signal_connect (GTK_OBJECT (clist), "select_row",
		      GTK_SIGNAL_FUNC(on_signals_clist_select), clist);

  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(dialog->action_area), button, TRUE, TRUE, 0);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);
  /* Save pointer to value to use when OK pressed */
  gtk_object_set_data(GTK_OBJECT(clist), GbValueKey, value);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC(on_signals_dialog_ok), clist);

  button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(dialog->action_area), button, TRUE, TRUE, 0);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_widget_show(GTK_WIDGET(dialog));
}


void
on_signals_dialog_ok(GtkWidget *widget, GtkWidget *clist)
{
  gint row, page;
  GtkWidget *value, *handler;
  GList *selection = GTK_CLIST(clist)->selection;
  gchar *name, *handler_text;

  if (selection) {
    row = (gint) selection->data;
    value = gtk_object_get_data(GTK_OBJECT(clist), GbValueKey);
    g_return_if_fail(value != NULL);
    gtk_clist_get_text(GTK_CLIST(clist), row, 0, &name);
    gtk_entry_set_text(GTK_ENTRY(value), name);

    /* If we're on the Signals page, and the current handler is empty,
       insert an initial value of "on_<widget name>_<signal name>" */
    page = gtk_notebook_current_page(GTK_NOTEBOOK(main_notebook));
    if (page == GB_PAGE_SIGNALS) {
      handler = (GtkWidget*) g_hash_table_lookup(gb_property_values,
						 GbSignalHandler);
      g_return_if_fail(handler != NULL);
      handler_text = gtk_entry_get_text(GTK_ENTRY(handler));
      
      if (strlen(handler_text) == 0) {
	gchar buf[128];
	gchar *widget_name;
	widget_name = gtk_widget_get_name(property_widget);
	if (strlen(widget_name) + strlen(name) + 5 < 128) {
	  sprintf(buf, "on_%s_%s", widget_name, name);
	  gtk_entry_set_text(GTK_ENTRY(handler), buf);
	  gtk_entry_select_region(GTK_ENTRY(handler), 0, -1);
	}
	gtk_widget_grab_focus(handler);
      }
    }
  }
  gtk_widget_destroy(gtk_widget_get_toplevel(clist));
}


void
on_signals_clist_select(GtkWidget *widget, gint row, gint column,
			GdkEventButton *bevent, gpointer data)
{
  /* Don't allow selection of widget class names */
  if (gtk_clist_get_row_data(GTK_CLIST(widget), row))
    gtk_clist_unselect_row(GTK_CLIST(widget), row, 0);
  if (bevent && bevent->type == GDK_2BUTTON_PRESS) {
    on_signals_dialog_ok(widget, widget);
  }
}


/*
 * Accelerators
 */

void
on_accelerator_add(GtkWidget *widget, GtkWidget *clist)
{
  gchar modifiers[4];
  gchar *row[3];
  gchar *key, *signal;

  key = property_get_string(GbAccelKey);
  if (strlen(key) == 0) {
    show_message_box("You need to set the accelerator key");
    return;
  }
  signal = property_get_string(GbAccelSignal);
  if (strlen(signal) == 0) {
    show_message_box("You need to set the signal to emit");
    return;
  }

  modifiers[0] = modifiers[1] = modifiers[2] = ' ';
  modifiers[3] = '\0';
  if (GTK_TOGGLE_BUTTON(accel_control_button)->active) modifiers[0] = 'C';
  if (GTK_TOGGLE_BUTTON(accel_shift_button)->active)   modifiers[1] = 'S';
  if (GTK_TOGGLE_BUTTON(accel_alt_button)->active)     modifiers[2] = 'A';

  row[0] = modifiers;
  row[1] = key;
  row[2] = signal;
  gtk_clist_append(GTK_CLIST(clist), row);

  /* clear the key & signal fields */
  property_set_string(GbAccelKey, "");
  property_set_string(GbAccelSignal, "");
}


void
on_accelerator_update(GtkWidget *widget, GtkWidget *clist)
{
  gchar modifiers[4];
  gchar *key, *signal;
  GList *selection = GTK_CLIST(clist)->selection;
  gint row;

  if (!selection) return;
  row = (gint) selection->data;

  key = property_get_string(GbAccelKey);
  if (strlen(key) == 0) {
    show_message_box("You need to set the accelerator key");
    return;
  }
  signal = property_get_string(GbAccelSignal);
  if (strlen(signal) == 0) {
    show_message_box("You need to set the signal to emit");
    return;
  }

  modifiers[0] = modifiers[1] = modifiers[2] = ' ';
  modifiers[3] = '\0';
  if (GTK_TOGGLE_BUTTON(accel_control_button)->active) modifiers[0] = 'C';
  if (GTK_TOGGLE_BUTTON(accel_shift_button)->active)   modifiers[1] = 'S';
  if (GTK_TOGGLE_BUTTON(accel_alt_button)->active)     modifiers[2] = 'A';

  gtk_clist_set_text(GTK_CLIST(clist), row, 0, modifiers);
  gtk_clist_set_text(GTK_CLIST(clist), row, 1, key);
  gtk_clist_set_text(GTK_CLIST(clist), row, 2, signal);
}


void
on_accelerator_delete(GtkWidget *widget, GtkWidget *clist)
{
  GList *selection = GTK_CLIST(clist)->selection;
  gint row;

  if (!selection) return;
  row = (gint) selection->data;
  gtk_clist_remove(GTK_CLIST(clist), row);
  /* clear the key & signal fields */
  property_set_string(GbAccelKey, "");
  property_set_string(GbAccelSignal, "");
}


void
on_accelerator_clear(GtkWidget *widget, GtkWidget *clist)
{
  property_set_string(GbAccelKey, "");
  property_set_string(GbAccelSignal, "");

  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (accel_control_button), FALSE);
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (accel_shift_button), FALSE);
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON (accel_alt_button), FALSE);
}


void
on_accelerator_select (GtkWidget *clist, gint row, gint column,
		       GdkEventButton *event, gpointer user_data)
{
  gchar *key, *signal;
  gtk_clist_get_text(GTK_CLIST(clist), row, 1, &key);
  gtk_clist_get_text(GTK_CLIST(clist), row, 2, &signal);
  property_set_string(GbAccelKey, key);
  property_set_string(GbAccelSignal, signal);
}


/*
 * Signals
 */

void
on_signal_add(GtkWidget *widget, GtkWidget *clist)
{
  gchar *row[5];
  gchar *signal, *handler, *object, *data;
  gboolean after;

  signal = property_get_string(GbSignalName);
  if (strlen(signal) == 0) {
    show_message_box("You need to set the signal name");
    return;
  }
  handler = property_get_string(GbSignalHandler);
  if (strlen(handler) == 0) {
    show_message_box("You need to set the handler for the signal");
    return;
  }
  object = property_get_string(GbSignalObject);
  after = property_get_bool(GbSignalAfter);
  data = property_get_string(GbSignalData);

  row[0] = signal;
  row[1] = handler;
  row[2] = object;
  row[3] = after ? "Y": "";
  row[4] = data;
  gtk_clist_append(GTK_CLIST(clist), row);

  /* clear the fields */
  property_set_string(GbSignalName, "");
  property_set_string(GbSignalHandler, "");
  property_set_string(GbSignalObject, "");
  property_set_string(GbSignalData, "");
}


void
on_signal_update(GtkWidget *widget, GtkWidget *clist)
{
  gchar *signal, *handler, *object, *data;
  gboolean after;
  GList *selection = GTK_CLIST(clist)->selection;
  gint row;

  if (!selection) return;
  row = (gint) selection->data;

  signal = property_get_string(GbSignalName);
  if (strlen(signal) == 0) {
    show_message_box("You need to set the signal name");
    return;
  }
  handler = property_get_string(GbSignalHandler);
  if (strlen(handler) == 0) {
    show_message_box("You need to set the handler for the signal");
    return;
  }
  object = property_get_string(GbSignalObject);
  after = property_get_bool(GbSignalAfter);
  data = property_get_string(GbSignalData);

  gtk_clist_set_text(GTK_CLIST(clist), row, 0, signal);
  gtk_clist_set_text(GTK_CLIST(clist), row, 1, handler);
  gtk_clist_set_text(GTK_CLIST(clist), row, 2, object);
  gtk_clist_set_text(GTK_CLIST(clist), row, 3, after ? "Y" : "");
  gtk_clist_set_text(GTK_CLIST(clist), row, 4, data);
}


void
on_signal_delete(GtkWidget *widget, GtkWidget *clist)
{
  GList *selection = GTK_CLIST(clist)->selection;
  gint row;

  if (!selection) return;
  row = (gint) selection->data;
  gtk_clist_remove(GTK_CLIST(clist), row);
  /* clear the key & signal fields */
  property_set_string(GbSignalName, "");
  property_set_string(GbSignalHandler, "");
  property_set_string(GbSignalObject, "");
  property_set_string(GbSignalData, "");
}


void
on_signal_clear(GtkWidget *widget, GtkWidget *clist)
{
  property_set_string(GbSignalName, "");
  property_set_string(GbSignalHandler, "");
  property_set_string(GbSignalObject, "");
  property_set_string(GbSignalData, "");
}


void
on_signal_select (GtkWidget *clist,
		  gint row,
		  gint column,
		  GdkEventButton *event,
		  gpointer user_data)
{
  gchar *signal, *handler, *object, *after, *data;
  gtk_clist_get_text(GTK_CLIST(clist), row, 0, &signal);
  gtk_clist_get_text(GTK_CLIST(clist), row, 1, &handler);
  gtk_clist_get_text(GTK_CLIST(clist), row, 2, &object);
  gtk_clist_get_text(GTK_CLIST(clist), row, 3, &after);
  gtk_clist_get_text(GTK_CLIST(clist), row, 4, &data);
  property_set_string(GbSignalName, signal);
  property_set_string(GbSignalHandler, handler);
  property_set_string(GbSignalObject, data);
  if (!strcmp(after, "Y")) property_set_bool(GbSignalAfter, TRUE);
  else  property_set_bool(GbSignalAfter, FALSE);
  property_set_string(GbSignalData, data);
}


/*
void
on_apply(GtkWidget *widget, gpointer data)
{
  if (!property_widget) return;
  gb_widget_apply_properties(property_widget, NULL);
}
*/

void
property_set_auto_apply(gboolean value)
{
  auto_apply = value;
}

void
on_property_changed(GtkWidget *widget, gpointer data)
{
  /*g_print("In on_property_changed\n");*/
  if (!property_widget || !auto_apply) return;
  /*g_print("   applying...\n");*/
  gb_widget_apply_properties(property_widget, data);
}


/*
 * Adjustments - handles all 6 properties
 */
void
property_add_adjustment(gchar *Values[], gint label_type)
{
  gchar *default_labels[] = { "Value:", "Min:", "Max:", "Step Inc:",
			      "Page Inc:", "Page Size:" };
  gchar *horz_labels[]    = { "H Value:", "H Min:", "H Max:", "H Step Inc:",
			      "H Page Inc:", "H Page Size:" };
  gchar *vert_labels[]    = { "V Value:", "V Min:", "V Max:", "V Step Inc:",
			      "V Page Inc:", "V Page Size:" };
  gchar *tips[] = {
    "The initial position",
    "The minimum position",
    "The maximum position",
    "The step increment",
    "The page increment",
    "The page size",
  };
  gchar **labels;
  gint i;

  if      (label_type == GB_ADJUST_H_LABELS) labels = horz_labels;
  else if (label_type == GB_ADJUST_V_LABELS) labels = vert_labels;
  else labels = default_labels;

  for (i = 0; i < 6; i++)
    property_add_float(Values[i], labels[i], tips[i]);
}


void
property_show_adjustment(gchar *Values[], GtkAdjustment *adjustment)
{
  property_set_float(Values[0], adjustment->value);
  property_set_float(Values[1], adjustment->lower);
  property_set_float(Values[2], adjustment->upper);
  property_set_float(Values[3], adjustment->step_increment);
  property_set_float(Values[4], adjustment->page_increment);
  property_set_float(Values[5], adjustment->page_size);
}


gboolean
property_apply_adjustment(gchar *Values[], GtkAdjustment *adjustment)
{
  gfloat value, lower, upper, step_inc, page_inc, page_size;
  gboolean myApply;

  value = property_get_float(Values[0]);
  myApply = Apply;
  lower = property_get_float(Values[1]);
  myApply |= Apply;
  upper = property_get_float(Values[2]);
  myApply |= Apply;
  step_inc = property_get_float(Values[3]);
  myApply |= Apply;
  page_inc = property_get_float(Values[4]);
  myApply |= Apply;
  page_size = property_get_float(Values[5]);
  myApply |= Apply;

  if (myApply) {
    adjustment->value = value;
    adjustment->lower = lower;
    adjustment->upper = upper;
    adjustment->step_increment = step_inc;
    adjustment->page_increment = page_inc;
    adjustment->page_size = page_size;
    return TRUE;
  }
  return FALSE;
}
